/**
 * This file contains the implementation of #Condition.
 * 
 * @file  Condition.cpp
 * @brief Implementation of Condition
 * 
 * @author Christian Keil
 *
 * $Id: Condition.cpp 533 2008-06-12 20:08:12Z keil $
 * $Name:  $
 */
#include <Condition.h>
#include <Constants.h>
#include <Functions.h>
#include <assert.h>

/**
 * @brief Constructor
 *
 * @param[in] lurupa   Lurupa reference for bound computations.
 * @param[in] preport  Report reference for reporting and debugging.
 */
Condition::Condition(Lurupa *lurupa, Report *preport)
{
  this->lurupa = lurupa;
  this->preport = preport;
} /* Condition::Condition */

/**
 * @brief Destructor
 */
Condition::~Condition()
{
} /* Condition::~Condition */

/**
 * Generate the lp to compute the distance to primal infeasibility. Ordonez and
 * Freund's lp (3.14) translates to the following lp. Note the transformation
 * from @f$y_L@f$ to @f$-y_L@f$.
 * @f{align*}
 *   &\min && \begin{pmatrix} 1 & 0 & 0 & 0 & 0 & 0 & 0 \end{pmatrix}
 *            \begin{pmatrix} x \\ h \\ y_L \\ y_E \\ v \\ s^+ \\ s^- \end{pmatrix}
 *            \\
 *   &s.t. && \begin{pmatrix}
 *              -1 &  e &    0 &    0 &  0 &  0 &  0 \\
 *              -1 &  0 & -a^T &  b^T & -1 &  0 &  0 \\
 *              -1 &  0 &  a^T & -b^T &  1 &  0 &  0 \\
 *               0 & -I & -A^T &  B^T &  0 &  I & -I \\
 *               0 & -I &  A^T & -B^T &  0 & -I &  I \\
 *               0 &  0 &    0 &    0 & -1 & -\underline{x}^T & +\overline{x}^T \\
 *             \end{pmatrix}
 *             \begin{pmatrix} x \\ h \\ y_L \\ y_E \\ v \\ s^+ \\ s^- \end{pmatrix}
 *             \le
 *             \begin{pmatrix} 0 \\ 0 \\ 0 \\ 0 \\ 0 \\ 0 \end{pmatrix} \\
 *   &      && x, h, y_L, s_{L_B}^+, s_{U_B}^- \ge 0 \\
 *   &      && s_{N \setminus L_B}^+, s_{N \setminus U_B}^- = 0
 * @f}
 */
Lp *
Condition::generate_lp_rho_p(const Lp *lp)
{
  int n = Dimension(lp->ix);
  int m = RowDimension(lp->IA);
  int p = RowDimension(lp->IB);
  int vars = 3*n + m + p + 2;
  int cons = 2*n + 4;
  Lp *lp_rho_p = new Lp;
  lp_rho_p->lurupa = lp->lurupa;
  lp_rho_p->infinite = lp->infinite;
  lp_rho_p->non_fixed_vars = vars;
  lp_rho_p->name = new char[strlen(lp->name) + 6 + 1];
  sprintf(lp_rho_p->name, "%s-rho_p", lp->name);
  INTERVAL izero = Hull(0.0);
  INTERVAL ione = Hull(1.0);

  /* Wanna build an lp? Huh? Let's start with the objective function. */
  lp_rho_p->maximize = false;
  Resize(lp_rho_p->ix, vars);
  Resize(lp_rho_p->ic, vars);
  Clear(lp_rho_p->ic);
  lp_rho_p->ic(1) = ione;

  /* now the constraint matrix - we only have inequalities. */
  Resize(lp_rho_p->IA, cons, vars);
  Clear(lp_rho_p->IA);
  INTERVAL_VECTOR row(vars);

  Clear(row);
  row(1) = -ione;
  for (int i = 2; i < 2 + n; ++i) {
    row(i) = ione;
  }
  SetRow(lp_rho_p->IA, 1, row);

  Clear(row);
  row(1) = -ione;
  int j = 2 + n;
  for (int i = 1; i <= m; ++i, ++j) {
    row(j) = -lp->ia(i);
  }
  for (int i = 1; i <= p; ++i, ++j) {
    row(j) = lp->ib(i);
  }
  row(j) = -ione;
  SetRow(lp_rho_p->IA, 2, row);
  row(1) = ione;
  SetRow(lp_rho_p->IA, 3, -row);

  for (int i = 4; i < 4 + n; ++i) {
    Clear(row);
    row(i - 2) = -ione;
    j = 2 + n;
    for (int k = 1; k <= m; ++k, ++j) {
      row(j) = -lp->IA(k, i - 3);
    }
    for (int k = 1; k <= p; ++k, ++j) {
      row(j) = lp->IB(k, i - 3);
    }
    row(  n + m + p + i - 1) =  ione;
    row(2*n + m + p + i - 1) = -ione;
    SetRow(lp_rho_p->IA, i, row);
    row(i - 2) = ione;
    SetRow(lp_rho_p->IA, i + n, -row);
  }

  Clear(row);
  row(n + m + p + 2) = -ione;
  j = n + m + p + 3;
  for (int i = 1; i <= n; ++i, ++j) {
    if (lp->xl(i) != -lp->infinite) row(j) = -lp->xl(i);
  }
  for (int i = 1; i <= n; ++i, ++j) {
    if (lp->xu(i) != lp->infinite) row(j) = lp->xu(i);
  }
  SetRow(lp_rho_p->IA, cons, row);

  /* Off to the right hand side - this one 's easy =) */
  Resize(lp_rho_p->ia, cons);
  Clear(lp_rho_p->ia);

  /* The corresponding duals */
  Resize(lp_rho_p->iy, cons);

  /* Last the simple bounds 
   * Basically all variables are >= 0 */
  Resize(lp_rho_p->xl, vars);
  Clear(lp_rho_p->xl);
  Resize(lp_rho_p->xu, vars);
  Initialize(lp_rho_p->xu, lp_rho_p->infinite);
  /* The s+ / s- multipliers are zero if the corresponding bounds are infinite */
  j = n + m + p + 3;
  for (int i = 1; i <= n; ++i, ++j) {
    if (lp->xl(i) == -lp->infinite) {
      lp_rho_p->xu(j) = 0;
      --lp_rho_p->non_fixed_vars;
    }
    if (lp->xu(i) ==  lp->infinite) {
      lp_rho_p->xu(j + n) = 0;
      --lp_rho_p->non_fixed_vars;
    }
  }
  /* The ys corresponding to equality constraints and v are free */
  lp_rho_p->free_variables_size = p + 1;
  lp_rho_p->free_variables = new int[p+1];
  for (int i = 0, di = n + m + 2; i <= p; ++i) {
    lp_rho_p->xl(i + di) = -lp_rho_p->infinite;
    lp_rho_p->free_variables[i] = i + di;
  }

  /* Now we set the dimensions of the remaining variables */
  Resize(lp_rho_p->IB, 0, vars);
  Resize(lp_rho_p->ib, 0);
  Resize(lp_rho_p->iz, 0);

  return lp_rho_p;
}

/**
 * Swap the elements with index @p i and @p iT of array @p rg.
 *
 * @param[in,out] rg  the array
 * @param[in]     i   index to swap
 * @param[in]     iT  index to swap
 */
void
Condition::swap(int *rg, int i, int iT)
{
  int swap = rg[i];
  rg[i] = rg[iT];
  rg[iT] = swap;
} /* Condition::swap */

/**
 * Rotate the elements with index @p i1, @p i2 and @p i3 of array @p rg clockwise. 
 *
 * @param[in,out] rg  the array
 * @param[in]     i1  index to rotate
 * @param[in]     i2  index to rotate
 * @param[in]     i3  index to rotate
 */
void
Condition::rot(int *rg, int i1, int i2, int i3)
{
  int swap = rg[i1];
  rg[i1] = rg[i3];
  rg[i3] = rg[i2];
  rg[i2]= swap;
} /* Condition::rot */

/**
 * Compute distance to primal infeasibility @f$\rho_p@f$.
 * Consists of the following steps:
 * - generate new lp
 * - tell module to transform lp to solver representation
 * - iterate over i/j settings maintaining enclosure of optimal value, i.e.,
 *   - set y_i = j
 *   - compute lower/upper bound
 *   - update @p enclosure, set to smallest lower/upper bound
 *
 * @param[in]  lp         Lp to compute distance to primal infeasibility for
 * @param[out] enclosure  %Enclosure of distance to primal infeasibility
 */
void
Condition::rho_p(const Lp *lp, INTERVAL &enclosure)
{
  Lp *lp_rho_p = generate_lp_rho_p(lp);
  lurupa->solver_module.lp2solver(lp_rho_p);
  enclosure = Hull(Machine::PosInfinity);

  int dummy;
  int base = 1 + Dimension(lp->ix);
  double optimal;
  double lower_bound, upper_bound;
  bool f;

  for (int i = 1; 
       (Sup(enclosure) > 0) && (i <= RowDimension(lp->IA));
       ++i) {
    preport->print(3, "--=[ Rho_p iteration %d/%d ]=--",
                   i, RowDimension(lp->IA) + RowDimension(lp->IB));
    /* This loop handles the dual parameters of the inequalities of the original
     * problem, thus the RowDimension(lp->IA). These are transformed in the
     * rho_p problem to being greater than zero, so we only need to solve for
     * them fixed to 1. The feasible region is trivially empty for y = -1. */
    preport->print(6, "Fixing y(%d) to 1.0", i);
    lurupa->solver_module.set_bounds(lp_rho_p, base + i, 1.0, 1.0);
    lp_rho_p->xl(base + i) = lp_rho_p->xu(base + i) = 1.0;
    --lp_rho_p->non_fixed_vars;
    lurupa->solve_lp(lp_rho_p, optimal);
    lurupa->lower_bound(lp_rho_p, lower_bound, dummy);
    lurupa->upper_bound(lp_rho_p, upper_bound, dummy);
    // lower_bound = upper_bound = optimal;
    lurupa->solver_module.set_bounds(lp_rho_p, base + i, 0, lp_rho_p->infinite);
    lp_rho_p->xl(base + i) = 0;
    lp_rho_p->xu(base + i) = lp_rho_p->infinite;
    ++lp_rho_p->non_fixed_vars;

    f = false;
    if (lower_bound < Inf(enclosure)) {
      enclosure = Hull(lower_bound, Sup(enclosure));
      f = true;
    }
    if (upper_bound < Sup(enclosure)) {
      enclosure = Hull(Inf(enclosure), upper_bound);
      f |= true;
    }
    if (f) preport->print(5, "Rho_p = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));
  }

  base += RowDimension(lp->IA);
  /* We're fixing a free variable in each loop iteration */
  --lp_rho_p->free_variables_size;
  --lp_rho_p->non_fixed_vars;
  /* We want to swap the first variables _out of_ the free ones, so substitute with
   * free_variables_size and not free_variables_size-1 ! */
  swap(lp_rho_p->free_variables, 0, lp_rho_p->free_variables_size);
  for (int i = 1; 
       i <= RowDimension(lp->IB);
       ++i) {
    preport->print(3, "--=[ Rho_p iteration %d/%d ]=--",
                   i + RowDimension(lp->IA),
                   RowDimension(lp->IA) + RowDimension(lp->IB));
    /* This loop handles the dual parameters for the equations of the original
     * problem, thus the conditional i <= RowDimension(lp->IB). Here we have to
     * solve for them being fixed to 1 and -1. */
    preport->print(6, "Fixing z(%d) to 1.0", i);
    lurupa->solver_module.set_bounds(lp_rho_p, base + i, 1.0, 1.0);
    lp_rho_p->xl(base + i) = lp_rho_p->xu(base + i) = 1.0;
    lurupa->solve_lp(lp_rho_p, optimal);
    lurupa->lower_bound(lp_rho_p, lower_bound, dummy);
    lurupa->upper_bound(lp_rho_p, upper_bound, dummy);
    // lower_bound = upper_bound = optimal;

    f = false;
    if (lower_bound < Inf(enclosure)) {
      enclosure = Hull(lower_bound, Sup(enclosure));
      f = true;
    }
    if (upper_bound < Sup(enclosure)) {
      enclosure = Hull(Inf(enclosure), upper_bound);
      f |= true;
    }
    if (f) {
      preport->print(5, "Rho_p = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));
      if (Sup(enclosure) <= 0) break;
    }

    preport->print(6, "Fixing z(%d) to -1.0", i);
    lurupa->solver_module.set_bounds(lp_rho_p, base + i, -1.0, -1.0);
    lp_rho_p->xl(base + i) = lp_rho_p->xu(base + i) = -1.0;
    lurupa->solve_lp(lp_rho_p, optimal);
    lurupa->lower_bound(lp_rho_p, lower_bound, dummy);
    lurupa->upper_bound(lp_rho_p, upper_bound, dummy);
    // lower_bound = upper_bound = optimal;
    preport->print(6, "Freeing z(%d)", i);
    lurupa->solver_module.set_bounds(lp_rho_p, base + i, -lp_rho_p->infinite, lp_rho_p->infinite);
    lp_rho_p->xl(base + i) = -lp_rho_p->infinite;
    lp_rho_p->xu(base + i) =  lp_rho_p->infinite;

    f = false;
    if (lower_bound < Inf(enclosure)) {
      enclosure = Hull(lower_bound, Sup(enclosure));
      f = true;
    }
    if (upper_bound < Sup(enclosure)) {
      enclosure = Hull(Inf(enclosure), upper_bound);
      f |= true;
    }
    if (f) {
      preport->print(5, "Rho_p = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));
      if (Sup(enclosure) <= 0) break;
    }

    /* We want to rotate the next variable _out of_ the free variables. Hence the 3rd parameter
     * is free_variables_size and not free_variables_size - 1 ! */
    rot(lp_rho_p->free_variables, i - 1, i, lp_rho_p->free_variables_size);
  }
#if 0
  /* We don't need this, lp_rho_p is deleted afterwards. */
  swap(lp_rho_p->free_variables, 
       lp_rho_p->free_variables_size - 2, lp_rho_p->free_variables_size - 1);
  ++lp_rho_p->free_variables_size;
  ++lp_rho_p->non_fixed_vars;
#endif

  if (Sup(enclosure) < 0) enclosure = Hull(Inf(enclosure), 0);
  if (Inf(enclosure) < 0) enclosure = Hull(0, Sup(enclosure));
  preport->print(5, "Rho_p = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));

  delete lp_rho_p;
} /* Condition::rho_p */

/**
 * Generate the lp to compute the distance to dual infeasibility. Ordonez and
 * Freund's lp (3.15) translates to the following lp. Note the transformation
 * from @f$p_L@f$ to @f$-p_L@f$.
 * @f{align*}
 *   &\min && \begin{pmatrix} 1 & 0 &  0 & 0 & 0 & 0 \end{pmatrix}
 *            \begin{pmatrix} y \\ h_L \\ h_E \\ x \\ p \\ g \end{pmatrix}
 *            \\
 *   &s.t. && \begin{pmatrix}
 *              -1 &  e &  e &    0 &  0 &  0 \\
 *              -1 &  0 &  0 &  c^T &  0 &  1 \\
 *              -1 &  0 &  0 & -c^T &  0 & -1 \\
 *               0 & -I &  0 &    A &  I &  0 \\
 *               0 &  0 & -I &    B &  0 &  0 \\
 *               0 & -I &  0 &   -A & -I &  0 \\
 *               0 &  0 & -I &   -B &  0 &  0 \\
 *             \end{pmatrix}
 *             \begin{pmatrix} y \\ h_L \\ h_E \\ x \\ p_L \\ g \end{pmatrix}
 *             \le
 *             \begin{pmatrix} 0 \\ 0 \\ 0 \\ 0 \\ 0 \\ 0 \\ 0 \end{pmatrix} \\
 *   &      && x_{L_B} \ge 0, x_{U_B} \le 0 \\
 *   &      && p \ge 0, g \ge 0 \\
 *   &      && y, h_L, h_E \ge 0 \\
 * @f}
 */
Lp *
Condition::generate_lp_rho_d(const Lp *lp)
{
  int n = Dimension(lp->ix);
  int m = RowDimension(lp->IA);
  int p = RowDimension(lp->IB);
  int vars = 2 + 2*m + p + n;
  int cons = 2*(m + p) + 3;
  Lp *lp_rho_d = new Lp;
  lp_rho_d->name = new char[strlen(lp->name) + 6 + 1];
  sprintf(lp_rho_d->name, "%s-rho_d", lp->name);
  lp_rho_d->lurupa = lp->lurupa;
  lp_rho_d->infinite = lp->infinite;
  lp_rho_d->non_fixed_vars = vars;
  lp_rho_d->free_variables_size = lp->free_variables_size;
  lp_rho_d->free_variables = new int[lp_rho_d->free_variables_size];
  memcpy(lp_rho_d->free_variables, lp->free_variables, lp->free_variables_size * sizeof (int));
  INTERVAL izero = Hull(0.0);
  INTERVAL ione = Hull(1.0);

  /* Wanna build an lp? Huh? Let's start with the objective function. */
  lp_rho_d->maximize = false;
  Resize(lp_rho_d->ix, vars);
  Resize(lp_rho_d->ic, vars);
  Clear(lp_rho_d->ic);
  lp_rho_d->ic(1) = ione;

  /* now the constraint matrix - we only have inequalities. */
  Resize(lp_rho_d->IA, cons, vars);
  Clear(lp_rho_d->IA);
  INTERVAL_VECTOR row(vars);

  Clear(row);
  row(1) = -ione;
  for (int i = 2; i < 2 + m + p; ++i) {
    row(i) = ione;
  }
  SetRow(lp_rho_d->IA, 1, row);

  Clear(row);
  row(1) = -ione;
  int j = 2 + m + p;
  for (int i = 1; i <= n; ++i, ++j) {
    row(j) = lp->ic(i);
  }
  row(vars) = ione;
  SetRow(lp_rho_d->IA, 2, row);
  row(1) = ione;
  SetRow(lp_rho_d->IA, 3, -row);

  for (int i = 4; i < 4 + m; ++i) {
    Clear(row);
    row(i - 2) = -ione;
    j = 2 + m + p;
    for (int k = 1; k <= n; ++k, ++j) {
      row(j) = lp->IA(i - 3, k);
    }
    row(  n + m + p + i - 2) = ione;
    SetRow(lp_rho_d->IA, i, row);
    row(i - 2) = ione;
    SetRow(lp_rho_d->IA, i + m + p, -row);
  }
  for (int i = 4 + m; i < 4 + m + p; ++i) {
    Clear(row);
    row(i - 2) = -ione;
    j = 2 + m + p;
    for (int k = 1; k <= n; ++k, ++j) {
      row(j) = lp->IB(i - m - 3, k);
    }
    SetRow(lp_rho_d->IA, i, row);
    row(i - 2) = ione;
    SetRow(lp_rho_d->IA, i + m + p, -row);
  }

  /* Off to the right hand side - this one 's easy =) */
  Resize(lp_rho_d->ia, cons);
  Clear(lp_rho_d->ia);

  /* The corresponding duals */
  Resize(lp_rho_d->iy, cons);

  /* Last the simple bounds 
   * most variables are >= 0 */
  Resize(lp_rho_d->xl, vars);
  Clear(lp_rho_d->xl);
  Resize(lp_rho_d->xu, vars);
  Initialize(lp_rho_d->xu, lp_rho_d->infinite);
  /* x_{L_B} >= 0, x_{U_B} <= 0 */
  j = 2 + m + p;
  for (int i = 1; i <= n; ++i, ++j) {
    if (lp->xl(i) == -lp->infinite) {
      lp_rho_d->xl(j) = -lp_rho_d->infinite;
      if (lp->xu(i) != lp->infinite) lp_rho_d->xu(j) = 0;
    } else {
      if (lp->xu(i) != lp->infinite) {
        lp_rho_d->xu(j) = 0;
        --lp_rho_d->non_fixed_vars;
      }
    }
  }

  /* Now we set the dimensions of the remaining variables */
  Resize(lp_rho_d->IB, 0, vars);
  Resize(lp_rho_d->ib, 0);
  Resize(lp_rho_d->iz, 0);

  return lp_rho_d;
}

/**
 * Compute distance to dual infeasibility @f$\rho_d@f$. Consists of the following steps:
 * - generate new lp
 * - tell module to transform lp to solver representation
 * - iterate over i/j settings maintaining enclosure of optimal value, i.e.,
 *   - set x_i = j
 *   - compute lower/upper bound
 *   - update @p enclosure, set to smallest lower/upper bound
 *
 * @param[in]  lp         %Lp to compute distance to primal infeasibility for
 * @param[out] enclosure  %Enclosure of distance to primal infeasibility
 */
void
Condition::rho_d(const Lp *lp, INTERVAL &enclosure)
{
  Lp *lp_rho_d = generate_lp_rho_d(lp);
  lurupa->solver_module.lp2solver(lp_rho_d);
  enclosure = Hull(Machine::PosInfinity);

  int dummy;
  int base = 1 + RowDimension(lp->IA) + RowDimension(lp->IB);
  double optimal;
  double lower_bound, upper_bound;
  double xlT, xuT;
  bool f;

  --lp_rho_d->non_fixed_vars;
  int ifree_variables = 0;
  for (int i = 1;
       (Sup(enclosure) > 0) && (i <= Dimension(lp->ix));
       ++i) {
    preport->print(3, "--=[ Rho_d iteration %d/%d ]=--",
                   i, Dimension(lp->ix));
    /* If lower and upper bound are equal they are both 0 and the feasible region is empty for 
     * x = +- 1 .*/
    if (lp_rho_d->xl(base + i) == lp_rho_d->xu(base + i)) continue;

    xlT = lp_rho_d->xl(base + i);
    xuT = lp_rho_d->xu(base + i);
    if ((xlT == -lp_rho_d->infinite) && (xuT == lp_rho_d->infinite)) {
      assert(lp_rho_d->free_variables[ifree_variables] == base + i);
      swap(lp_rho_d->free_variables, ifree_variables, lp_rho_d->free_variables_size - 1);
      --lp_rho_d->free_variables_size;
    }

    /* If there is a finite lower bound on the original i-th variable (i.e., i
     * is in L) the feasible region is trivially empty for x(i) = -1 */
    if (xlT == -lp->infinite) {
      preport->print(6, "Fixing x(%d) to -1.0", i);
      lurupa->solver_module.set_bounds(lp_rho_d, base + i, -1.0, -1.0);
      lp_rho_d->xl(base + i) = lp_rho_d->xu(base + i) = -1.0;

      lurupa->solve_lp(lp_rho_d, optimal);
      lurupa->lower_bound(lp_rho_d, lower_bound, dummy);
      lurupa->upper_bound(lp_rho_d, upper_bound, dummy);
      // lower_bound = upper_bound = optimal;

      f = false;
      if (lower_bound < Inf(enclosure)) {
        enclosure = Hull(lower_bound, Sup(enclosure));
        f = true;
      }
      if (upper_bound < Sup(enclosure)) {
        enclosure = Hull(Inf(enclosure), upper_bound);
        f |= true;
      }
      if (f) {
        preport->print(5, "Rho_d = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));
        if (Sup(enclosure) <= 0) break;
      }
    }

    /* If there is a finite upper bound on the original i-th variable (i.e., i
     * is in U) the feasible region is trivially empty for x(i) = 1 */
    if (xuT == lp->infinite) {
      preport->print(6, "Fixing x(%d) to 1.0", i);
      lurupa->solver_module.set_bounds(lp_rho_d, base + i, 1.0, 1.0);
      lp_rho_d->xl(base + i) = lp_rho_d->xu(base + i) = 1.0;

      lurupa->solve_lp(lp_rho_d, optimal);
      lurupa->lower_bound(lp_rho_d, lower_bound, dummy);
      lurupa->upper_bound(lp_rho_d, upper_bound, dummy);
      // lower_bound = upper_bound = optimal;

      f = false;
      if (lower_bound < Inf(enclosure)) {
        enclosure = Hull(lower_bound, Sup(enclosure));
        f = true;
      }
      if (upper_bound < Sup(enclosure)) {
        enclosure = Hull(Inf(enclosure), upper_bound);
        f |= true;
      }
      if (f) preport->print(5, "Rho_d = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));
    }

    /* Now restore the simple bounds on x(i) */
    preport->print(6, "Freeing x(%d)", i);
    lurupa->solver_module.set_bounds(lp_rho_d, base + i, xlT, xuT);
    lp_rho_d->xl(base + i) = xlT;
    lp_rho_d->xu(base + i) = xuT;
    if ((xlT == -lp_rho_d->infinite) && (xuT == lp_rho_d->infinite)) {
      ++lp_rho_d->free_variables_size;
      swap(lp_rho_d->free_variables, ifree_variables, lp_rho_d->free_variables_size - 1);
    }
  }
#if 0 
  /* We don't need this, lp_rho_d is deleted afterwards */
  ++lp_rho_d->non_fixed_vars;
#endif

  if (Sup(enclosure) < 0) enclosure = Hull(Inf(enclosure), 0);
  if (Inf(enclosure) < 0) enclosure = Hull(0, Sup(enclosure));
  preport->print(5, "Rho_d = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));

  delete lp_rho_d;
} /* Condition::rho_d */

/**
 * Compute an enclosure of the 1 norm of @p ix.
 *
 * @param[in] ix  the vector
 * @return the norm enclosure
 */
INTERVAL
Condition::norm_1(const INTERVAL_VECTOR &ix)
{
  INTERVAL N(0);
  for (int i = Dimension(ix); i > 0; --i) {
    N += Abs(ix(i));
  }
  return N;
} /* Condition::norm_1 */

/**
 * Compute an enclosure of the 1,1 norm of the matrix 
 * @f[
 * \left\| \begin{pmatrix} A\\B \end{pmatrix} \right\|
 * := \text{max}_j \left\| \begin{pmatrix} A\\B \end{pmatrix}_{:j} \right\|_1.
 * @f]
 *
 * @param[in] IA  upper part of the matrix
 * @param[in] IB  lower part of the matrix
 * @return the norm enclosure
 */
INTERVAL
Condition::norm_11(const INTERVAL_MATRIX &IA, const INTERVAL_MATRIX &IB)
{
  INTERVAL N(0), NT(0);
  for (int i = ColDimension(IA); i > 0; --i) {
    NT = norm_1(Col(IA, i)) + norm_1(Col(IB, i));
    N = Hull(Inf(NT) > Inf(N) ? Inf(NT) : Inf(N),
             Sup(NT) > Sup(N) ? Sup(NT) : Sup(N));
  }
  return N;
} /* Condition::norm_11 */

/**
 * Compute an enclosure of the Frobenius norm of the matrix 
 * @f$\left( \begin{smallmatrix} A\\B \end{smallmatrix} \right) @f$.
 *
 * @param[in] IA  upper part of the matrix
 * @param[in] IB  lower part of the matrix
 * @return the norm enclosure
 */
INTERVAL
Condition::norm_F(const INTERVAL_MATRIX &IA, const INTERVAL_MATRIX &IB)
{
  INTERVAL N(0);
  for (int i = RowDimension(IA); i > 0; --i) {
    N += Sqr(Row(IA, i));
  }
  for (int i = RowDimension(IB); i > 0; --i) {
    N += Sqr(Row(IB, i));
  }
  return Sqrt(N);
} /* Condition::norm_F */

/**
 * Compute an enclosure of the L1 norm of the matrix 
 * @f$\left( \begin{smallmatrix} A\\B \end{smallmatrix} \right) @f$.
 *
 * @param[in] IA  upper part of the matrix
 * @param[in] IB  lower part of the matrix
 * @return the norm enclosure
 */
INTERVAL
Condition::norm_L1(const INTERVAL_MATRIX &IA, const INTERVAL_MATRIX &IB)
{
  INTERVAL N(0);
  for (int i = RowDimension(IA); i > 0; --i) {
    N += norm_1(Row(IA, i));
  }
  for (int i = RowDimension(IB); i > 0; --i) {
    N += norm_1(Row(IB, i));
  }
  return N;
} /* Condition::norm_L1 */

#if 0
/* x hat does not generalize easily to interval matrices. Deferred.
VECTOR
Condition::x_hat(const INTERVAL_MATRIX &IA)
{
} /* Condition::x_hat */
#endif

/**
 * Compute an enclosure of the norm of @p lp. This is bounded by
 */
int
Condition::norm(const Lp *lp, INTERVAL &enclosure)
{
  INTERVAL norm_rhs(norm_1(lp->ia));
  norm_rhs += norm_1(lp->ib);

  INTERVAL norm_operator(0, BiasPosInf);
  INTERVAL normT(0);
  normT = norm_11(lp->IA, lp->IB);
  preport->print(5, "Maximum absolute column sum norm of operator = [%.16e, %.16e]", Inf(normT), Sup(normT));
  if (Inf(normT) > Inf(norm_operator)) norm_operator = Hull(Inf(normT), Sup(norm_operator));

  normT = norm_F(lp->IA, lp->IB);
  preport->print(5, "Frobenius Norm of operator = [%.16e, %.16e]", Inf(normT), Sup(normT));
  if (Inf(normT) > Inf(norm_operator)) norm_operator = Hull(Inf(normT), Sup(norm_operator));

  VECTOR v(Dimension(lp->ix));
  Initialize(v, 1.);
  normT = norm_1(lp->IA * v) + norm_1(lp->IB * v);
  preport->print(5, "1-norm of operator * e = [%.16e, %.16e]", Inf(normT), Sup(normT));
  if (Inf(normT) > Inf(norm_operator)) norm_operator = Hull(Inf(normT), Sup(norm_operator));
  /*
     v = x_hat(lp->IA);
     normT = norm_1(lp->IA * v);
     */

  normT = norm_L1(lp->IA, lp->IB);
  preport->print(5, "L1-norm of operator = [%.16e, %.16e]", Inf(normT), Sup(normT));
  if (Sup(normT) < Sup(norm_operator)) norm_operator = Hull(Inf(norm_operator), Sup(normT));
  preport->print(5, "Norm of operator = [%.16e, %.16e]", Inf(norm_operator), Sup(norm_operator));

  enclosure = norm_1(lp->ic);
  preport->print(5, "1-norm of objective vector = [%.16e, %.16e]", Inf(normT), Sup(normT));

  enclosure = Hull(Inf(norm_rhs) > Inf(enclosure) ? Inf(norm_rhs) : Inf(enclosure),
                   Sup(norm_rhs) > Sup(enclosure) ? Sup(norm_rhs) : Sup(enclosure));
  enclosure = Hull(Inf(norm_operator) > Inf(enclosure) ? Inf(norm_operator) : Inf(enclosure),
                   Sup(norm_operator) > Sup(enclosure) ? Sup(norm_operator) : Sup(enclosure));
  preport->print(5, "Norm of problem = [%.16e, %.16e]", Inf(enclosure), Sup(enclosure));

  return 0;
} /* Condition::norm */

void
Condition::cond(const Lp *lp, INTERVAL &enclosure)
{
  INTERVAL norm;
  INTERVAL rho_p;
  INTERVAL rho_d;

  preport->print(3, "Bounding problem norm");
  Condition::norm(lp, norm);

  preport->print(3, "Computing distance to primal infeasibility");
  Condition::rho_p(lp, rho_p);
  preport->print(3, "Computing distance to dual infeasibility");
  Condition::rho_d(lp, rho_d);

  if ((Inf(rho_d) == 0) || (Inf(rho_p) == 0)) {
    enclosure = Hull(Inf(norm) / (Sup(rho_p) < Sup(rho_d) ? Sup(rho_p) : Sup(rho_d)),
                     BiasPosInf);
  } else {
    enclosure = norm / Hull(Inf(rho_p) < Inf(rho_d) ? Inf(rho_p) : Inf(rho_d),
                            Sup(rho_p) < Sup(rho_d) ? Sup(rho_p) : Sup(rho_d));
    
  }
} /* Condition::cond */
