/**
 * This file contains the definition of #Condition.
 * 
 * @file  Condition.h
 * @brief Definition of Condition
 * 
 * @author Christian Keil
 *
 * $Id: Condition.h 533 2008-06-12 20:08:12Z keil $
 * $Name:  $
 */
#ifndef CONDITION_H
#define CONDITION_H

#include <lurupa/Lurupa.h>
#include <lurupa/Lp.h>
#include <lurupa/globals.h>
#include <IntervalVector.h>

/**
 * @class Condition Condition.h lurupa/Condition.h
 *
 * @brief %Condition numbers
 *
 * This class implements the computation of verified condition numbers. It is based on Ordonez and
 * Freund's \cite OrdFre  linear programming characterization of the condition number. This in turn
 * can be bounded with Lurupa.
 */
class Condition {
public:
  Condition(Lurupa *lurupa, Report *report);
  ~Condition();

	/** @brief Compute distance to primal infeasibility. */
  void rho_p(const Lp *lp, INTERVAL &enclosure);
	/** @brief Compute distance to dual infeasibility. */
  void rho_d(const Lp *lp, INTERVAL &enclosure);
	/** @brief Compute verified condition number. */
  void cond(const Lp *lp, INTERVAL &enclosure);

private:
  Lurupa *lurupa;   /**< @brief Lurupa reference for bound computation. */
  Report *preport;  /**< @brief Reporting and debugging. */

	/** @brief Swap elements of array. */
  void swap(int *rg, int i, int iT);
	/** @brief Rotate elements of array. */
  void rot(int *rg, int i1, int i2, int i3);

	/** @brief Generate lp to compute distance to primal infeasibility. */
  Lp *generate_lp_rho_p(const Lp *lp);
	/** @brief Generate lp to compute distance to dual infeasibility. */
  Lp *generate_lp_rho_d(const Lp *lp);

	/** @brief %Enclosure of the norm of an lp. */
  int norm(const Lp *lp, INTERVAL &enclosure);

	/** @brief %Enclosure of the 1 norm of a vector. */
  INTERVAL norm_1(const INTERVAL_VECTOR &ix);
	/** @brief %Enclosure of the 1,1 operator norm of a matrix. */
  INTERVAL norm_11(const INTERVAL_MATRIX &IA, const INTERVAL_MATRIX &IB);
	/** @brief %Enclosure of the Frobenius norm of a matrix. */
  INTERVAL norm_F(const INTERVAL_MATRIX &IA, const INTERVAL_MATRIX &IB);
	/** @brief %Enclosure of the L1 norm of a matrix. */
  INTERVAL norm_L1(const INTERVAL_MATRIX &IA, const INTERVAL_MATRIX &IB);

#if 0
	/** @brief Generate vector x hat. */
	/* This does not easily generalize to interval matrices. Deferred. */
  VECTOR x_hat(const INTERVAL_MATRIX &IA);
#endif
}; /* class Condition */

#endif /* CONDITION_H */
