/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This storage class agregates all the model data and the lp-solver's judgement
 * of its feasibility.
 *
 * @file  Lp.cpp
 * @brief Definition of %Lp storage class
 *
 * @author Christian Keil
 *
 * $Id: Lp.cpp 530 2008-06-12 08:57:52Z keil $
 */
#include <lurupa/Lp.h>

Lp::Lp()
{
  free_variables_size = 0;
  free_variables = NULL;
  non_fixed_vars = 0;
  name = NULL;
  module = NULL;
  lurupa = NULL;
};

Lp::Lp(Lurupa *lurupa)
{
  free_variables_size = 0;
  free_variables = NULL;
  non_fixed_vars = 0;
  name = NULL;
  module = NULL;
	this->lurupa = lurupa;
};

Lp::~Lp()
{
  if (name != NULL) {
    free(name);
  }
  if (free_variables != NULL) {
    delete [] free_variables;
  }
  delete module;
};

void
Lp::set_lurupa(Lurupa *lurupa)
{
  this->lurupa = lurupa;
}

bool
Lp::is_consistent() const
{
  bool fconsistent = true;

  /* First start with a dimension check */
  int n = Dimension(ix);
  int m = RowDimension(IA);
  int p = RowDimension(IB);

  if (ColDimension(IA) != n) {
    lurupa->report.print(1, "ColDimension(IA) = %d, expected %d", ColDimension(IA), n);
    fconsistent = false;
  }
  if (ColDimension(IB) != n) {
    lurupa->report.print(1, "ColDimension(IB) = %d, expected %d", ColDimension(IB), n);
    fconsistent = false;
  }
  if (Dimension(ia) != m) {
    lurupa->report.print(1, "Dimension(ia) = %d, expected %d", Dimension(ia), m);
    fconsistent = false;
  }
  if (Dimension(ib) != p) {
    lurupa->report.print(1, "Dimension(ib) = %d, expected %d", Dimension(ib), p);
    fconsistent = false;
  }
  if (Dimension(ic) != n) {
    lurupa->report.print(1, "Dimension(ic) = %d, expected %d", Dimension(ic), n);
    fconsistent = false;
  }
  if (Dimension(xl) != n) {
    lurupa->report.print(1, "Dimension(xl) = %d, expected %d", Dimension(xl), n);
    fconsistent = false;
  }
  if (Dimension(xu) != n) {
    lurupa->report.print(1, "Dimension(xu) = %d, expected %d", Dimension(xu), n);
    fconsistent = false;
  }
  if (Dimension(iy) != m) {
    lurupa->report.print(1, "Dimension(iy) = %d, expected %d", Dimension(iy), m);
    fconsistent = false;
  }
  if (Dimension(iz) != p) {
    lurupa->report.print(1, "Dimension(iz) = %d, expected %d", Dimension(iz), p);
    fconsistent = false;
  }

  bool f = false, fT = false;
  int cnon_fixed = 0;
  for (int i = n; i > 0; --i) {
    if ((xl(i) == -infinite) && (xu(i) == infinite)) f = true;

    for (int j = free_variables_size - 1; j >= 0; --j) {
      if (free_variables[j] == i) {
        fT = true;
        break;
      }
    }

    if (f != fT) {
      lurupa->report.print(1, "Variable %d is %sfree but %scontained in free_variables",
                             i, (f ? "" : "not "), (fT ? "" : "not "));
      fconsistent = false;
    }

    f = false;
    fT = false;

    if (xl(i) != xu(i)) ++cnon_fixed;
  }
  if (cnon_fixed != non_fixed_vars) {
    lurupa->report.print(1, "non_fixed_vars = %d, expected %d", non_fixed_vars, cnon_fixed);
    fconsistent = false;
  }
  
  /* fconsistent &= module->is_consistent(); */

  return fconsistent;
}
