/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This storage class aggregates all the model data and the lp-solver's
 * judgement of its feasibility.
 *
 * @file  Lp.h
 * @brief Declaration of %Lp storage class
 *
 * @author Christian Keil
 *
 * $Id: Lp.h 533 2008-06-12 20:08:12Z keil $
 */
#ifndef LURUPA_LP_H
#define LURUPA_LP_H

#include <lurupa/globals.h>
#include <lurupa/Lurupa.h>
#include <lurupa/Solver_module.h>
#include <IntervalMatrix.h>

#include <IntervalVector.h>

/**
 * @struct Primal_deflation Lp.h lurupa/Lp.h
 *
 * @brief Primal deflation parameters
 * 
 * This structure stores the primal deflation parameters for the lp.
 */
struct Primal_deflation {
  VECTOR a;   /**< @brief Deflation parameter for the right hand side of the
                inequalities */ 
  VECTOR xl;  /**< @brief Deflation parameter for the simple lower bounds */
  VECTOR xu;  /**< @brief Deflation parameter for the simple upper bounds */
};

/**
 * @class Lp Lp.h lurupa/Lp.h
 *
 * @brief A linear program
 *
 * This class is Lurupa's representation of an lp.
 * The problem is stored in PROFILs interval types.
 */
class LU_LURUPA_SCOPE Lp {
public:
  INTERVAL_MATRIX IA;         /**< @brief The inequality matrix */
  INTERVAL_MATRIX IB;         /**< @brief The equation matrix */

  INTERVAL_VECTOR ia;         /**< @brief Right hand sides of the inequalities */ 
  INTERVAL_VECTOR ib;         /**< @brief Right hand sides of the equations */
  INTERVAL_VECTOR ic;         /**< @brief The cost vector */
  VECTOR          xl;         /**< @brief The simple lower bound */
  VECTOR          xu;         /**< @brief The simple upper bound */
  INTERVAL_VECTOR ix;         /**< @brief Solution of the last solved problem
                                (changes during algorithm as perturbed problems are solved) */
  INTERVAL_VECTOR iy;         /**< @brief Duals of the inequalities of the last solved problem
                                (changes during algorithm as perturbated problems are solved) */
  INTERVAL_VECTOR iz;         /**< @brief Duals of the equations of the last solved problem 
                                (changes during algorithm as perturbated problems are solved) */
  REAL infinite;              /**< @brief Used lp-solvers representation of infinity 
                                (e.g. as bound value) */
  int free_variables_size;    /**< @brief Number of free variables in model */
  int *free_variables;        /**< @brief The indices of the free variables */
  int non_fixed_vars;         /**< @brief Number of non fixed variables in model */
  Solver_status feasibility;  /**< @brief Lp-solver's judgement of problem's feasibility */
  bool maximize;              /**< @brief True if objective function is to be maximized, false
                                otherwise  */
  char *name;                 /**< @brief Name of the lp model */

  Lp_solver *module;          /**< @brief Additional storage for module data */

  Lurupa *lurupa;             /**< @brief Reference to #Lurupa instance */

  /** @brief Constructor */
  Lp();
  /** @brief Constructor */
	Lp(Lurupa *lurupa);

  /** @brief Destructor */
  ~Lp();

  /** @brief Set #Lurupa instance. */
  void set_lurupa(Lurupa *lurupa);

	/** @brief Check %lp for consistency. */
  bool is_consistent() const;
};

#endif /* LURUPA_LP_H */
