/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file contains the declaration of #Lurupa.
 * 
 * @file  Lurupa.h
 * @brief Declaration of %Lurupa's core
 * 
 * @author Christian Keil
 *
 * $Id: Lurupa.h 533 2008-06-12 20:08:12Z keil $
 */
#ifndef LURUPA_LURUPA_H
#define LURUPA_LURUPA_H

#include <lurupa/Solver_module.h>
#include <lurupa/Report.h>
#include <lurupa/globals.h>
#include <IntervalMatrix.h>
#include <IntervalVector.h>
#include <Vector.h>
#include <ltdl.h>

#ifndef LU_LURUPA_SCOPE
#  if defined(_WIN32) || defined(WIN32)
#    ifdef LU_LURUPA_IMPORT
#      define LU_LURUPA_SCOPE __declspec(dllimport)
#    elif defined DLL_EXPORT
#      define LU_LURUPA_SCOPE __declspec(dllexport)
#    endif
#  endif
#  ifndef LU_LURUPA_SCOPE
#    define LU_LURUPA_SCOPE
#  endif
#endif

class Condition;
class Lp;

/**
 * @class Lurupa Lurupa.h lurupa/Lurupa.h
 *
 * @brief Lurupa's core
 *
 * This class implements Lurupa's core logic. It implements the interface
 * visible to the user as well as the algorithm logic to compute the verified
 * bounds using the interface to the solver modules.
 */
class LU_LURUPA_SCOPE Lurupa {
public:
  Lurupa();
  ~Lurupa();

  /** @brief Get core version string */
  const char *get_core_version() const;
  /** @brief Print core information */
  void print_core_version() const;
  /** @brief Print brief core information */
  void print_core_brief_version() const;
  /** @brief Print module information */
  void print_module_version() const;
  /** @brief Print brief module information */
  void print_module_brief_version() const;

  /*@{*/
  /** @brief Get algorithm parameter alpha */
  double get_alpha() const;
  /** @brief Set algorithm parameter alpha */
  void set_alpha(double alpha);

  /** @brief Get algorithm parameter eta */
  double get_eta() const;
  /** @brief Set algorithm parameter eta */
  void set_eta(double eta);

  /** @brief Is inflation set */
  bool is_inflate() const;
  /** @brief Set inflation */
  void set_inflate(bool inflate);
  /*@}*/

  /*@{*/
  /** @brief Set solver module */
  bool set_module(char *module_path);

  /** @brief Set solver module */
  void set_interface(Solver_module_interface module);

  /** @brief Get solver module version */
  const char *get_module_version() const;

  /** @brief Print options supported by module */
  void print_module_options() const;

  /** @brief Get solver's accuracy */
  double get_solver_eps() const;

  /** @brief Read linear program */
  Lp *read_lp(FILE *in, const double &relative_interval_radius);

  /** @brief Set linear program's solution */
  bool set_solution(Lp *lp, void *p);

  /** @brief Set linear program from solver */
  bool set_lp(Lp *lp, const double relative_interval_radius);

	/** @brief Set linear program from Lurupa */
	void lp2solver(Lp *lp);

  /** @brief Set solver options */
  bool set_module_options(Lp *lp, int argc, char *argv[]);

  /** @brief Solve linear program */
  Solver_status solve_lp(Lp *lp, double &optimal_value);
  /*@}*/

  /*@}*/
	/** @brief Compute distance to primal infeasibility. */
  int rho_p(Lp *lp, double &lower, double &upper);
	/** @brief Compute distance to dual infeasibility. */
  int rho_d(Lp *lp, double &lower, double &upper);
	/** @brief Compute verified condition number. */
  int cond(Lp *lp, double &lower, double &upper);
  /*@}*/

  /** @brief Compute lower bound */
  Bound_status lower_bound(Lp *lp, double &bound, int &iterations);

  /** @brief Compute upper bound */
  Bound_status upper_bound(Lp *lp, double &bound, int &iterations);

  /** @brief Compute certificate for primal infeasibility */
  Bound_status dual_certificate(Lp *lp);

  /** @brief Compute certificate for dual infeasibility */
  Bound_status primal_certificate(Lp *lp);

  /** @brief Get linear program's name */
  char *get_lp_name() const;

  /** @brief Return whether linear program is maximized */
  bool is_lp_maximize() const;

  Report report;          /**< @brief Reporting and debugging */

  friend class Condition;
  friend class Lp;

private:
  /** @brief Interface to solver module */
  Solver_module_interface solver_module;

  /** @brief Handle to solver module object */
  lt_dlhandle mod_handle;

  int iteration;          /**< @brief Current iteration count */
  int max_iterations;     /**< @brief Upper limit for iterations */
  bool inflate;           /**< @brief Try inflation to compute bound */
  double eta;             /**< @brief Algorithm parameter */
  double eps;             /**< @brief Algorithm parameter */
  double alpha;           /**< @brief Algorithm parameter */

  /** @name Find basis
   * @brief Functions to find basis
   *
   * These functions are used to find a regular submatrix to rigorously verify
   * the primal or dual constraints.
   */
  /*@{*/
  /** @brief Find pivot element */
  double pivot_element(MATRIX LU, int step, int &pivot_index);
  /** @brief Do the pivot */
  void do_pivot(MATRIX &LU, int pivot_index, int step, int *basis_indices);
  /** @brief Perform an elimination step */
  void eliminate(MATRIX &LU, int step);
  /** @brief Add column to basis */
  void add_column_to_basis(int step, INTERVAL_MATRIX &A,
                           INTERVAL_MATRIX &IBasis, int *basis_indices);
  /** @brief Shorten basis indices array */
  void shorten_basis_indices(int *&basis_indices, int basis_size);
  /** @brief Find basis for linear system */
  bool find_basis(INTERVAL_MATRIX &A,
                  int *&basis_indices, int size_basis_indices,
                  INTERVAL_MATRIX &IBasis, MATRIX &Basis_mid_inverse);
  /*@}*/

  /** @name Primal feasibility
   * @brief Establish primal feasibility
   *
   * These functions are used to establish a primal feasible solution starting
   * from an approximate one delivered by the solver.
   */
  /*@{*/
  /** @brief Establish simple bounds on Lp#ix */
  void establish_ix_simple_bounds(Lp *lp);
  /** @brief Establish equality constraints */
  bool establish_equality_constraints(Lp *lp, const INTERVAL_MATRIX &IBasis,
                                      const MATRIX &Basis_mid_inverse,
                                      const INTERVAL_MATRIX &IResidual,
                                      const int *basis_indices);
  /** @brief Check for primal feasibility */
  bool primal_feasible(Lp *lp);
  /** @brief Establish primal feasibility */
  Bound_status establish_primal_feasibility(Lp *lp,
                                            const INTERVAL_MATRIX &IBasis,
                                            const MATRIX &Basis_mid_inverse,
                                            const INTERVAL_MATRIX &IResidual,
                                            const int *basis_indices);
  /*@}*/

  /** @name Dual feasibility
   * @brief Establish dual feasibility
   *
   * These functions are used to establish a dual feasible solution starting
   * from an approximate one delivered by the solver.
   * */
  /*@{*/
  /** @brief Establish simple bounds on Lp#iy */
  void establish_iy_simple_bounds(Lp *lp);
  /** @brief Process free variables */
  Bound_status process_free_variables(Lp *lp, const INTERVAL_MATRIX &IBasis,
                                      const MATRIX &Basis_mid_inverse,
                                      const INTERVAL_MATRIX &IResidual,
                                      const int *basis_indices);
  /** @brief Process bounded variables */
  Bound_status process_bounded_variables(Lp *lp, INTERVAL_VECTOR &id_pos,
                                         INTERVAL_VECTOR &id_neg);
  /** @brief Establish lagrange parameteres */
  Bound_status establish_lagrange_parameters(Lp *lp, const INTERVAL_MATRIX &IBasis,
                                             const MATRIX &Basis_mid_inverse,
                                             const INTERVAL_MATRIX &IResidual,
                                             const int *basis_indices,
                                             INTERVAL_VECTOR &id_pos,
                                             INTERVAL_VECTOR &id_neg);
  /** @brief Establish dual feasibility */
  Bound_status establish_dual_feasibility(Lp *lp, const INTERVAL_MATRIX &IBasis,
                                          const MATRIX &Basis_mid_inverse,
                                          const INTERVAL_MATRIX &IResidual,
                                          const int *basis_indices,
                                          INTERVAL_VECTOR &id_pos,
                                          INTERVAL_VECTOR &id_neg);
  /*@}*/

  /** @name Upper bound
   * @brief Compute minimizing upper bound
   *
   * These functions are used to compute an upper bound for the minimal
   * objective value.
   */
  /*@{*/
  /** @brief Find equation basis */
  bool find_equation_basis(Lp *lp, INTERVAL_MATRIX &IResidual, int *&basis_indices, 
                           INTERVAL_MATRIX &IBasis, MATRIX &Basis_mid_inverse);
  /** @brief Process initial primal solution */
  void process_initial_primal_solution(Lp *lp, Bound_status &status, double &bound,
                                       const INTERVAL_MATRIX &IBasis,
                                       const MATRIX &Basis_mid_inverse,
                                       const INTERVAL_MATRIX &IResidual,
                                       const int *basis_indices);
  /** @brief Compute primal deflation parameters */
  void compute_primal_deflation(Lp *lp, Primal_deflation &d);
  /** @brief Increase primal deflation parameters */
  void increase_primal_deflation(Lp *lp, Primal_deflation &d);
  /** @brief Decrease primal deflation parameters */
  void decrease_primal_deflation(Lp *lp, Primal_deflation &d);
  /** @brief Process perturbed primal solution */
  void process_perturbed_primal_solution(Lp *lp, Bound_status &status, double &bound,
                                         const INTERVAL_MATRIX &IBasis,
                                         const MATRIX &Basis_mid_inverse,
                                         const INTERVAL_MATRIX &IResidual,
                                         const int *basis_indices,
                                         Primal_deflation &deflation);
  /** @brief Compute minimizing upper bound */
  Bound_status compute_upper(Lp *lp, double &bound);
  /*@}*/

  /** @name Lower bound
   * @brief Compute minimizing lower bound
   *
   * These functions are used to compute a lower bound for the minimal objective
   * value.
   */
  /*@{*/
  /** @brief Find constraint basis */
  bool find_constraint_basis(Lp *lp, INTERVAL_MATRIX &IResidual, int *&basis_indices,
                             INTERVAL_MATRIX &IBasis, MATRIX &Basis_mid_inverse);
  /** @brief Process initial dual solution */
  void process_initial_dual_solution(Lp *lp, Bound_status &status, double &bound,
                                     const INTERVAL_MATRIX &IBasis,
                                     const MATRIX &Basis_mid_inverse,
                                     const INTERVAL_MATRIX &IResidual,
                                     const int *basis_indices);
  /** @brief Compute dual deflation parameters */
  void compute_dual_deflation(Lp *lp, VECTOR &d_c);
  /** @brief Increase dual deflation parameters */
  void increase_dual_deflation(Lp *lp, VECTOR &d_c,
                               const INTERVAL_VECTOR &id_pos,
                               const INTERVAL_VECTOR &id_neg);
  /** @brief Decrease dual deflation parameters */
  void decrease_dual_deflation(Lp *lp, VECTOR &d_c);
  /** @brief Process perturbed dual solution */
  void process_perturbed_dual_solution(Lp *lp, Bound_status &status, double &bound,
                                       const INTERVAL_MATRIX &IBasis,
                                       const MATRIX &Basis_mid_inverse,
                                       const INTERVAL_MATRIX &IResidual,
                                       const int *basis_indices,
                                       VECTOR &deflation_c);
  /** @brief Compute minimizing lower bound */
  Bound_status compute_lower(Lp *lp, double &bound);
  /*@}*/

	/** @brief Set primal phase 1 %lp */
  void set_primal_phase1(Lp *lp, bool *maximize);
	/** @brief Restore %lp after primal phase 1 approach */
  void restore_primal_phase1(Lp *lp, const INTERVAL_VECTOR &icT, const bool maxmimize);
	/** @brief Set dual phase 1 %lp */
  void set_dual_phase1(Lp *lp);
	/** @brief Restore %lp after dual phase 1 approach */
  void restore_dual_phase1(Lp *lp, const INTERVAL_VECTOR &iaT, const INTERVAL_VECTOR &ibT, 
                           const VECTOR &xlT, const VECTOR &xuT, const int non_fixed_varsT,
                           const int free_variables_sizeT);

}; /* class Lurupa */

#endif /* LURUPA_LURUPA_H */
