/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file contains the declaration of the Report class.
 * 
 * @file  Report.h
 * @brief Declaration of %Report class
 *
 * @author Christian Keil
 *
 * $Id: Report.h 530 2008-06-12 08:57:52Z keil $
 */
#ifndef LURUPA_REPORT_H
#define LURUPA_REPORT_H

#include <lurupa/globals.h>
#include <Vector.h>
#include <Matrix.h>
#include <stdarg.h>

#ifndef LU_REPORT_SCOPE
#  if defined(_WIN32) || defined(WIN32)
#    ifdef LU_REPORT_IMPORT
#      define LU_REPORT_SCOPE __declspec(dllimport)
#    elif defined DLL_EXPORT
#      define LU_REPORT_SCOPE __declspec(dllexport)
#    endif
#  endif
#  ifndef LU_REPORT_SCOPE
#    define LU_REPORT_SCOPE
#  endif
#endif

/**
 * @class Report Report.h lurupa/Report.h
 *
 * @brief Functions for reporting and debugging 
 */
class LU_REPORT_SCOPE Report {
public:
  /** @brief Constructor */
  Report();

  /** @brief Set verbosity */
  void set_verbosity(short level, bool print_time, bool write_vm);

  /** @brief Set verbosity level */
  void set_level(short level);
  /** @brief Get verbosity level */
  short get_level() const;

  /** @brief Set time prepending */
  void set_print_time(bool print_time);
  /** @brief Get time prepending */
  bool get_print_time() const;

  /** @brief Set vector and matrix writing */
  void set_write_vm(bool write_vm);
  /** @brief Get vector and matrix writing */
  bool get_write_vm() const;

  /** @brief Set csv style */
  void set_csv_style(Csv_style style);
  /** @brief Get csv style */
  Csv_style get_csv_style() const;

  /** @brief Set destination of normal messages */
  void set_normal(FILE *normal);

  /** @brief Set destination of warning messages */
  void set_warning(FILE *warning);

  /** @brief Set destination of error messages */
  void set_error(FILE *error);

#ifdef WITH_MATLAB
	/** @brief Set whether running under mex interface */
	void set_mex(bool mex, int (*mexPrintf)(const char *,...), void (*mexWarnMsgTxt)(const char *));
	/** @brief Get whether running under mex interface */
	bool is_mex();
#endif

  /** @brief Print message */
  void print(short type, const char *format, ...) const;

  /** @brief Write vector to disk */
  void write_vector(const char *file, const VECTOR &x, const char *name) const;

  /** @brief Write matrix to disk */
  void write_matrix(const char *file, const MATRIX &X, const char *name) const;

private:
  bool       print_time; /**< @brief Prepend time to messages */
  bool       write_vm;   /**< @brief Write vectors and matrices to disk */
  short      level;      /**< @brief Maximum level of messages to be printed */
  Csv_style  csv_style;  /**< @brief Use which csv style */
  FILE      *normal;     /**< @brief Destination of normal messages */
  FILE      *warning;    /**< @brief Destination of warning messages */
  FILE      *error;      /**< @brief Destination of error messages */

#ifdef WITH_MATLAB
	bool mex;                             /**< @brief Running under mex interface */
	int (*mexPrintf)(const char *, ...);  /**< @brief Pointer to mexPrintf */
	void (*mexWarnMsgTxt)(const char *);  /**< @brief Pointer to mexWarnMsgTxt */
#endif
			  
}; /* class Report */
#endif /* LURUPA_REPORT_H */
