/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file contains the interface declaration for solver modules.
 *
 * @file  Solver_module.h
 * @brief Declaration of Solver module interface
 *
 * @author Christian Keil
 *  
 * $Id: Solver_module.h 533 2008-06-12 20:08:12Z keil $
 */

#ifndef LURUPA_SOLVER_MODULE_H
#define LURUPA_SOLVER_MODULE_H

#include <lurupa/globals.h>
#include <lurupa/Report.h>
#include <Vector.h>

struct Primal_deflation;
class Lp;

/**
 * @class Lp_solver
 *
 * @brief Information about solver specific %lp representation.
 */
class Lp_solver {
public:
  virtual ~Lp_solver() {};
};

/**
 * @struct Solver_module_interface Solver_module.h lurupa/Solver_module.h
 *
 * @brief Function pointer version of %Solver_module
 *
 * This struct is a function pointer version of Solver_module.
 */
struct Solver_module_interface {
  const char *(*get_version)           ();
  void        (*print_version)         ();
  void        (*print_brief_version)   ();
  bool        (*init)                  (Report *);
  void        (*print_options)         ();
  bool        (*solve_primal_perturbed)(const Primal_deflation &, Lp *, const int);
  bool        (*solve_dual_perturbed)  (const VECTOR &, Lp *, const int);
  bool        (*solve_original)        (Lp *, double &);
  bool        (*get_dual_ray)          (Lp *, int &);
  bool	      (*get_primal_ray)	       (Lp *, int &);
  void        (*set_primal_phase1)	   (Lp *);
  void		    (*set_dual_phase1)	     (Lp *);
  void        (*restore_primal)        (Lp *);
  void        (*restore_dual)          (Lp *);
  void		    (*restore_primal_phase1) (Lp *);
  void        (*restore_dual_phase1)	 (Lp *);
  bool        (*read_lp)               (Lp *, FILE *, const double, double &);
  bool        (*set_solution)	         (Lp *, void *);
  bool        (*set_lp)		             (Lp *, const double, double &);
  bool        (*set_module_options)    (Lp *, int, char *[]);
  double      (*get_accuracy)          ();
  void        (*lp2solver)             (Lp *);
  void        (*set_bounds)            (Lp *, int, double, double);
};

/**
 * @class Solver_module Solver_module.h lurupa/Solver_module.h
 *
 * @brief Interface for an lp-solver module
 *
 * This class is the prototype for the lp-solver modules. The interface between
 * Lurupa and the lp-solver is specified.
 */
class Solver_module {
public:
  /** @brief Get module version string */
  static const char *get_version();

  /** @brief Print version information */
  static void print_version();
  /** @brief Print brief version information */
  static void print_brief_version();

  /** @brief Initialize the module */
  static bool init(Report *preport);

  /*@{*/
  /** @brief Get accuracy used by solver */
  static double get_accuracy();

  /** @brief Set module options */
  static bool set_module_options(int argc, char *argv[]);

  /** @brief Read an lp */
  static bool read_lp(Lp *lp, FILE *in, const double relative_interval_radius, double &eta);
  /*@}*/

  /** @brief Print options supported by module */
  static void print_options();

#if 0
  static bool set_solution(Lp *lp, void *p);
#endif

  /** @brief Set an lp */
  static bool set_lp(Lp *lp, const double relative_interval_radius, double &eta);

  /** @brief Compute solution of unperturbed lp */
  static bool solve_original(Lp *lp, double &optimal_value);

  /** @brief Compute solution of primal perturbed lp */
  static bool solve_primal_perturbed(const Primal_deflation &deflation,
                                     Lp *lp, const int iteration);

  /** @brief Compute solution of dual perturbed lp */
  static bool solve_dual_perturbed(const VECTOR &deflation_c, Lp *lp,
                                   const int iteration);

  /** @brief Get dual improving ray */
  static bool get_dual_ray(Lp *lp, int &iyzray);

  /** @brief Get primal improving ray */
  static bool get_primal_ray(Lp *lp, int &ixray);

  /** @brief Set primal phase 1 lp */
  static void set_primal_phase1(Lp *lp);

  /** @brief Set dual phase 1 lp */
  static void set_dual_phase1(Lp *lp);

  /** @brief Restore lp after primal perturbation */
  static void restore_primal(Lp *lp);

  /** @brief Restore lp after dual perturbation */
  static void restore_dual(Lp *lp);

  /** @brief Restore lp after primal phase 1 perturbation */
  static void restore_primal_phase1(Lp *);

  /** @brief Restore lp after dual phase 1 perturbation */
  static void restore_dual_phase1(Lp *);

	/** @brief Transform an lp into solver's representation */
  static void lp2solver(Lp *lp_lurupa);

	/** @brief Set simple bounds on a variable */
  static void set_bounds(Lp *lp, int var, double lower, double upper);

protected:
  static Report *preport;  /**< @brief pointer to report class instance */

private:
  Solver_module();
  ~Solver_module();
}; /* Solver_module */

extern "C" {
  /** @brief Get pointers to member functions */
  void getFuncPointers(Solver_module_interface &i);
}

#endif /* ifdef LURUPA_SOLVER_MODULE_H */

