/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * @file  cLurupa.h
 * @brief Declaration of C wrapper for %Lurupa
 *
 * This is work in progess.
 *
 * @author Christian Keil
 *
 * $Id: cLurupa.h 533 2008-06-12 20:08:12Z keil $
 */
#ifndef CLURUPA_H
#define CLURUPA_H

#include <lurupa/globals.h>
#include <stdio.h>

#ifdef __cplusplus
/**
 * @name C++ version of variable types
 *
 * These definitions are used when including cLurupa.h in a C++ source.
 */
/*@{*/
typedef Bound_status BOUND_STATUS;
typedef Solver_status SOLVER_STATUS;
typedef bool BOOL;
class Lp;
typedef Lp *PLP;
/*@}*/
#else
/**
 * @name C version of variable types
 *
 * These definitions are used when including cLurupa.h in a C source.
 */
/*@{*/
typedef enum Bound_status BOUND_STATUS;
typedef enum Solver_status SOLVER_STATUS;
typedef char BOOL;
typedef void *PLP;
/*@}*/
#endif

/** @brief Initialize library wrapper */
void clu_init();
/** @brief Cleanup library wrapper */
void clu_cleanup();

/** @brief Get core version string */
const char *clu_get_core_version();

/** @brief Get algorithm parameter alpha */
double clu_get_alpha();
/** @brief Set algorithm parameter alpha */
void clu_set_alpha(double alpha);

/** @brief Get algorithm parameter eta */
double clu_get_eta();
/** @brief Set algorithm parameter eta */
void clu_set_eta(double eta);

/** @brief Is inflation set */
BOOL clu_is_inflate();
/** @brief Set inflation */
void clu_set_inflate(BOOL inflate);

/** @brief Set solver module */
BOOL clu_set_module(char *module_path);

/** @brief Get solver module version */
const char *clu_get_module_version();

/** @brief Print options supported by module */
void clu_print_module_options();

/** @brief Get solver's accuracy */
double clu_get_solver_eps();

/** @brief Compute lower bound */
BOUND_STATUS clu_lower_bound(PLP Lp, double *bound, int *iterations);

/** @brief Compute upper bound */
BOUND_STATUS clu_upper_bound(PLP Lp, double *bound, int *iterations);

/** @brief Get linear program's name */
char *clu_get_lp_name(PLP Lp);

/** @brief Return whether linear program is maximized */
BOOL clu_is_lp_maximize(PLP Lp);

/** @brief Read linear program */
PLP clu_read_lp(FILE *in, const double relative_interval_radius);

/** @brief Set a linear program's feasibility */
BOOL clu_set_solution(PLP Lp, void *p);

/** @brief Set linear program */
BOOL clu_set_lp(PLP *lp, const double relative_interval_radius);

/** @brief Set solver options */
BOOL clu_set_module_options(PLP Lp, int argc, char *argv[]);

/** @brief Solve linear program */
BOOL clu_solve_lp(double *optimal_value, SOLVER_STATUS *status);

#endif /* CLURUPA_H */
