/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file contains global properties shared by all components of %Lurupa. It
 * contains possible return values and parameters of methods as enums or defines
 * for enhanced readibility as well as string arrays with textual descriptions
 * of some of these for displaying.
 *
 * @file  globals.h
 * @brief Global properties
 *
 * @author Christian Keil
 *
 * $Id: globals.h 533 2008-06-12 20:08:12Z keil $
 */

/*
 * Documentation of the Hungarian prefixes used throughout:
 *
 * eq  - Equality constraint
 * le  - Less-equal constraint
 * con - Any constraint
 * st  - Status value (maybe Bound_status, Solver_status, etc.)
 */

#ifndef LURUPA_GLOBALS_H
#define LURUPA_GLOBALS_H

#ifdef HAVE_CONFIG_H
#include <lurupa/config.h>
#endif

/** @name %Bound status 
*/ 
/*@{*/
/** 
 * @brief %Bound status
 *
 * The possible status values of a bound computation. 
 */
enum Bound_status {
  bs_verified,    /**< %Bound successfully computed */
  bs_priminf,     /**< %Bound successfully computed; problem is verified to be infeasible */
  bs_dualinf,     /**< %Bound successfully computed; problem is verified to be unbounded */
  bs_orginf,      /**< %Bound not computed; the lp-solver claims infeasibility */
  bs_orgunb,      /**< %Bound not computed; the lp-solver claims unboundedness */
  bs_pertinf,     /**< %Bound not computed; a perturbed problem seems to be infeasible and inflation
                    is not set */
  bs_noenc,       /**< %Bound not computed; no enclosure of the solution set of linear system could
                    be found */
  bs_warn,        /**< %Bound not computed; the lp-solver gave a warning while solving a perturbed
                    problem */
  bs_rank,        /**< %Bound not computed; the equation matrix has not full rank */
  bs_iter,        /**< %Bound not computed; the maximum number of iterations is exceeded */
  bs_failure,     /**< %Bound not computed; an unexpected error occured */
  bs_timeout,     /**< %Bound not computed; the lp-solver timed out solving a perturbed problem */
  bs_running      /**< %Bound computation in progress */
};
/**
 * @brief %Bound status strings
 *
 * Status string array converting Bound_status values to real text for
 * displaying.
 */
static const char *bound_status_string[] = {
  "ok",
  "primal infeasible",
  "dual infeasible",
  "lp-solver infeasible",
  "lp-solver unbounded",
  "perturbed lp infeasible",
  "no ls enclosure",
  "solver warning",
  "rank deficiency",
  "iteration limit reached",
  "solver failure",
  "solver timeout",
  "running"
};
/*@}*/

/** @name Lp-solver status
*/
/*@{*/
/** 
 * @brief Lp-solver status
 * 
 * The possible status values returned by lp-solvers.
 */
enum Solver_status {
  ss_feasible,   /**< Lp-solver judges model to be feasible */
  ss_unbounded,  /**< Lp-solver judges model to be unbounded */
  ss_infeasible, /**< Lp-solver judges model to be infeasible */
  ss_failure,    /**< Lp-solver returned from solving with a failure */
  ss_timeout,    /**< Lp-solver timed out solving */
  ss_unknown     /**< Lp-solver status not yet known */
};
/**
 * @brief Lp-solver status strings
 *
 * Status string array converting Solver_status values to real text for
 * displaying.
 */
static const char *solver_status_string[] = {
  "feasible",
  "unbounded",
  "infeasible",
  "failure",
  "timeout",
  "unknown"
};
/*@}*/

/** 
 * @brief CSV style
 *
 * The style which is used to save csv files in Report#write_vector and
 * Report#write_matrix
 */
enum Csv_style {
  octave, /**< Octave style; one file (<em>problem_name</em>.csv) with dimension and
            name as meta information */
  matlab  /**< Matlab style; all variables in different files without dimension
            in one directory (<em>problem_name</em>) */
};

/** @name Tables
 *
 * The kind of tables that are written to disk by the command line client.
 */
/*@{*/
#define LU_T_LATEX 1 /**< @brief LaTeX table */
#define LU_T_CSV   2 /**< @brief CSV table */
/*@}*/

/**
 * @brief Write %lps as MPS file
 *
 * Which intermediate %lps are written as MPS files to disk.
 */
enum Write_mps {
  wm_all,   /**< All intermediate %lps. */
  wm_ask,   /**< Ask for all intermediate %lps. */
  wm_none   /**< No intermedate %lp. */
};

#endif /* LURUPA_GLOBALS_H */
