#include <mex.h>
#include <matrix.h>
#include <mat.h>
#include <lurupa/Lurupa.h>
#include <lurupa/Lp.h>
#include <IntervalMatrix.h>
#include <string.h>

#ifndef MX_API_VERSION
typedef int mwSize;
typedef int mwIndex;
#endif

static Lurupa lurupa;

bool
is_valid_ls(const mxArray *ls) {
	if (!mxIsStruct(ls)) {
		mexPrintf("Expecting structure for Lp.\n");
		return false;
	}
	if (7 != mxGetNumberOfFields(ls)) {
		mexPrintf("Expecting 7 fields in Lp structure.\n");
		return false;
	}
	if (1 != mxGetM(ls) * mxGetN(ls)) {
		mexPrintf("Expecting ONE Lp structure.\n");
		return false;
	}

	int ceq, cle, cv;
	mxArray *mx;
	mx  = mxGetFieldByNumber(ls, 0, 0);
	cv  = mxGetM(mx);
	mx  = mxGetFieldByNumber(ls, 0, 1);
	cle = mxGetM(mx);
	mx  = mxGetFieldByNumber(ls, 0, 3);
	ceq = mxGetM(mx);

	if (cle) {
		mx = mxGetFieldByNumber(ls, 0, 1);
		if (cv != mxGetN(mx)) {
			mexPrintf("Wrong number of columns in A (expected %d, found %d).\n", cv, mxGetN(mx));
			return false;
		}

		mx = mxGetFieldByNumber(ls, 0, 2);
		if ((cle != mxGetM(mx)) || (1 != mxGetN(mx))) {
			mexPrintf("Wrong dimension for a (expected %dx1, found %dx%d).\n", cle, mxGetM(mx), mxGetN(mx));
			return false;
		}
	}

	if(ceq) {
		mx = mxGetFieldByNumber(ls, 0, 3);
		if (cv != mxGetN(mx)) {
			mexPrintf("Wrong number of columns in B (expected %d, found %d).\n", cv, mxGetN(mx));
			return false;
		}

		mx = mxGetFieldByNumber(ls, 0, 4);
		if ((ceq != mxGetM(mx)) || (1 != mxGetN(mx))) {
			mexPrintf("Wrong dimension for b (expected %dx1, found %dx%d).\n", ceq, mxGetM(mx), mxGetN(mx));
			return false;
		}
	}

	mx = mxGetFieldByNumber(ls, 0, 5);
	if ((cv != mxGetM(mx)) || (1 != mxGetN(mx))) {
		mexPrintf("Wrong dimension for xl (expected %dx1, found %dx%d).\n", cv, mxGetM(mx), mxGetN(mx));
		return false;
	}

	mx = mxGetFieldByNumber(ls, 0, 6);
	if ((cv != mxGetM(mx)) || (1 != mxGetN(mx))) {
		mexPrintf("Wrong dimension for xu (expected %dx1, found %dx%d).\n", cv, mxGetM(mx), mxGetN(mx));
		return false;
	}

	return true;
}

MATRIX
matrix_from_mx(const mxArray *mx, int m, int n)
{
	MATRIX mT(m, n);
	Clear(mT);
	MakeTemporary(mT);

	if (mxIsSparse(mx)) {
		mwIndex *ir, *jc;
		double *pr;
		ir = mxGetIr(mx);
		jc = mxGetJc(mx);
		pr = mxGetPr(mx);
		for (int j = 0; j < n; ++j) {
			for (int i = 0; i < jc[1] - jc[0]; ++i, ++ir, ++pr) {
				mT((*ir) + 1, j + 1) = *pr;
			}
			++jc;
		}
	} else {
		double *pr = mxGetPr(mx);
		for (int i = 0; i < m; ++i) {
			for (int j = 0; j < n; ++j) {
				mT(i+1, j+1) = pr[i + j*m];
			}
		}
	}

	return mT;
}

Lp *
lp_from_ls(const mxArray *mx)
{
	Lp *lp = new Lp(&lurupa);

	int ceq, cle, cv;
	mxArray *mxT;
	mxT = mxGetFieldByNumber(mx, 0, 0);
	cv  = mxGetM(mxT);
	mxT = mxGetFieldByNumber(mx, 0, 1);
	cle = mxGetM(mxT);
	mxT = mxGetFieldByNumber(mx, 0, 3);
	ceq = mxGetM(mxT);

	Resize(lp->IA, cle, cv);
	Resize(lp->IB, ceq, cv);
	Resize(lp->ic, cv);
	Resize(lp->ia, cle);
	Resize(lp->ib, ceq);
	Resize(lp->xl, cv);
	Resize(lp->xu, cv);
	Resize(lp->ix, cv);
	Resize(lp->iy, cle);
	Resize(lp->iz, ceq);

	mxT = mxGetFieldByNumber(mx, 0, 0);
	VECTOR v(cv);
	memcpy(v.theElements, mxGetPr(mxT), cv * sizeof (REAL));
	lp->ic = Hull(v);

	if (cle) {
		mxT = mxGetFieldByNumber(mx, 0, 1);
		lp->IA = Hull(matrix_from_mx(mxT, cle, cv));

		mxT = mxGetFieldByNumber(mx, 0, 2);
		Resize(v, cle);
		memcpy(v.theElements, mxGetPr(mxT), cle * sizeof (REAL));
		lp->ia = Hull(v);
	}

	if (ceq) {
		mxT = mxGetFieldByNumber(mx, 0, 3);
		lp->IB = Hull(matrix_from_mx(mxT, ceq, cv));

		mxT = mxGetFieldByNumber(mx, 0, 4);
		Resize(v, ceq);
		memcpy(v.theElements, mxGetPr(mxT), ceq * sizeof (REAL));
		lp->ib = Hull(v);
	}

	mxT = mxGetFieldByNumber(mx, 0, 5);
	memcpy(lp->xl.theElements, mxGetPr(mxT), cv * sizeof (REAL));

	mxT = mxGetFieldByNumber(mx, 0, 6);
	memcpy(lp->xu.theElements, mxGetPr(mxT), cv * sizeof (REAL));

	lp->name = new char[10 + 1];
	sprintf(lp->name, "FromMatlab");
	/* Using lp_solve 5.5's values */
	lp->infinite = 1e30;

	lp->non_fixed_vars = 0;
	lp->free_variables_size = 0;
	lp->free_variables = NULL;
	for (int i = 1; i <= cv; ++i) {
		if (lp->xl(i) != lp->xu(i)) ++lp->non_fixed_vars;
		if ((lp->xl(i) == BiasNegInf) && (lp->xu(i) == BiasPosInf)) ++lp->free_variables_size;
	}
	if (lp->free_variables_size > 0) {
		lp->free_variables = new int[lp->free_variables_size];
		for (int i = 1, *j = lp->free_variables; i <= cv; ++i) {
			if ((lp->xl(i) == BiasNegInf) && (lp->xu(i) == BiasPosInf)) *(j++) = i;
		}
	}
	lp->maximize = false;

	if (!lp->is_consistent()) {
		mexErrMsgTxt("Lp is not consistent!\n");
	}

	return lp;
}

void 
lower_bound(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs > 2) {
		mexErrMsgTxt("Need exactly two input arguments for 'lower_bound'.\n");
	}
	if (!is_valid_ls(prhs[1])) {
		mexErrMsgTxt("Lp structure seems to be invalid.\n");
	}
	if (((nlhs >= 2 ) && (nlhs <= 5)) || (nlhs == 7) || (nlhs >= 9)) {
		mexErrMsgTxt("Invalid number of output arguments for 'lower_bound'.\n");
	}

	Lp *lp;
	lp = lp_from_ls(prhs[1]);
	lurupa.lp2solver(lp);

	double f, fl;
	int it;
	lurupa.solve_lp(lp, f);

	if (nlhs == 8) {
		plhs[6] = mxCreateDoubleScalar(f);
		int cv = Dimension(lp->ix);
		plhs[7] = mxCreateDoubleMatrix(cv, 1, mxREAL);
		/* lp->ix is currently the point approximate solution */
		memcpy(mxGetPr(plhs[7]), Inf(lp->ix).theElements, cv * sizeof (REAL));
	}

	int32_t flag = lurupa.lower_bound(lp, fl, it);

	plhs[0] = mxCreateDoubleScalar(fl);
	if (nlhs >= 6) {
		plhs[1] = mxCreateNumericMatrix(1, 1, mxINT32_CLASS, mxREAL);
		int32_t *p = static_cast<int32_t *>(mxGetData(plhs[1]));
		p[0] = flag;
		int cle, ceq;
		cle = Dimension(lp->iy);
		ceq = Dimension(lp->iz);
		plhs[2] = mxCreateDoubleMatrix(cle, 1, mxREAL);
		plhs[3] = mxCreateDoubleMatrix(cle, 1, mxREAL);
		memcpy(mxGetPr(plhs[2]), Inf(lp->iy).theElements, cle * sizeof (REAL));
		memcpy(mxGetPr(plhs[3]), Sup(lp->iy).theElements, cle * sizeof (REAL));

		plhs[4] = mxCreateDoubleMatrix(ceq, 1, mxREAL);
		plhs[5] = mxCreateDoubleMatrix(ceq, 1, mxREAL);
		memcpy(mxGetPr(plhs[4]), Inf(lp->iz).theElements, ceq * sizeof (REAL));
		memcpy(mxGetPr(plhs[5]), Sup(lp->iz).theElements, ceq * sizeof (REAL));
	}
	delete lp;
}

void 
upper_bound(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs > 2) {
		mexErrMsgTxt("Need exactly two input arguments for 'lower_bound'.\n");
	}
	if (!is_valid_ls(prhs[1])) {
		mexErrMsgTxt("Lp structure seems to be invalid.\n");
	}
	if (((nlhs >= 2 ) && (nlhs <= 3)) || (nlhs == 5) || (nlhs >= 7)) {
		mexErrMsgTxt("Invalid number of output arguments for 'lower_bound'.\n");
	}

#if 0
	Lurupa lurupa;
	lurupa.report.set_mex(true);
	if (!lurupa.set_module("Sm_lps5_5")) {
		mexErrMsgTxt("Failed to set solver module Sm_lps5_5.\n");
	}
#endif
	Lp *lp;
	lp = lp_from_ls(prhs[1]);
	lurupa.lp2solver(lp);

	double f, fu;
	int it;
	lurupa.solve_lp(lp, f);

	if (nlhs == 6) {
		plhs[4] = mxCreateDoubleScalar(f);
		int cv = Dimension(lp->ix);
		plhs[5] = mxCreateDoubleMatrix(cv, 1, mxREAL);
		/* lp->ix is currently the point approximate solution */
		memcpy(mxGetPr(plhs[5]), Inf(lp->ix).theElements, cv * sizeof (REAL));
	}

	int32_t flag = lurupa.upper_bound(lp, fu, it);

	plhs[0] = mxCreateDoubleScalar(fu);
	if (nlhs >= 4) {
		plhs[1] = mxCreateNumericMatrix(1, 1, mxINT32_CLASS, mxREAL);
		int32_t *p = static_cast<int32_t *>(mxGetData(plhs[1]));
		p[0] = flag;
		int cv = Dimension(lp->ix);
		plhs[2] = mxCreateDoubleMatrix(cv, 1, mxREAL);
		plhs[3] = mxCreateDoubleMatrix(cv, 1, mxREAL);
		memcpy(mxGetPr(plhs[2]), Inf(lp->ix).theElements, cv * sizeof (REAL));
		memcpy(mxGetPr(plhs[3]), Sup(lp->ix).theElements, cv * sizeof (REAL));
	}
	delete lp;
}

void
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	static int init = 0;

	if (0 == init) {
		lurupa.report.set_mex(true, mexPrintf, mexWarnMsgTxt);
		if (!lurupa.set_module("Sm_lps5_5")) {
			mexErrMsgTxt("Failed to set solver module Sm_lps5_5.\n");
		}
		init = 1;
	}

	/* First we'll do some input validation */
	if (nrhs == 0) {
		mexErrMsgTxt("At least one input argument required. See help lurupa.\n");
	}

	if (!mxIsChar(prhs[0])) {
		mexErrMsgTxt("Function to perform has to be a string.\n");
	}

	mwSize cch = mxGetM(prhs[0]) * mxGetN(prhs[0]) * sizeof (mxChar) + 1;
	char *sz = (char *) mxMalloc(cch);
	mxGetString(prhs[0], sz, cch);

	if (0 == strcmp(sz, "lower_bound")) {
		lower_bound(nlhs, plhs, nrhs, prhs);
	} else if (0 == strcmp(sz, "upper_bound")) {
		upper_bound(nlhs, plhs, nrhs, prhs);
	} else {
		mexErrMsgTxt("Not implemented yet.\n");
	}

	mxFree(sz);
}

