/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file defines the routines for reporting and debugging.
 *
 * @file  Report.cpp
 * @brief Definition of %Report class
 *
 * @author Christian Keil
 *
 * $Id: Report.cpp 530 2008-06-12 08:57:52Z keil $
 */
#include <lurupa/Report.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

/**
 * Set the default values: 
 * - Only warning and error messages are printed.
 * - Messages are printed without prepended time.
 * - Vectors and matrices are not written to disk.
 * - #Csv_style is set to #matlab.
 * - Warning and error messages go to stderr all others to stdout.
 */
Report::Report()
{
	print_time = false;
	write_vm = false;
	level = 2;
	csv_style = matlab;
	normal = stdout;
	warning = stderr;
	error = stderr;

#ifdef WITH_MATLAB
	mex = false;
	mexPrintf = 0;
	mexWarnMsgTxt = 0;
#endif
} /* Report::Report */

/**
 * Set the verbosity. The @p level determines which messages are displayed. If a
 * message's severity is below this threshold, it is displayed. If @p print_time
 * is set, each message will be prepended with the current time.  Intermediate
 * vectors and matrices will be written to disk if @p write_vm is set.
 *
 * @param[in] level       messages verbosity
 * @param[in] print_time  prepend messages with current time
 * @param[in] write_vm    write vectors and matrices to disk
 */
void
Report::set_verbosity(short level, bool print_time, bool write_vm)
{
	this->level  = level; 
	this->print_time = print_time;
	this->write_vm   = write_vm;
} /* Report::set_verbosity */

void
Report::set_level(short level)
{
	this->level = level;
} /* Report::set_level */

short
Report::get_level() const
{ 
	return level; 
} /* Report::get_level */

void
Report::set_print_time(bool print_time)
{
	this->print_time = print_time;
} /* Report::set_print_time */

bool
Report::get_print_time() const
{
	return print_time;
} /* Report::get_print_time */

void
Report::set_write_vm(bool write_vm)
{
	this->write_vm = write_vm;
} /* Report::set_write_vm */

bool
Report::get_write_vm() const
{
	return write_vm;
} /* Report::get_write_vm */

/**
 * Set which csv style is to be used when writing vectors and matrices to disk.
 *
 * @param[in] style  the csv style to use
 *
 * @usage
 * @code
 *   Report report;
 *   report.set_csv_style(octave);
 * @endcode
 */
void
Report::set_csv_style(Csv_style style) 
{
	csv_style = style;
} /* Report::set_csv_style */

Csv_style 
Report::get_csv_style() const
{
	return csv_style;
} /* Report::get_csv_style */

void
Report::set_normal(FILE *normal)
{
	this->normal = normal;
} /* Report::set_normal */

void
Report::set_warning(FILE *warning)
{
	this->warning = warning;
} /* Report::set_warning */

void
Report::set_error(FILE *error)
{
	this->error = error;
} /* Report::set_error */

#ifdef WITH_MATLAB
void
Report::set_mex(bool mex, int (*mexPrintf)(const char *,...), void (*mexWarnMsgTxt)(const char *))
{
	this->mex = mex;
	this->mexPrintf = mexPrintf;
	this->mexWarnMsgTxt = mexWarnMsgTxt;
} /* Report::set_mex */

bool
Report::is_mex()
{
	return mex;
} /* Report::is_mex */
#endif

/**
 * Print the supplied message if its @p severity is below #level. The message is
 * assumed to be in standard printf format. Messages with a @p severity of @c 1
 * are prepended with '<tt>ERROR:</tt>', messages with a @p severity of @c 2
 * with '<tt>Warning:</tt>'. Higher @p severity values cause an indentation of
 * two spaces for each level higher than @c 3. Messages with a @p severity value
 * of 0 are printed without a prefix and without indentation.
 *
 * @param[in] severity  severity of message
 * @param[in] format    format string
 * @param[in] ...       any necessary parameters belonging to the format string
 *
 * @usage
 * @code
 *   Report report;
 *   report.print(2, "test");
 *   report.print(1, "test %d \t %f \n", 2, 3.141);
 * @endcode
 */
void
Report::print(short severity, const char * format, ...) const
{
	if (severity < 0) severity = 1;
	if (severity <= level) {
#ifdef WITH_MATLAB
		static char sz[200 + 1];
		int cch = 0;
#endif
		FILE *out = normal;
		if (severity == 1) out = error;
		else if (severity == 2) out = warning;

		va_list l;
		va_start(l, format);

		if (print_time) {
			time_t t = time(NULL);
			char *ct = ctime(&t);
			/* kill this bl.... \n at the end of the string */
			ct[strlen(ct) - 1] = '\0';
#ifdef WITH_MATLAB
			if (mex) {
				cch += snprintf(sz, 201, "[%s] ", ct);
			} else
#endif
			fprintf(out, "[%s] ", ct);
		}

		switch (severity) {
		case 0:
			/* message comes from the solver */
			break;
		case 1:
			/* ERROR prefix */
#ifdef WITH_MATLAB
			if (mex && (cch >= 0) && (cch < 200)) {
				cch += snprintf(sz + cch, 201 - cch, "ERROR: ");
			} else
#endif
			fprintf(out, "ERROR: ");
			break;
		case 2:
			/* Warning prefix */
#ifdef WITH_MATLAB
			if (mex && (cch >= 0) && (cch < 200)) {
				cch += snprintf(sz + cch, 201 - cch, "Warning: ");
			} else
#endif
			fprintf(out, "Warning: ");
			break;
		default:
			/* indent */
#ifdef WITH_MATLAB
			if (mex && (cch >= 0) && (cch < 200)) {
				cch += snprintf(sz + cch, 201 - cch, "%*s", (severity - 3) * 2, "");
			} else
#endif
			fprintf(out, "%*s", (severity - 3) * 2, "");
			break;
		}
#ifdef WITH_MATLAB
		if (mex && (cch >= 0) && (cch < 200)) {
			cch += vsnprintf(sz + cch, 201 - cch, format, l);
		} else
#endif
		vfprintf(out, format, l);

		va_end(l);
#ifdef WITH_MATLAB
		if (mex) {
			if ((cch < 0) || (cch > 200)) {
				sprintf(sz + 197, "...");
			}
			if ((1 == severity) || (2 == severity)) {
				mexWarnMsgTxt(sz);
			} else {
				mexPrintf("%s\n", sz);
			}
		} else {
#endif
		fputc('\n', out);

		/* flush the output stream, otherwise we have error strings displayed out of
		 * order when redirecting output */
		if (fflush(out) != 0) {
			fprintf(error, "ERROR: Report::print: %s\n", strerror(errno));
		}
#ifdef WITH_MATLAB
		}
#endif
	}
} /* Report::print */

/**
 * Write a vector to a file. What happens depends on #csv_style.
 *
 * If this is set to #octave, append the data to @p file with extension
 * '<tt>.csv</tt>' First write a block of meta data containing the name and
 * dimension of the vector in the form
 * @code
 *   # name: <name of vector>
 *   # type: global matrix
 *   # rows: <dimension of vector>
 *   # columns: 1
 * @endcode
 * Then write the data as csv (comma seperated value) using '<tt>\\n</tt>' to
 * seperate the components.
 *
 * If #csv_style is set to #matlab create a directory @p file if necessary and
 * write the data without meta information to a file @p name with extension
 * '<tt>.csv</tt>'. The format of the data is the same as with #octave.
 *
 * @param[in] file  in octave style basename of the file to append to,
 *                  in matlab style name of the directory to store data in
 * @param[in] x     the vector to be written
 * @param[in] name  the name of the vector, in matlab style basename of the file
 *                  to write to
 *
 * @usage
 * @code
 *   Report report;
 *   report.write_vector("output", x, "costs");
 * @endcode
 */
void
Report::write_vector(const char *file, const VECTOR &x, const char *name) const
{
	switch (csv_style) {
	case octave: 
		{
			char* file_csv = new char[strlen(file) + 4 + 1];
			strcpy(file_csv, file);
			sprintf(file_csv + strlen(file), ".csv");
			FILE *out = NULL;
			if ((out = fopen(file_csv, "a")) == NULL) {
				print(1, "Report::write_vector: opening %s for appending", file_csv);
			} else {
				fprintf(out, "# name: %s\n", name);
				fprintf(out, "# type: global matrix\n");
				fprintf(out, "# rows: %d\n", Dimension(x));
				fprintf(out, "# columns: 1\n");
				for (int comp = 1; comp <= Dimension(x); ++comp)
					fprintf(out, " %.16e\n", x(comp));
				fclose(out);
			}
		}
		break;
	case matlab:
		{
			/* check if we need to create a directory */
			struct stat buf;
			int status = stat(file, &buf);
			if (status == -1) {
				/* error stating entity */
				if (errno == ENOENT) {
					/* no entity exists - make directory */
#if MKDIR_ACCEPTS_MODE
					if (mkdir(file, 0744) != 0)
#else
					if (mkdir(file) != 0)
#endif
					{
						print(1, "Report::write_vector: creating directory %s", file);
						print(1, strerror(errno));
					}
				} else {
					/* stating failed with other reason - failure */
					print(1, "Report::write_vector: stating %s", file);
					print(1, strerror(errno));
				}
			} else if ((status == 0) && !(S_ISDIR(buf.st_mode))) {
				/* entity exists and is no directory - failure */
				print(1, "Report::write_vector: cannot create directory %s, " 
							"non-directory entity with that name exists", file);
			} else {
				/* at this point a directory exists - either created by us or already
				 * existent */
				char * file_csv = new char[strlen(file) + 1 + strlen(name) + 4 + 1];
				sprintf(file_csv, "%s/%s.csv", file, name);
				FILE * out = NULL;
				if ((out = fopen(file_csv, "w")) == NULL) {
					print(1, "Report::write_vector: opening %s for writing", file_csv);
				} else {
					for (int comp = 1; comp <= Dimension(x); ++comp)
						fprintf(out, "%.16e\n", x(comp));
					fclose(out);
				}
			}
		}
	}
} /* Report::write_vector */

/**
 * Write a matrix to a file. What happens depends on #csv_style.
 *
 * If this is set to #octave, append the matrix to @p file with extension
 * '<tt>.csv</tt>'. First write a block of meta data containing the name and
 * dimension of the matrix in the form
 * @code
 *   # name: <name of matrix>
 *   # type: global matrix
 *   # rows: <row dimension of matrix>
 *   # columns: <column dimension of matrix>
 * @endcode
 * Then write the data as csv (comma seperated value) using spaces to seperate
 * row entries and '<tt>\\n</tt>' to seperate the columns.
 *
 * If #csv_style is set to #matlab create a directory @p file if necessary and
 * write the data without meta information to a file @p name with extension
 * '<tt>.csv</tt>'. The format of the data is the same as with #octave.
 *
 * @param[in] file  in octave style basename of the file to append to, 
 *                  in matlab style name of the directory to store data in
 * @param[in] X     the matrix to be written
 * @param[in] name  the name of the matrix, in matlab style basename of the file
 *                  to write to
 *
 * @usage
 * @code
 *   Report report;
 *   report.write_matrix("output", X, "Equations");
 * @endcode
 */
void
Report::write_matrix(const char *file, const MATRIX &X, const char *name) const
{
	switch (csv_style) {
	case octave:
		{
			char *file_csv = new char[strlen(file) + 4 + 1];
			strcpy(file_csv, file);
			sprintf(file_csv + strlen(file), ".csv");
			FILE *out = NULL;
			if ((out = fopen(file_csv, "a")) == NULL) {
				print(1, "Report::write_matrix: opening %s for appending", file_csv);
			} else {
				fprintf(out, "# name: %s\n", name);
				fprintf(out, "# type: global matrix\n");
				fprintf(out, "# rows: %d\n", RowDimension(X));
				fprintf(out, "# columns: %d\n", ColDimension(X));
				for (int i = 1; i <= RowDimension(X); ++i) {
					for (int j = 1; j <= ColDimension(X); ++j)
						fprintf(out, " %.16e", X(i, j));
					fprintf(out, "\n");
				}
				fclose(out);
			}
		}
		break;
	case matlab:
		{
			/* check if we need to create a directory */
			struct stat buf;
			int status = stat(file, &buf);
			if (status == -1) {
				/* error stating entity */
				if (errno == ENOENT) {
					/* no entity exists - make directory */
#if MKDIR_ACCEPTS_MODE
					if (mkdir(file, 0744) != 0)
#else
					if (mkdir(file) != 0)
#endif
					{
						print(1, "Report::write_matrix: creating directory %s", file);
						print(1, strerror(errno));
					}
				} else {
					/* stating failed with other reason - failure */
					print(1, "Report::write_matrix: stating %s", file);
					print(1, strerror(errno));
				}
			} else if ((status == 0) && !(S_ISDIR(buf.st_mode))) {
				/* entity exists and is no directory - failure */
				print(1, "Report::write_matrix: cannot create directory %s," 
							"existent non-directory entity with that name", file);
			} else {
				/* at this point a directory exists - either created by us or already
				 * existent */
				char *file_csv = new char[strlen(file) + 1 + strlen(name) + 4 + 1];
				sprintf(file_csv, "%s/%s.csv", file, name);
				FILE *out = NULL;
				if ((out = fopen(file_csv, "w")) == NULL) {
					print(1, "Report::write_matrix: opening %s for writing", file_csv);
				} else {
					for (int i = 1; i <= RowDimension(X); ++i) {
						for (int j = 1; j < ColDimension(X); ++j)
							fprintf(out, "%.16e,", X(i, j));
						fprintf(out, "%.16e\n", X(i, ColDimension(X)));
					}
					fclose(out);
				}
			}
		}
	}
} /* Report::write_matrix */
