/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file defines the solver module for lp_solve 3.2, Sm_lps3_2.
 *
 * @file  Sm_lps3_2.cpp
 * @brief Definition of lp_solve 3.2 solver module
 *
 * @author Christian Keil
 *
 * $Id: Sm_lps3_2.cpp 533 2008-06-12 20:08:12Z keil $
 */
#include "Sm_lps3_2.h"
#include <lurupa/Lp.h>
#include <lpglob.h>
#include <patchlevel.h>
#include <string.h>

/** @brief Version string */
static const char *version = "Lp_solve 3.2 module 1.0";

Report *Solver_module::preport = NULL;

Lp_lp_solve::Lp_lp_solve()
{
  mp_eq_con = NULL;
  mp_le_con = NULL;
}

Lp_lp_solve::~Lp_lp_solve()
{
  delete_lp(lp);
  lp = NULL;
  delete [] mp_eq_con;
  delete [] mp_le_con;
  delete [] mp_parts;
}

/**
 * Get the version information of the solver module.
 */
const char *
Sm_lps3_2::get_version()
{
  return version;
}

void
Sm_lps3_2::print_version()
{
  printf("solver module\n"
         "  version : %s\n"
         "  solver parameters\n"
         "    version : %s\n"
         "    eps     : %g\n", 
         version, PATCHLEVEL, DEF_EPSD);
}

void
Sm_lps3_2::print_brief_version()
{
  printf("solver module       : %s, solver version = %s, solver eps = %g\n",
         version, PATCHLEVEL, DEF_EPSD);
}

/**
 * Set #preport to \p *report.
 *
 * \param[in] report  the #Report instance to use
 *
 * \retval true
 */
bool
Sm_lps3_2::init(Report *report)
{
  preport = report;
  preport->print(3, "Using %s", version);

  if (0 != strcmp(PATCHLEVEL, "3.2")) {
    preport->print(2, "Lp_solve version differs, expected: 3.2, found: %s", PATCHLEVEL);
  }

  return true;
}

/**
 * Print the command line options supported by the solver module.
 */
void
Sm_lps3_2::print_options()
{
  printf("No options supported by %s\n", version);
}

/**
 * Read general lp properties from lp_solve's structures and store them in @p lp. This includes the
 * value representing infinity, the direction of optimization, and the name of the lp.
 *
 * @param[out] lp  stores the read data
 */
void
Sm_lps3_2::read_general_data(Lp *lp)
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;
  lp->infinite = lp_lps->infinite;
  lp->maximize = (bool) lp_lps->maximise;
  lp->name = strdup(lp_lps->lp_name);
}

/**
 * Build a mapping from Lurupa's equality and inequality constraints to lp_solve's constraints,
 * which is storing all constraints in one big matrix. Store the mapping in @p lp->module->mp_eq_con
 * and @p lp->module->mp_le_con. Check if the number of constraints is less than the number of
 * variables.
 *
 * @param[out] c_eq  number of equations in lp
 * @param[out] c_le  number of inequalities in lp
 * @param[in]  rows  number of rows of constraint matrix
 * @param[in]  cols  number of columns of constraint matrix
 * @param[in]  lp    the %lp
 *
 * @retval true
 * if number of constraints is less than number of variables,
 * @retval false
 * otherwise
 */
bool
Sm_lps3_2::build_constraint_maps(int &c_eq, int &c_le, const int rows, const int cols, Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* count equations and inequalities */
  for (int i = 1; i <= rows; ++i)
    if (lp_lps->lp->orig_upbo[i] == 0)  {
      ++c_eq;
    } else {
      ++c_le;
    }

  if (c_eq > cols) {
    preport->print(1, "Sm_lps3_2::build_constraint_maps: model contains more "
                   "equations (%d) than variables (%d)", c_eq, cols);
    return false;
  }

  /* build maps */
  lp_lps->mp_eq_con = new int[c_eq];
  lp_lps->mp_le_con = new int[c_le];
  c_eq = -1; c_le = -1;
  for (int i = 1; i <= rows; ++i) {
    if (lp_lps->lp->orig_upbo[i] == 0) {
      lp_lps->mp_eq_con[++c_eq] = i;
    } else {
      lp_lps->mp_le_con[++c_le] = i;
    }
  }

  ++c_le;
  ++c_eq;

  return true;
}

/**
 * Find free variables in the lp. Store the indices of the free variables in @p lp->free_variables
 * and their number in @p lp->free_variables_size.
 *
 * Adjust the number of variables @p vars to reflect the actual count of variables and not include
 * the split variables used in lp_solve to represent free variables.
 *
 * @param[in,out] vars    number of variables in lp
 * @param[out]    lp      stores number and indices of the free variables
 */
void
Sm_lps3_2::find_free_variables(int &vars, Lp *lp)
{
  /* Lp_solve 3.2 does not support free variables directly, but splits them into two variables that
   * are just connected through their names. The split variable has the same name with '_' appended.
   * Search through the variable names for trailing '_' and match with the corresponding original
   * variable. This is heuristic, but the best we can do. */

  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);
  int cch;
  nstring s_pos_part;
  lp_lps->mp_parts = new int[vars];
  memset(lp_lps->mp_parts, 0, vars * sizeof (int));

  /* count free variables */
  for (int i = vars; i > 0; --i) {
    cch = strlen(lp_lps->lp->col_name[i]);
    if (lp_lps->lp->col_name[i][cch - 1] == '_') {
      /* we likely found a free variable */
      strncpy(s_pos_part, lp_lps->lp->col_name[i], cch - 1);
      s_pos_part[cch - 1] = '\0';

      hashelem *p_pos_part = findhash(s_pos_part, lp_lps->lp->colname_hashtab);
      if (p_pos_part != NULL) {
        lp_lps->mp_parts[i - 1] = -p_pos_part->index;
        lp_lps->mp_parts[p_pos_part->index - 1] = i;
        ++lp->free_variables_size;
      } else
        /* as the free variables are appended after the bounded ones there can't be any free left */
        break;
    }
  }
  lp->free_variables = new int[lp->free_variables_size];

  /* store indices */
  int var = vars, free_var = lp->free_variables_size - 1;
  while (free_var >= 0) {
    while (lp_lps->mp_parts[--var] >= 0) {};
    lp->free_variables[free_var--] = -lp_lps->mp_parts[var];
  };

  /* Only the 'real' variables are important to us, not the helper ones. */
  vars = vars - lp->free_variables_size;
}

/**
 * Resize the vectors and matrices in Lurupa's data structure to the sizes of
 * the lp.
 *
 * @param[out] lp     Lurupa's structure with the vectors and matrices to be resized
 * @param[in]  cols   number of variables in the lp
 * @param[in]  c_eq   number of equality constraints in the lp
 * @param[in]  c_le   number of inequality constraints in the lp
 */
void
Sm_lps3_2::resize_lp(Lp *lp, const int cols, const int c_eq, const int c_le)
{
  Resize(lp->IA, c_le, cols);
  Resize(lp->IB, c_eq, cols);
  Resize(lp->ic, cols);
  Resize(lp->ia, c_le);
  Resize(lp->ib, c_eq);
  Resize(lp->xl, cols);
  Resize(lp->xu, cols);
  Resize(lp->ix, cols);
  Resize(lp->iy, c_le);
  Resize(lp->iz, c_eq);
}

/**
 * Read lp_solve's @c lp_mat structure. This stores the left hand sides of the
 * constraints together with the objective function. Split the read coefficients
 * into equality and inequality constraints, transform inequality constraints to
 * all less equal, and the direction of optimization to minimize.
 *
 * @param[in]  rows   number of rows in constraint matrix
 * @param[out] lp     stores the constraints and objective function
 */
void
Sm_lps3_2::read_lp_mat(const int rows, Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  REAL *column = new REAL[rows + 1];
  int d_j = 0;
  for (int j = 1; j <= ColDimension(lp->IA); ++j) {
    /* get_column returns the parameters of the _original_ lp, so we need to
     * transform into our representation */
    while (lp_lps->mp_parts[j + d_j - 1] < 0) ++d_j;

    get_column(lp_lps->lp, j + d_j, column);

    /* maximizing linear programs become minimizing ones: multiply objective
     * function coefficients by -1 */
    lp->ic(j) = lp->maximize ? -Hull(column[0]) : Hull(column[0]);

    for (int i = 1; i <= RowDimension(lp->IA); ++i) {
      /* >= constraints become <= constraints: multiply coefficients by -1 */
      if (!lp_lps->lp->ch_sign[lp_lps->mp_le_con[i - 1]]) {
        lp->IA(i, j) =  Hull(column[lp_lps->mp_le_con[i - 1]]);
      } else {
        lp->IA(i, j) = -Hull(column[lp_lps->mp_le_con[i - 1]]);
      }
    }

    for (int i = 1; i <= RowDimension(lp->IB); ++i) {
      lp->IB(i, j) = Hull(column[lp_lps->mp_eq_con[i - 1]]);
    }
  }
  delete [] column;
} /* Sm_lps3_2::read_lp_mat */

/**
 * Read the right hand sides of the constraints. Split the read right hand sides
 * into equality and inequality constraints, transform inequality constraints to
 * all less equal. Store the values in @p lp->ia and @p lp->ib.
 *
 * @param[out] lp  stores the right hand sides
 */
void
Sm_lps3_2::read_right_hand_sides(Lp *lp)
{
  /* lp_solve 3.2 does not offer a function which returns the right hand sides 
   * => we have to get these from the internal structures and check if scaling is in use */
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  if (lp_lps->lp->scaling_used) {
    preport->print(6, "Scaling in use");

    for (int i = 1; i <= Dimension(lp->ia); ++i) {
      /* internally lp_solve only uses <= constraints, we do _not_ have to transform these (s.a.) */
      lp->ia(i) = (Hull(lp_lps->lp->orig_rh[lp_lps->mp_le_con[i - 1]])
                   / lp_lps->lp->scale[lp_lps->mp_le_con[i - 1]]);
    }

    for (int i = 1; i <= Dimension(lp->ib); ++i) {
      lp->ib(i) = (Hull(lp_lps->lp->orig_rh[lp_lps->mp_eq_con[i - 1]])
                   / lp_lps->lp->scale[lp_lps->mp_eq_con[i - 1]]);
    }
  } else { 
    preport->print(6, "No scaling in use");

    for (int i = 1; i <= Dimension(lp->ia); ++i) {
      /* internally lp_solve only uses <= constraints, we do _not_ have to transform these (s.a.) */
      lp->ia(i) = Hull(lp_lps->lp->orig_rh[lp_lps->mp_le_con[i - 1]]);
    }

    for (int i = 1; i <= Dimension(lp->ib); ++i) {
      lp->ib(i) = Hull(lp_lps->lp->orig_rh[lp_lps->mp_eq_con[i - 1]]);
    }
  }
} /* Sm_lps3_2::read_right_hand_sides */

/**
 * Read the simple bounds of the variables and store them in @p lp->xl and
 * @p lp->xu.
 *
 * @param[out] lp     stores the simple bounds
 * @param[in]  rows   number of rows in lp_solve's constraint matrix
 */
void
Sm_lps3_2::read_simple_bounds(Lp *lp, const int rows)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);
  REAL *up_bo   = &lp_lps->lp->orig_upbo[rows + 1];
  REAL *low_bo  = &lp_lps->lp->orig_lowbo[rows + 1];
  int d_i       = 0;

  if (lp_lps->lp->scaling_used) {
    REAL scale;

    for (int i = 1; i <= Dimension(lp->xl); ++i, ++up_bo, ++low_bo) {
      while (lp_lps->mp_parts[i + d_i - 1] < 0) {
        ++d_i; 
        ++up_bo; ++low_bo;
      }

      if (lp_lps->mp_parts[i + d_i - 1] > 0) {
        lp->xl(i) = -lp->infinite;
        lp->xu(i) =  lp->infinite;
        ++lp->non_fixed_vars;
      } else {
        scale = lp_lps->lp->scale[rows + i + d_i];
        lp->xl(i) = (-lp->infinite == *low_bo ? *low_bo : *low_bo * scale);
        lp->xu(i) = ( lp->infinite ==  *up_bo ?  *up_bo :  *up_bo * scale);
        if (lp->xl(i) != lp->xu(i)) ++lp->non_fixed_vars;
      }
    }
  } else {
    /* no scaling used */
    for (int i = 1; i <= Dimension(lp->xl); ++i, ++up_bo, ++low_bo) {
      while (lp_lps->mp_parts[i + d_i - 1] < 0) {
        ++d_i;
        ++up_bo; ++low_bo;
      }

      if (lp_lps->mp_parts[i + d_i - 1] > 0) {
        lp->xl(i) = -lp->infinite;
        lp->xu(i) =  lp->infinite;
        ++lp->non_fixed_vars;
      } else {
        lp->xl(i) = *low_bo;
        lp->xu(i) = *up_bo;
        if (lp->xl(i) != lp->xu(i)) ++lp->non_fixed_vars;
      }
    }
  }
} /* Sm_lps3_2::read_simple_bounds */

/**
 * Adjust the algorithm parameter @p eta to be problem dependent.  It becomes
 * the maximum of its previous value and \f$10^{-20}\f$ times the
 * maximum norm of the right hand sides of the inequality constraints,
 * \f[
 *   \eta = \max\{\eta, 10^{-20} \|a\|_{\infty}\}.
 * \f]
 *
 * @param[out] eta  parameter to be adjusted
 * @param[in]  lp   lp @p eta is adjusted to
 */
void
Sm_lps3_2::adjust_eta(double &eta, const Lp *lp)
{
  /* <lp->ia> is a point vector here */
  VECTOR v = 1e-20 * Abs(lp->ia);

  for (int i = 1; i <= Dimension(v); ++i) {
    if (v(i) > eta) eta = v(i);
  }

  preport->print(5, "Adjusting eta to %.16e", eta);
}

/**
 * Inflate all lp parameters to intervals with relative radius @p r,
 * @f[
 *   x \quad \rightarrow \quad \mathbf{x} \triangleq x \cdot [1-r, 1+r]
 * @f]
 *
 * @param[out] lp  lp to be inflated
 * @param[in]  r   relative interval radius
 */
void
Sm_lps3_2::inflate_lp(Lp *lp, const double r)
{
  lp->IA = Hull((1 - r) * Inf(lp->IA), (1 + r) * Sup(lp->IA));
  lp->IB = Hull((1 - r) * Inf(lp->IB), (1 + r) * Sup(lp->IB));
  lp->ia = Hull((1 - r) * Inf(lp->ia), (1 + r) * Sup(lp->ia));
  lp->ib = Hull((1 - r) * Inf(lp->ib), (1 + r) * Sup(lp->ib));
  lp->ic = Hull((1 - r) * Inf(lp->ic), (1 + r) * Sup(lp->ic));
}

/**
 * Transform lp from the data structure of lp_solve 3.2 into the representation used in Lurupa.
 * Adjust algorithm parameters @p eta according to the lp, and inflate the lp to an interval valued
 * one if @p relative_interval_radius is greater than 0.
 *
 * Write the (midpoint) problem with #preport if Report#write_vm is set.
 *
 * @param[out]  lp                        receives the transformed data
 * @param[in]   relative_interval_radius  relative interval radius to inflate lp parameters to
 * @param[out]  eta                       gets adjusted to the model
 * 
 * @retval true
 * if the transformation succeeds,
 * @retval false
 * otherwise, @p lp->module->lp is not set or @c build_constraint_maps returns false
 *
 * @sa
 * #adjust_eta,
 * #build_constraint_maps,
 * #find_free_variables,
 * #inflate_lp,
 * #read_general_data,
 * #read_lp_mat,
 * #read_right_hand_sides,
 * #read_simple_bounds,
 * #resize_lp
 */
bool
Sm_lps3_2::transform_lp(Lp *lp, const double relative_interval_radius, double &eta)
{
  preport->print(3, "Reading lp from lp_solve 3.2");

  if (static_cast<Lp_lp_solve *>(lp->module)->lp == NULL) {
    preport->print(1, "Sm_lps3_2::transform_lp: model not set");
    return false;
  }

  /* read general data */
  read_general_data(lp);

  int c_eq = 0, c_le = 0;
  int rows = static_cast<Lp_lp_solve *>(lp->module)->lp->rows;
  int cols = static_cast<Lp_lp_solve *>(lp->module)->lp->columns;

  /* build constraint maps and count equations and inequalities */
  if (!build_constraint_maps(c_eq, c_le, rows, cols, lp)) {
    return false;
  }

  /* collect free variables if present in the lp */
  find_free_variables(cols, lp);

  preport->print(3, "%d variables: %d of them free",
                 cols, lp->free_variables_size);
  preport->print(3, "%d constraints: %d equalities, %d inequalities",
                 rows, c_eq, c_le);

  /* resize the data structures */
  resize_lp(lp, cols, c_eq, c_le);

  /* read the matrix and the objective function */
  read_lp_mat(rows, lp);

  /* read the right hand sides of the constraints */
  read_right_hand_sides(lp);

  /* read the simple bounds of the variables */
  read_simple_bounds(lp, rows);

  /* make eta problem dependent */
  adjust_eta(eta, lp);

  if (preport->get_write_vm()) {
    /* write the model to a file */
    preport->write_matrix(lp->name, Mid(lp->IA), "A");
    preport->write_vector(lp->name, Mid(lp->ia), "a");
    preport->write_matrix(lp->name, Mid(lp->IB), "B");
    preport->write_vector(lp->name, Mid(lp->ib), "b");
    preport->write_vector(lp->name, Mid(lp->ic), "c");
    preport->write_vector(lp->name, lp->xl, "xl");
    preport->write_vector(lp->name, lp->xu, "xu");
  }

  if (relative_interval_radius > 0)
    /* inflate the lp to an interval one */
    inflate_lp(lp, relative_interval_radius);

  return true;
}

/**
 * Read an lp from a file into lp_solve and @p lp. Inflate the lp to an interval valued one if @p
 * relative_interval_radius is greater than 0, and adjust @p eta according to the lp's parameters.
 *
 * @param[out] lp                        receives the lp
 * @param[in]  in                        file pointer to %lp
 * @param[in]  relative_interval_radius  interval radius to inflate lp parameters to
 * @param[out] eta                       algorithm parameter to be adjusted
 *
 * @retval true
 * if reading and transforming the lp succeeds,
 * @retval false
 * otherwise, @c transform_lp returns false
 *
 * @sa
 * #transform_lp
 */
bool
Sm_lps3_2::read_lp(Lp *lp, FILE *in, const double relative_interval_radius, double &eta)
{
  lp->module = new Lp_lp_solve;
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  BiasRoundNear();
  /* we need to initialize some values that don't get set in lp_solve
   * otherwise */
  Trej = 1e-4;

  lp_lps->lp = read_mps(in, FALSE);

  /* and some more... */
  lp_lps->lp->anti_degen = FALSE;
  lp_lps->lp->do_presolve = FALSE;

  return transform_lp(lp, relative_interval_radius, eta);
} /* Sm_lps3_2::read_lp */

bool
Sm_lps3_2::set_lp(Lp *lp, const double relative_interval_radius, double &eta)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  return transform_lp(lp, relative_interval_radius, eta);
} /* Sm_lps3_2::set_lp */

/**
 * Solve an lp with lp_solve. Store lp_solve's status in @p lp->feasibility.
 *
 * @param[out] lp  receives lp_solve's status 
 *
 * @retval true
 * if lp_solve returns a known status code (i.e., OPTIMAL, INFEASIBLE,
 * UNBOUNDED, or FAILURE)
 * @retval false
 * otherwise
 */
bool
Sm_lps3_2::solve_lp(Lp *lp) 
{
  BiasRoundNear();
  int code = solve(static_cast<Lp_lp_solve *>(lp->module)->lp);

  bool success = true;
  switch(code) {
  case OPTIMAL:
    lp->feasibility = ss_feasible; 
    preport->print(5, "OPTIMAL, objective = %.16e", 
                   static_cast<Lp_lp_solve *>(lp->module)->lp->best_solution[0]);
    break;
  case INFEASIBLE:
    lp->feasibility = ss_infeasible; 
    preport->print(5, "INFEASIBLE");
    break;
  case UNBOUNDED:
    lp->feasibility = ss_unbounded;
    preport->print(5, "UNBOUNDED");
    break;
  case FAILURE:
    lp->feasibility = ss_failure;
    preport->print(5, "FAILURE");
    break;
  default:
    lp->feasibility = ss_failure;
    preport->print(1, "Sm_lps3_2::solve_lp: unknown code %d", code);
    success = false;
  }

  return success;
} /* Sm_lps3_2::solve_lp */

/**
 * Read the primal solution from lp_solve and store it in @p lp->ix. If the lp seems to be
 * unbounded construct the last primal feasible point and read this. Join the solution parts of
 * split variables.
 *
 * @param[out] lp    receives the primal solution
 * @param[in]  rows  number of rows in constraint matrix
 */
void
Sm_lps3_2::read_primals(Lp *lp, const int rows)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  if (lp->feasibility == ss_unbounded) {
    /* construct last feasible point */
    construct_solution(lp_lps->lp);
    memcpy(lp_lps->lp->best_solution, lp_lps->lp->solution, (lp_lps->lp->sum + 1) * sizeof(REAL));
  }

  int d_i = 0;
  for (int i = 1; i <= Dimension(lp->ix); ++i) {
    while (lp_lps->mp_parts[i + d_i - 1] < 0) {
      ++d_i;
    }

    if (lp_lps->mp_parts[i + d_i - 1] > 0) {
      lp->ix(i) = Hull(lp_lps->lp->best_solution[rows + i + d_i] - 
                       lp_lps->lp->best_solution[rows + lp_lps->mp_parts[i + d_i - 1]]);
    } else {
      lp->ix(i) = Hull(lp_lps->lp->best_solution[rows + i + d_i]);
    }
  }

#ifndef NDEBUG
  /* What kind of output do we get in case of infeasible/unbounded linear
   * programs? */
  preport->print(6, "lp_solve's c'x = %.16e", lp_lps->lp->best_solution[0]);
  preport->print(6, "read c'x       = %.16e", Inf(lp->ic * lp->ix));
  /* Set rounding mode to nearest - we only need this if we actually print these
   * debug infos. */
  BiasRoundNear();
#endif
} /* Sm_lps3_2::read_primals */

/**
 * Read the dual solution from lp_solve and store it in @p lp->iy and @p lp->iz. Lp_solve's
 * convention to determine the sign of the duals is, that the costs must be a linear combination of
 * the gradients of the constraints. 
 *
 * The duals returned by lp_solve, however, are the duals for the original lp,
 * that is the duals for greater equal constraints are multiplied by \f$-1\f$,
 * and all duals are again multiplied by \f$-1\f$ if the lp is a maximizing one.
 * As %Lurupa stores all linear programs as minimizing and all inequality
 * constraints as less equal undo these multiplications.
 *
 * @param[out] lp  receives the dual solution
 */
void
Sm_lps3_2::read_duals(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* this can only happen, if lp->feasibility != ss_unbounded */
  if (lp->feasibility == ss_infeasible) calculate_duals(lp_lps->lp);

  /* lp_solve 3.2 uses the same convention to determine the sign of the duals as
   * we do. The costs must be a linear combination of the gradients of the
   * constraints however these get multiplied by -1 for greater equal
   * constraints and all are multiplied by -1 if the problem is a maximizing
   * one. */
  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->iy); ++i) {
      if (lp_lps->lp->ch_sign[lp_lps->mp_le_con[i - 1]]) {
        lp->iy(i) =  Hull(lp_lps->lp->duals[lp_lps->mp_le_con[i - 1]]);
      } else {
        lp->iy(i) = -Hull(lp_lps->lp->duals[lp_lps->mp_le_con[i - 1]]);
      }
    }

    for (int i = 1; i <= Dimension(lp->iz); ++i) {
      lp->iz(i) = -Hull(lp_lps->lp->duals[lp_lps->mp_eq_con[i - 1]]);
    }
  } else {
    for (int i = 1; i <= Dimension(lp->iy); ++i) {
      if (lp_lps->lp->ch_sign[lp_lps->mp_le_con[i - 1]]) {
        lp->iy(i) = -Hull(lp_lps->lp->duals[lp_lps->mp_le_con[i - 1]]);
      } else {
        lp->iy(i) =  Hull(lp_lps->lp->duals[lp_lps->mp_le_con[i - 1]]);
      }
    }

    for (int i = 1; i <= Dimension(lp->iz); ++i) {
      lp->iz(i) = Hull(lp_lps->lp->duals[lp_lps->mp_eq_con[i - 1]]);
    }
  }
} /* Sm_lps3_2::read_duals */

/**
 * Solve the original unperturbed lp with lp_solve and read the primal and dual
 * solution.
 *
 * @param[out] lp             receives the solver status and the solution
 * @param[out] optimal_value  receives the optimal value
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp returns false
 *
 * @sa
 * #read_duals,
 * #read_primals,
 * #solve_lp
 */
bool
Sm_lps3_2::solve_original(Lp *lp, double &optimal_value) 
{
  preport->print(3, "Solving original lp");

  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  bool success = solve_lp(lp);

  if (success) {
    optimal_value = (double) lp_lps->best_solution[0];

    /* read the solution and the value of the duals */
    read_primals(lp, lp_lps->rows);
    read_duals(lp);
  }

  return success;
} /* Sm_lps3_2::solve_original */

/**
 * Perturb the lp in the right hand sides of the inequalities and the simple bounds.
 *
 * @param[in]  d   deflation parameters
 * @param[out] lp  contains the base for the deflation
 */
void
Sm_lps3_2::primal_perturb(const Primal_deflation &d, const Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  if (lp_lps->lp->scaling_used) {
    for (int i = 1; i <= Dimension(d.a); ++i) {
      lp_lps->lp->orig_rh[lp_lps->mp_le_con[i - 1]] = 
        (Mid(lp->ia(i)) - d.a(i)) * lp_lps->lp->scale[lp_lps->mp_le_con[i - 1]];
    }
  } else {
    for (int i = 1; i <= Dimension(d.a); ++i) {
      lp_lps->lp->orig_rh[lp_lps->mp_le_con[i - 1]] = (Mid(lp->ia(i)) - d.a(i));
    }
  }

  int d_i = 0;
  for (int i = 1; i <= Dimension(d.xl); ++i) {
    while (lp_lps->mp_parts[i + d_i - 1] < 0) {
      ++d_i;
    }

    /* we don't have to deal with the positive part of free variables as their
     * bounds are never perturbed */
    if (d.xl(i) != 0) set_lowbo(lp_lps->lp, i + d_i, lp->xl(i) + d.xl(i));
    if (d.xu(i) != 0)  set_upbo(lp_lps->lp, i + d_i, lp->xu(i) - d.xu(i));
  }
}

/**
 * Perturb the lp and solve the new one. Read the primal solution and store it
 * in @p lp->ix.
 *
 * @param[in]  d          deflation parameters
 * @param[out] lp         receives the solver status and the primal solution
 * @param[in]  iteration  number of the current iteration
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp returns false
 *
 * @sa
 * #solve_lp
 */
bool
Sm_lps3_2::solve_primal_perturbed(const Primal_deflation &d, Lp *lp, const int iteration)
{
  preport->print(3, "Solving primal perturbed lp");

  primal_perturb(d, lp);

  bool success = solve_lp(lp);

  /* Read the primal solution if the problem is feasible or unbounded. We might
   * get a bound from a last primal feasible point in the latter case. */
  if (success) {
    switch (lp->feasibility) {
    case ss_unbounded:
    case ss_feasible:
      read_primals(lp, static_cast<Lp_lp_solve *>(lp->module)->lp->rows);
    }
  }

  return success;
}

/**
 * Perturb the lp in the cost vector.
 *
 * @param[in]  d_c  deflation parameters
 * @param[out] lp   contains the base for the deflation
 */
void
Sm_lps3_2::dual_perturb(const VECTOR &d_c, const Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* if the lp is a maximizing one we need to multiply our minimizing cost 
   * vector by -1 */
  int d_i = 0;
  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      while (lp_lps->mp_parts[i + d_i - 1] < 0) {
        ++d_i;
      }

      if (lp_lps->mp_parts[i + d_i - 1] == 0) {
        set_mat(lp_lps->lp, 0, i + d_i, -(Mid(lp->ic(i)) + d_c(i)));
      }
    }
  } else {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      while (lp_lps->mp_parts[i + d_i - 1] < 0) {
        ++d_i;
      }

      if (lp_lps->mp_parts[i + d_i - 1] == 0) {
        set_mat(lp_lps->lp, 0, i + d_i, Mid(lp->ic(i)) + d_c(i));
      }
    }
  }
}

/**
 * Perturb the lp and solve the new one. Read the dual solution from lp_solve
 * and store it in @p lp->iy and @p lp->iz.
 *
 * @param[in]  d_c        deflation parameter
 * @param[out] lp         receives the solver status and the dual solution
 * @param[in]  iteration  number of the current iteration
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp returns false
 *
 * @sa
 * #solve_lp
 */
bool
Sm_lps3_2::solve_dual_perturbed(const VECTOR &d_c, Lp *lp, const int iteration)
{
  preport->print(3, "Solving dual perturbed lp");

  dual_perturb(d_c, lp);

  bool success = solve_lp(lp);

  /* Read the dual solution if the problem is feasible or infeasible, which
   * means dual unbounded. We might get a bound from a last dual feasible point
   * in the latter case. */
  if ((success)
      && ((lp->feasibility == ss_feasible)
          || (lp->feasibility == ss_infeasible)))
    read_duals(lp);

  return success;
}

/**
 * Take the model parameters for the right hand sides and the simple bounds from @p lp and write to
 * @p lp->module->lp restoring the original model after computation of the upper bound.
 *
 * @param[in] lp  contains the original parameters
 */
void
Sm_lps3_2::restore_primal(Lp *lp) 
{
  preport->print(3, "Restoring original lp");

  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  if (lp_lps->lp->scaling_used) {
    for (int i = 1; i <= RowDimension(lp->IA); ++i) {
      lp_lps->lp->orig_rh[lp_lps->mp_le_con[i - 1]] = 
        Mid(lp->ia(i)) * lp_lps->lp->scale[lp_lps->mp_le_con[i - 1]];
    }
  } else {
    for (int i = 1; i <= RowDimension(lp->IA); ++i) {
      lp_lps->lp->orig_rh[lp_lps->mp_le_con[i - 1]] = Mid(lp->ia(i));
    }
  }

  int d_i = 0;
  for (int i = 1; i <= Dimension(lp->xl); ++i) {
    while (lp_lps->mp_parts[i + d_i - 1] < 0) {
      ++d_i;
    }

    set_lowbo(lp_lps->lp, i, lp->xl(i));
    set_upbo(lp_lps->lp, i, lp->xu(i));
  }
} /* Sm_lps3_2::restore_primal */

/**
 * Take the model parameters for the costs from @p lp and write to @p lp->module->lp restoring the
 * original model after computation of the lower bound.
 *
 * @param[in] lp  contains the original parameters
 */
void
Sm_lps3_2::restore_dual(Lp *lp) 
{
  preport->print(3, "Restoring original lp");

  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  int d_i = 0;
  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      while (lp_lps->mp_parts[i + d_i - 1] < 0) {
        ++d_i;
      }
      if (lp_lps->mp_parts[i + d_i - 1] == 0) {
        set_mat(lp_lps->lp, 0, i + d_i, -Mid(lp->ic(i)));
      }
    }
  } else {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      while (lp_lps->mp_parts[i + d_i - 1] < 0) {
        ++d_i;
      }

      if (lp_lps->mp_parts[i + d_i - 1] == 0) {
        set_mat(lp_lps->lp, 0, i + d_i, Mid(lp->ic(i)));
      }
    }
  }
} /* Sm_lps3_2::restore_dual */


/**
 * Get lp_solve's set accuracy
 *
 * @return
 * lp_solve's accuracy
 */
double
Sm_lps3_2::get_accuracy() 
{
  return DEF_EPSD;
} /* Sm_lps3_2::get_accuracy */

/**
 * Set the pointers of @p i to point to the member functions of Sm_lps3_2.
 *
 * @param[out]  i  receives the pointers to member functions
 */
extern "C" void
get_func_pointers(Solver_module_interface &i) 
{
  i.get_version            = Sm_lps3_2::get_version;
  i.print_version          = Sm_lps3_2::print_version;
  i.print_brief_version    = Sm_lps3_2::print_brief_version;
  i.print_options          = Sm_lps3_2::print_options;
  i.init                   = Sm_lps3_2::init;
  i.solve_original         = Sm_lps3_2::solve_original;
  i.solve_primal_perturbed = Sm_lps3_2::solve_primal_perturbed;
  i.solve_dual_perturbed   = Sm_lps3_2::solve_dual_perturbed;
  i.get_dual_ray           = NULL;
  i.get_primal_ray	       = NULL;
  i.set_primal_phase1	     = NULL;
  i.set_dual_phase1	       = NULL;
  i.restore_primal         = Sm_lps3_2::restore_primal;
  i.restore_dual           = Sm_lps3_2::restore_dual;
  i.restore_primal_phase1  = NULL;
  i.restore_dual_phase1    = NULL;
  i.read_lp                = Sm_lps3_2::read_lp;
  i.set_lp		             = Sm_lps3_2::set_lp;
  i.set_module_options     = NULL;
  i.get_accuracy           = Sm_lps3_2::get_accuracy;
} /* get_func_pointers */
