/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file defines the solver module for lp_solve 4.0.1.0, Sm_lps4_0_1_0.
 *
 * @file  Sm_lps4_0_1_0.cpp
 * @brief Definition of lp_solve 4.0.1.0 solver module
 *
 * @author Christian Keil
 *
 * $Id: Sm_lps4_0_1_0.cpp 533 2008-06-12 20:08:12Z keil $
 */
#include "Sm_lps4_0_1_0.h"
#include <lurupa/Lp.h>
#include <errno.h>
#include <string.h>

/** @brief Version string */
static const char *version = "Lp_solve 4.0.1.0 module 1.0";

Report *Solver_module::preport = NULL;

Lp_lp_solve::Lp_lp_solve()
{
  mp_eq_con = NULL;
  mp_le_con = NULL;
}

Lp_lp_solve::~Lp_lp_solve()
{
  delete_lp(lp);
  lp = NULL;
  delete [] mp_eq_con;
  delete [] mp_le_con;
}

/**
 * Get the version information of the solver module.
 */
const char *
Sm_lps4_0_1_0::get_version()
{
  return version;
} /* Sm_lps4_0_1_0::get_version */

void
Sm_lps4_0_1_0::print_version()
{
  int major, minor, release, build;
  lp_solve_version(&major, &minor, &release, &build);
  printf("solver module\n"
         "  version : %s\n"
         "  solver parameters\n"
         "    version : %d.%d.%d.%d\n"
         "    eps     : %g\n", 
         version, major, minor, release, build, DEF_EPSD);
} /* Sm_lps4_0_1_0::print_version */

void
Sm_lps4_0_1_0::print_brief_version()
{
  int major, minor, release, build;
  lp_solve_version(&major, &minor, &release, &build);
  printf("solver module       : %s, solver version = %d.%d.%d.%d, solver eps = %g\n",
         version, major, minor, release, build, DEF_EPSD);
} /* Sm_lps4_0_1_0::print_brief_version */

/**
 * Set #preport to \p *report. Check the version of lp_solve.
 *
 * \param[in] report  the #Report instance to use
 *
 * \retval true
 */
bool
Sm_lps4_0_1_0::init(Report *report)
{
  int major, minor, release, build;

  preport = report;
  preport->print(3, "Using %s", version);

  lp_solve_version(&major, &minor, &release, &build);
  if ((major != 4) || (minor != 0) || (release != 1) || (build != 0)) {
    preport->print(2, "Lp_solve version differs, expected: 4.0.1.0, found: %d.%d.%d.%d",
                   major, minor, release, build);
  }

  return true;
} /* Sm_lps4_0_1_0::init */

/**
 * Print the command line options supported by the solver module.
 */
void
Sm_lps4_0_1_0::print_options()
{
  printf("Options supported by %s:\n", version);
  printf("  -sm,timeout,i       Set solver timeout in seconds.\n");
  printf("  -sm,vn              Set solver verbosity:\n"
         "                        v0: NEUTRAL\n"
         "                        v1: CRITICAL\n"
         "                        v2: SEVERE\n"
         "                        v3: IMPORTANT (default)\n"
         "                        v4: NORMAL\n"
         "                        v5: DETAILED\n"
         "                        v6: FULL\n");
} /* Sm_lps4_0_1_0::print_options */

/**
 * Read general lp properties from lp_solve's structures and store them in @p lp. This includes the
 * value representing infinity, the direction of optimization, and the name of the lp.
 *
 * @param[out] lp  stores the read data
 */
void
Sm_lps4_0_1_0::read_general_data(Lp *lp)
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;
  lp->infinite = lp_lps->infinite;
  lp->maximize = (bool) lp_lps->maximise;
  lp->name = strdup(lp_lps->lp_name);
} /* Sm_lps4_0_1_0::read_general_data */

/**
 * Build a mapping from Lurupa's equality and inequality constraints to lp_solve's constraints,
 * which is storing all constraints in one big matrix. Store the mapping in @p lp->module->mp_eq_con
 * and @p lp->module->mp_le_con. Check if the number of constraints is less than the number of
 * variables.
 *
 * @param[out] c_eq  number of equations in lp
 * @param[out] c_le  number of inequalities in lp
 * @param[in]  rows  number of rows of constraint matrix
 * @param[in]  cols  number of columns of constraint matrix
 * @param[in]  lp    the lp
 *
 * @retval true
 * if number of constraints is less than number of variables,
 * @retval false
 * otherwise
 */
bool
Sm_lps4_0_1_0::build_constraint_maps(int &c_eq, int &c_le, const int rows, const int cols, Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* count equations and inequalities */
  for (int i = 1; i <= rows; ++i) {
    if (get_constr_type(lp_lps->lp, i) == EQ) {
      ++c_eq;
    } else {
      ++c_le;
    }
  }

  if (c_eq > cols) {
    preport->print(1, "Sm_lps4_0_1_0::build_constraint_maps: model contains more "
                   "equations (%d) than variables (%d)", c_eq, cols);
    return false;
  }

  /* build maps */
  lp_lps->mp_eq_con = new int[c_eq];
  lp_lps->mp_le_con = new int[c_le];
  c_eq = -1; c_le = -1;
  for (int i = 1; i <= rows; ++i) {
    if (get_constr_type(lp_lps->lp, i) == EQ) {
      lp_lps->mp_eq_con[++c_eq] = i;
    } else {
      lp_lps->mp_le_con[++c_le] = i;
    }
  }

  ++c_le;
  ++c_eq;

  return true;
} /* Sm_lps4_0_1_0::build_constraint_maps */

/**
 * Find free variables in the lp. Store the indices of the free variables in @p lp->free_variables
 * and their number in @p lp->free_variables_size.
 *
 * @param[in]  vars   number of variables in lp
 * @param[out] lp     stores number and indices of the free variables
 */
void
Sm_lps4_0_1_0::find_free_variables(const int vars, Lp *lp)
{
  /* Since we completed the postprocessing all free variables have their
   * infinite bounds restored, so we just have to check for variables with
   * infinite lower and upper bounds. */
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  /* count free variables */
  for (int i = 1; i <= vars; ++i) {
    if ((get_lowbo(lp_lps, i) == -lp->infinite) 
        && (get_upbo(lp_lps, i) ==  lp->infinite)) {
      ++lp->free_variables_size;
    }
  }
  lp->free_variables = new int[lp->free_variables_size];

  /* store indices */
  int var = 1, free_var = 0;
  while (free_var < lp->free_variables_size) {
    if ((get_lowbo(lp_lps, var) == -lp->infinite) 
        && (get_upbo(lp_lps, var) ==  lp->infinite)) {
      lp->free_variables[free_var++] = var;
    }
    ++var;
  }
}

/**
 * Resize the vectors and matrices in Lurupa's data structure to the sizes of
 * the lp.
 *
 * @param[out] lp     Lurupa's structure with the vectors and matrices to be resized
 * @param[in]  cols   number of variables in the lp
 * @param[in]  c_eq   number of equality constraints in the lp
 * @param[in]  c_le   number of inequality constraints in the lp
 */
void
Sm_lps4_0_1_0::resize_lp(Lp *lp, const int cols, const int c_eq, const int c_le)
{
  Resize(lp->IA, c_le, cols);
  Resize(lp->IB, c_eq, cols);
  Resize(lp->ic, cols);
  Resize(lp->ia, c_le);
  Resize(lp->ib, c_eq);
  Resize(lp->xl, cols);
  Resize(lp->xu, cols);
  Resize(lp->ix, cols);
  Resize(lp->iy, c_le);
  Resize(lp->iz, c_eq);
} /* Sm_lps4_0_1_0::resize_lp */

/**
 * Read lp_solve's @c lp_mat structure. This stores the left hand sides of the
 * constraints together with the objective function. Split the read coefficients
 * into equality and inequality constraints, transform inequality constraints to
 * all less equal, and the direction of optimization to minimize.
 *
 * @param[in]  rows   number of rows in constraint matrix
 * @param[out] lp     stores the constraints and objective function
 */
void
Sm_lps4_0_1_0::read_lp_mat(const int rows, Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  REAL *column = new REAL[rows + 1];
  for (int j = 1; j <= ColDimension(lp->IA); ++j) {
    /* get_column returns the parameters of the _original_ lp, so we need to
     * transform into our representation */
    get_column(lp_lps->lp, j, column);

    /* maximizing linear programs become minimizing ones: multiply objective
     * function coefficients by -1 */
    lp->ic(j) = lp->maximize ? -Hull(column[0]) : Hull(column[0]);

    for (int i = 1; i <= RowDimension(lp->IA); ++i) {
      /* >= constraints become <= constraints: multiply coefficients by -1 */
      if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
        lp->IA(i, j) =  Hull(column[lp_lps->mp_le_con[i - 1]]);
      } else {
        lp->IA(i, j) = -Hull(column[lp_lps->mp_le_con[i - 1]]);
      }
    }

    for (int i = 1; i <= RowDimension(lp->IB); ++i) {
      lp->IB(i, j) = Hull(column[lp_lps->mp_eq_con[i - 1]]);
    }
  }
  delete [] column;
} /* Sm_lps4_0_1_0::read_lp_mat */

/**
 * Read the right hand sides of the constraints. Split the read right hand sides
 * into equality and inequality constraints, transform inequality constraints to
 * all less equal. Store the values in @p lp->ia and @p lp->ib.
 *
 * @param[out] lp  stores the right hand sides
 */
void
Sm_lps4_0_1_0::read_right_hand_sides(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  for (int i = 1; i <= Dimension(lp->ia); ++i) {
    /* transform >= constraints to <= */
    if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
      lp->ia(i) =  Hull(get_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1]));
    } else {
      lp->ia(i) = -Hull(get_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1]));
    }
  }

  for (int i = 1; i <= Dimension(lp->ib); ++i) {
    lp->ib(i) = Hull(get_rh(lp_lps->lp, lp_lps->mp_eq_con[i - 1]));
  }
} /* Sm_lps4_0_1_0:read_right_hand_sides */

/**
 * Read the simple bounds of the variables and store them in @p lp->xl and
 * @p lp->xu.
 *
 * @param[out] lp  stores the simple bounds
 */
void
Sm_lps4_0_1_0::read_simple_bounds(Lp *lp)
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  for (int i = 1; i <= Dimension(lp->xl); ++i) {
    lp->xl(i) = get_lowbo(lp_lps, i);
    lp->xu(i) = get_upbo(lp_lps, i);

    if (lp->xl(i) != lp->xu(i)) ++lp->non_fixed_vars;
  }
} /* Sm_lps4_0_1_0::read_simple_bounds */

/**
 * Adjust the algorithm parameter @p eta to be problem dependent.  It becomes
 * the maximum of its previous value and \f$10^{-20}\f$ times the
 * maximum norm of the right hand sides of the inequality constraints,
 * \f[
 *   \eta = \max\{\eta, 10^{-20} \|a\|_{\infty}\}.
 * \f]
 *
 * @param[out] eta  parameter to be adjusted
 * @param[in]  lp   lp @p eta is adjusted to
 */
void
Sm_lps4_0_1_0::adjust_eta(double &eta, const Lp *lp)
{
  /* <lp->ia> is a point vector here */
  VECTOR v = 1e-20 * Abs(lp->ia);

  for (int i = 1; i <= Dimension(v); ++i) {
    if (v(i) > eta) eta = v(i);
  }

  preport->print(5, "Adjusting eta to %.16e", eta);
} /* Sm_lps4_0_1_0::adjust_eta */

/**
 * Inflate all lp parameters to intervals with relative radius @p r,
 * @f[
 *   x \quad \rightarrow \quad \mathbf{x} \triangleq x \cdot [1-r, 1+r]
 * @f]
 *
 * @param[out] lp  lp to be inflated
 * @param[in]  r   relative interval radius
 */
void
Sm_lps4_0_1_0::inflate_lp(Lp *lp, const double r)
{
  lp->IA = Hull((1 - r) * Inf(lp->IA), (1 + r) * Sup(lp->IA));
  lp->IB = Hull((1 - r) * Inf(lp->IB), (1 + r) * Sup(lp->IB));
  lp->ia = Hull((1 - r) * Inf(lp->ia), (1 + r) * Sup(lp->ia));
  lp->ib = Hull((1 - r) * Inf(lp->ib), (1 + r) * Sup(lp->ib));
  lp->ic = Hull((1 - r) * Inf(lp->ic), (1 + r) * Sup(lp->ic));
} /* Sm_lps4_0_1_0::inflate_lp */

/**
 * Callback function for lp_solve's log functionality. Print the supplied log
 * message marked as coming from lp_solve.
 */
extern "C" void WINAPI
Sm_lps4_0_1_0::lps_log(void *lp, void *userhandle, char *buf)
{
  /* Remove \n's from the start of buf */
  while (*buf == '\n') ++buf;
  preport->print(0, "[lps4010] %s", buf);
}

/**
 * Transform lp from the data structure of lp_solve 4.0.1.0 into the representation used in Lurupa.
 * Adjust algorithm parameters @p eta according to the lp, and inflate the lp to an interval valued
 * one if @p relative_interval_radius is greater than 0.
 *
 * Write the (midpoint) problem with #preport if Report#write_vm is set.
 *
 * @param[out]  lp                        receives the transformed data
 * @param[in]   relative_interval_radius  relative interval radius to inflate lp parameters to
 * @param[out]  eta                       gets adjusted to the model
 * 
 * @retval true
 * if the transformation succeeds,
 * @retval false
 * otherwise, @p lp->module->lp is not set or @c build_constraint_maps returns false
 *
 * @sa
 * #adjust_eta,
 * #build_constraint_maps,
 * #find_free_variables,
 * #inflate_lp,
 * #read_general_data,
 * #read_lp_mat,
 * #read_right_hand_sides,
 * #read_simple_bounds,
 * #resize_lp
 */
bool
Sm_lps4_0_1_0::transform_lp(Lp *lp, const double relative_interval_radius, double &eta)
{
  preport->print(3, "Reading lp from lp_solve 4.0.1.0");

  if (static_cast<Lp_lp_solve *>(lp->module)->lp == NULL) {
    preport->print(1, "Sm_lps4_0_1_0::transform_lp: model not set");
    return false;
  }

  /* read general data */
  read_general_data(lp);

  int c_eq = 0, c_le = 0;
  int rows = get_Nrows(static_cast<Lp_lp_solve *>(lp->module)->lp);
  int cols = get_Ncolumns(static_cast<Lp_lp_solve *>(lp->module)->lp);

  /* build constraint maps and count equations and inequalities */
  if (!build_constraint_maps(c_eq, c_le, rows, cols, lp)) {
    return false;
  }

  /* collect free variables if present in the lp */
  find_free_variables(cols, lp);

  preport->print(3, "%d variables: %d of them free",
                 cols, lp->free_variables_size);
  preport->print(3, "%d constraints: %d equalities, %d inequalities",
                 rows, c_eq, c_le);

  /* resize the data structures */
  resize_lp(lp, cols, c_eq, c_le);

  /* read the matrix and the objective function */
  read_lp_mat(rows, lp);

  /* read the right hand sides of the constraints */
  read_right_hand_sides(lp);

  /* read the simple bounds of the variables */
  read_simple_bounds(lp);

  /* make eta problem dependent */
  adjust_eta(eta, lp);

  if (preport->get_write_vm()) {
    /* write the model to a file */
    preport->write_matrix(lp->name, Mid(lp->IA), "A");
    preport->write_vector(lp->name, Mid(lp->ia), "a");
    preport->write_matrix(lp->name, Mid(lp->IB), "B");
    preport->write_vector(lp->name, Mid(lp->ib), "b");
    preport->write_vector(lp->name, Mid(lp->ic), "c");
    preport->write_vector(lp->name, lp->xl, "xl");
    preport->write_vector(lp->name, lp->xu, "xu");
  }

  if (relative_interval_radius > 0) {
    /* inflate the lp to an interval one */
    inflate_lp(lp, relative_interval_radius);
  }

  return true;
}

/**
 * Read an lp from a file into lp_solve and @p lp. Inflate the lp to an interval valued one if @p
 * relative_interval_radius is greater than 0, and adjust @p eta according to the lp's parameters.
 *
 * @param[out] lp                        receives the lp
 * @param[in]  in                        file pointer to %lp
 * @param[in]  relative_interval_radius  nterval radius to inflate lp parameters to
 * @param[out] eta                       algorithm parameter to be adjusted
 *
 * @retval true
 * if reading and transforming the lp succeeds,
 * @retval false
 * otherwise, reading the lp fails or @c transform_lp returns false
 *
 * @sa
 * #transform_lp
 */
bool
Sm_lps4_0_1_0::read_lp(Lp *lp, FILE *in, const double relative_interval_radius, double &eta)
{
  bool success = true;
  lp->module = new Lp_lp_solve;
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  BiasRoundNear();
  if (NULL == (lp_lps->lp = read_mps(in, IMPORTANT))) {
    success = false;
  }

  if (success) {
    /* Print log messages ourself, this way we can prefix them and thus
     * distinguish from ours. */
    put_logfunc(lp_lps->lp, lps_log, NULL);
    /* Set default verbosity level to 3 */
    set_verbose(lp_lps->lp, 3);

    success = transform_lp(lp, relative_interval_radius, eta);
  }

  if (success) {
    /* Disable presolving as lp_solve 4.0.1.0 offers no concept of mapping
     * presolved solutions to the original problem. */
    set_do_presolve(lp_lps->lp, FALSE);
  }

  return success;
} /* Sm_lps4_0_1_0::read_lp */

bool
Sm_lps4_0_1_0::set_lp(Lp *lp, const double relative_interval_radius, double &eta)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  return transform_lp(lp, relative_interval_radius, eta);
} /* Sm_lps4_0_1_0::set_lp */

/**
 * Construct 'last feasible points' if @p status indicates that lp_solve judges @p lp->module->lp to
 * be infeasible or unbounded. Postprocess the lp.
 *
 * When lp_solve determines an lp to be infeasible or unbounded it does not store any information
 * gathered in the simplex algorithm but discards it during the postprocessing. Deriving 'last
 * feasible points' from this information may deliver rigorous bounds to support lp_solve's claim.
 * Calling lp_solve's preprocessing routine explicitly, which does not alter the lp in a harmful way
 * regarding the verification, causes lp_solve not to postprocess the lp after solving. This way the
 * information in the simplex tableau is preserved.
 *
 * @param[in] lp      the lp
 * @param[in] status  return value of lp_solve solving the lp
 */
void
Sm_lps4_0_1_0::postprocess(Lp *lp, const int status) 
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  preport->print(4, "Postprocessing lp");

  /* construct 'last feasible points' if appropriate */
  if (status == INFEASIBLE) {
    calculate_duals(lp_lps);
  } else if (status == UNBOUNDED) {
    construct_solution(lp_lps);
    transfer_solution(lp_lps);
  }

  /* call lp_solves postprocess 
   * this is incomplete with regard to negative variables and split variables */
  ::postprocess(lp_lps);

  /* Postprocess negative variables
   * Lp_solve's postprocess restores the bounds and coefficients but does not
   * reset var_is_free, which flags negative variables transparently transformed
   * to positive ones.
   * Reset var_is_free entry of negative variables. Not resetting this results
   * (at least) in repetitive multiplication of the corresponding column with -1
   * during postprocessing which results in the first iteration in a variable
   * flagged as having been transformed to positive range which has a negative
   * range. Ugly!! 
   *
   * This should be no problem as the next preprocessing, which is always
   * performed prior solving the lp, transforms the negative variable back to
   * the state the flag suggests.
   * It could make problems if a variable is switched from being negative to not
   * being negative. This might result from inflation of the model. */
  for (int var = 1; var <= lp_lps->orig_columns; ++var) {
    if (lp_lps->var_is_free[var] == -1)  {
      lp_lps->var_is_free[var] = 0;
    }
  }

  /* Postprocess split variables 
   * Lp_solve's postprocess transfers the solution of split variables to the
   * original variable but does not restore the bounds. The flag indicating the
   * split is not reset and the helper column is not deleted. */
  int rows = get_Nrows(lp_lps);
  for (int var = get_Ncolumns(lp_lps); var > lp_lps->orig_columns; --var) {
    /* Reset var_is_free entry of original variable as a splitting candidate
     * might become a normal variable through perturbation. */
    lp_lps->var_is_free[-lp_lps->var_is_free[var]] = 0;
    /* Reset the simple bounds of the original variable as lp_solve handles it
     * this way: If the optimal solution is below the splitting threshold
     * <lp->negrange> the simple lower bound is reset to the - possibly
     * false - value <-lp->infinite>. If the optimal solution is in the upper
     * part of the split nothing is done. */
    set_lowbo(lp_lps, -lp_lps->var_is_free[var], -get_upbo(lp_lps, var));
    /* Delete the helper column. Not doing this would result in more columns
     * every iteration as postprocessing restores the original bounds and
     * preprocessing splits the variables again, appending helper columns.
     * If the helper column is in the basis this will invalidate the basis. To
     * keep it valid put the unsplitted variable into the basis. */
    if (lp_lps->basis[rows + var]) {
      /* lp_solve does not consider <lp->basis> during <del_column> so we
       * can savely leave the entry of the negative part <TRUE>. */
      lp_lps->basis[rows - lp_lps->var_is_free[var]] = TRUE;
      int bas = 1;
      while (lp_lps->bas[bas] != rows + var)
        ++bas;
      lp_lps->bas[bas] = rows - lp_lps->var_is_free[var];
    }
    del_column(lp_lps, var);
  }
} /* Sm_lps4_0_1_0::postprocess */

/**
 * Solve an lp with lp_solve. Store lp_solve's status in @p lp->feasibility.
 *
 * @param[out] lp  receives lp_solve's status 
 *
 * @retval true
 * if lp_solve returns a known status code (i.e., OPTIMAL, INFEASIBLE,
 * UNBOUNDED, TIMEOUT, or FAILURE)
 * @retval false
 * otherwise
 */
bool
Sm_lps4_0_1_0::solve_lp(Lp *lp) 
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  /* Explicitly call preprocess to keep lp_solve from postprocessing the lp by
   * itself. This way we can get hold of "last feasible points" and ease the
   * handling of free variables. */

  BiasRoundNear();
  preprocess(lp_lps);
  int code = solve(lp_lps);
  postprocess(lp, code);

  bool success = true;
  switch(code) {
  case OPTIMAL:
    lp->feasibility = ss_feasible; 
    preport->print(5, "OPTIMAL, objective = %.16e", get_objective(lp_lps));
    break;
  case INFEASIBLE:
    lp->feasibility = ss_infeasible; 
    preport->print(5, "INFEASIBLE");
    break;
  case UNBOUNDED:
    lp->feasibility = ss_unbounded;
    preport->print(5, "UNBOUNDED");
    break;
  case TIMEOUT:
    lp->feasibility = ss_timeout;
    preport->print(5, "TIMEOUT");
    break;
  case FAILURE:
    lp->feasibility = ss_failure;
    preport->print(5, "FAILURE");
    break;
  default:
    lp->feasibility = ss_failure;
    preport->print(1, "Sm_lps4_0_1_0::solve_lp: unknown code %d", code);
    success = false;
  }

  return success;
} /* Sm_lps4_0_1_0::solve_lp */

/**
 * Read the primal solution from lp_solve and store it in @p lp->ix.
 *
 * @param[out] lp     receives the primal solution
 * @param[in]  cols   number of variables in the lp
 *
 * @retval true
 * if reading the primal solution succeeds,
 * @retval false
 * otherwise, calling lp_solve's @c get_variables returns false
 */
bool
Sm_lps4_0_1_0::read_primals(Lp *lp, const int cols)
{
  REAL *solution = new REAL[cols];
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);
  if (!get_variables(lp_lps->lp, solution)) {
    preport->print(1, "Sm_lps4_0_1_0::read_primals: calling get_variables");
    return false;
  }

  for (int i = 1; i <= Dimension(lp->ix); ++i)
    lp->ix(i) =  Hull(solution[i - 1]);

  delete [] solution;

#ifndef NDEBUG
  /* What kind of output do we get in case of infeasible/unbounded linear
   * programs? */
  preport->print(6, "lp_solve's c'x = %.16e", get_objective(lp_lps->lp));
  preport->print(6, "read c'x       = %.16e", Inf(lp->ic * lp->ix));
  /* Set rounding mode to nearest - we only need this if we actually print these
   * debug infos. */
  BiasRoundNear();
#endif

  return true;
} /* Sm_lps4_0_1_0::read_primals */

/**
 * Read the dual solution from lp_solve and store it in @p lp->iy and @p lp->iz. Lp_solve's
 * convention to determine the sign of the duals is, that the costs must be a linear combination of
 * the gradients of the constraints. 
 *
 * The duals returned by lp_solve, however, are the duals for the original lp,
 * that is the duals for greater equal constraints are multiplied by \f$-1\f$,
 * and all duals are again multiplied by \f$-1\f$ if the lp is a maximizing one.
 * As %Lurupa stores all linear programs as minimizing and all inequality
 * constraints as less equal undo these multiplications.
 *
 * @param[out] lp  receives the dual solution
 *
 * @retval true
 * if reading the dual solution succeeded,
 * @retval false
 * otherwise, calling lp_solve's @c get_sensitivity_rhs returns false
 */
bool
Sm_lps4_0_1_0::read_duals(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  REAL *duals = new REAL[lp_lps->lp->sum];

  if (!get_sensitivity_rhs(lp_lps->lp, duals, NULL, NULL)) {
    preport->print(1, "Sm_lps4_0_1_0::read_duals: calling get_sensitivity_rhs");
    return false;
  }

  /* if we have a maximizing lp all returned duals have to be multiplied by -1 */
  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->iy); ++i) {
      /* if the constraint is a <= one no additional multiplication is necessary */
      if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
        lp->iy(i) = -Hull(duals[lp_lps->mp_le_con[i - 1] - 1]);
      } else {
        lp->iy(i) =  Hull(duals[lp_lps->mp_le_con[i - 1] - 1]);
      }
    }

    for (int i = 1; i <= Dimension(lp->iz); ++i) {
      lp->iz(i) = -Hull(duals[lp_lps->mp_eq_con[i - 1] - 1]);
    }
  } else {
    for (int i = 1; i <= Dimension(lp->iy); ++i) {
      /* if the constraint is a <= one no multiplication is necessary */
      if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
        lp->iy(i) =  Hull(duals[lp_lps->mp_le_con[i - 1] - 1]);
      } else {
        lp->iy(i) = -Hull(duals[lp_lps->mp_le_con[i - 1] - 1]);
      }
    }

    for (int i = 1; i <= Dimension(lp->iz); ++i) {
      lp->iz(i) = Hull(duals[lp_lps->mp_eq_con[i - 1] - 1]);
    }
  }

  delete [] duals;

  return true;
} /* Sm_lps4_0_1_0::read_duals */

/**
 * Solve the original unperturbed lp with lp_solve and read the primal and dual
 * solution.
 *
 * @param[out] lp             receives the solver status and the solution
 * @param[out] optimal_value  receives the optimal value
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp, @c read_primals, or @c read_duals returns false
 *
 * @sa
 * #read_duals,
 * #read_primals,
 * #solve_lp
 */
bool
Sm_lps4_0_1_0::solve_original(Lp *lp, double &optimal_value)
{
  preport->print(3, "Solving original lp");

  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  bool success = solve_lp(lp);

  if (success) {
    optimal_value = get_objective(lp_lps);

    /* read the solution and the value of the duals */
    if (!read_primals(lp, get_Ncolumns(lp_lps)) || !read_duals(lp))
      success = false;
  }

  return success;
} /* Sm_lps4_0_1_0::solve_original */

/**
 * Perturb the lp in the right hand sides of the inequalities and the simple bounds.
 *
 * @param[in]  d   deflation parameters
 * @param[out] lp  contains the base for the deflation
 */
void
Sm_lps4_0_1_0::primal_perturb(const Primal_deflation &d, const Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* perturb the right hand sides */
  for (int i = 1; i <= Dimension(d.a); ++i) {
    /* If the constraint is a "less or equal" one we can directly use the
     * perturbed right hand side. Otherwise we have to multiply it by -1 as
     * lp_solve expects the right hand side for a "greater or equal" constraint
     * in this case. */
    if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1],   Mid(lp->ia(i)) - d.a(i));
    } else {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1], -(Mid(lp->ia(i)) - d.a(i)));
    }
  }

  /* perturb the simple bounds */
  for (int i = 1; i <= Dimension(d.xl); ++i) {
    if (d.xl(i) != 0) set_lowbo(lp_lps->lp, i, lp->xl(i) + d.xl(i));
    if (d.xu(i) != 0)  set_upbo(lp_lps->lp, i, lp->xu(i) - d.xu(i));
  }
} /* Sm_lps4_0_1_0::primal_perturb */

/**
 * Perturb the lp and solve the new one. Read the primal solution and store it
 * in @p lp->ix.
 *
 * @param[in]  d          deflation parameters
 * @param[out] lp         receives the solver status and the primal solution
 * @param[in]  iteration  number of the current iteration
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp or @c read_primals returns false
 *
 * @sa
 * #read_primals,
 * #solve_lp
 */
bool
Sm_lps4_0_1_0::solve_primal_perturbed(const Primal_deflation &d, Lp *lp, const int iteration)
{
  preport->print(3, "Solving primal perturbed lp");

  primal_perturb(d, lp);

  bool success = solve_lp(lp);

  /* Read the primal solution if the problem is feasible or unbounded. We might
   * get a bound from a last primal feasible point in the latter case. */
  if (success
      && ((lp->feasibility == ss_feasible) || (lp->feasibility == ss_unbounded))) {
    success = read_primals(lp, get_Ncolumns(static_cast<Lp_lp_solve *>(lp->module)->lp));
  }

  return success;
}

/**
 * Perturb the lp in the cost vector.
 *
 * @param[in]  d_c  deflation parameters
 * @param[out] lp   contains the base for the deflation
 */
void
Sm_lps4_0_1_0::dual_perturb(const VECTOR &d_c, const Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* if the lp is a maximizing one we need to multiply our minimizing cost 
   * vector by -1 */
  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      if (0 != d_c(i)) {
        set_mat(lp_lps->lp, 0, i, -(Mid(lp->ic(i)) + d_c(i)));
      }
    }
  } else {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      if (0 != d_c(i)) {
        set_mat(lp_lps->lp, 0, i, Mid(lp->ic(i)) + d_c(i));
      }
    }
  }
}

/**
 * Perturb the lp and solve the new one. Read the dual solution from lp_solve
 * and store it in @p lp->iy and @p lp->iz.
 *
 * @param[in]  d_c        deflation parameter
 * @param[out] lp         receives the solver status and the dual solution
 * @param[in]  iteration  number of the current iteration
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp or @c read_duals returns false
 *
 * @sa
 * #read_duals,
 * #solve_lp
 */
bool
Sm_lps4_0_1_0::solve_dual_perturbed(const VECTOR &d_c, Lp *lp, const int iteration)
{
  preport->print(3, "Solving dual perturbed lp");

  dual_perturb(d_c, lp);

  bool success = solve_lp(lp);

  /* Read the dual solution if the problem is feasible or infeasible, which
   * means dual unbounded. We might get a bound from a last dual feasible point
   * in the latter case. */
  if (success
      && ((lp->feasibility == ss_feasible) || (lp->feasibility == ss_infeasible))) {
    success = read_duals(lp);
  }

  return success;
} /* Sm_lps4_0_1_0::solve_dual_perturbed */

/**
 * Take the model parameters for the right hand sides and the simple bounds from @p lp and write to
 * @p lp->module->lp restoring the original model after computation of the upper bound.
 *
 * @param[in] lp  contains the original parameters
 */
void
Sm_lps4_0_1_0::restore_primal(Lp *lp) 
{
  preport->print(3, "Restoring original lp");

  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  for (int i = 1; i <= RowDimension(lp->IA); ++i) {
    if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1],  Mid(lp->ia(i)));
    } else {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1], -Mid(lp->ia(i)));
    }
  }

  for (int i = 1; i <= Dimension(lp->xl); ++i) {
    set_lowbo(lp_lps->lp, i, lp->xl(i));
    set_upbo(lp_lps->lp, i, lp->xu(i));
  }
} /* Sm_lps4_0_1_0::restore_primal */

/**
 * Take the model parameters for the costs from @p lp and write to @p lp->module->lp restoring the
 * original model after computation of the lower bound.
 *
 * @param[in] lp  contains the original parameters
 */
void
Sm_lps4_0_1_0::restore_dual(Lp *lp) 
{
  preport->print(3, "Restoring original lp");

  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      set_mat(lp_lps, 0, i, -Mid(lp->ic(i)));
    }
  } else {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      set_mat(lp_lps, 0, i, Mid(lp->ic(i)));
    }
  }
} /* Sm_lps4_0_1_0::restore_dual */

/**
 * Set the specified module options. These are supplied in @p argc and @p argv
 * in a command line version. Each option is supplied as one string entry in the
 * string array @p argv, @p argc specifies the number of options.
 * 
 * The module supports the following options
 * - '<tt>-sm,timeout,n</tt>': Set lp_solve's timeout to <tt>n</tt> seconds.
 * - '<tt>-sm,vn</tt>': Set lp_solve's verbosity level to <tt>n</tt>. The
 *   available levels are
 *   - <tt>v0</tt> NEUTRAL
 *   - <tt>v1</tt> CRITICAL
 *   - <tt>v2</tt> SEVERE
 *   - <tt>v3</tt> IMPORTANT (default)
 *   - <tt>v4</tt> NORMAL
 *   - <tt>v5</tt> DETAILED
 *   - <tt>v6</tt> FULL
 *
 * @param[out] lp     lp to set the options for
 * @param[in]  argc   number of arguments
 * @param[in]  argv   array of arguments
 *
 * @retval true
 * if options are set
 * @retval false
 * otherwise
 */
bool
Sm_lps4_0_1_0::set_module_options(Lp *lp, int argc, char *argv[]) 
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  for (int i = 0; i < argc; ++i) {
    if (strncmp(argv[i], "-sm,timeout,", 12) == 0) {
      errno = 0;
      char *end;
      long timeout = strtol(argv[i] + 12, &end, 10);
      if (errno != 0) {
				preport->print(1, "Sm_lps4_0_1_0::set_module_options: %s", strerror(errno));
        return false;
      } else if (*end != '\0') {
        preport->print(1, "Sm_lps4_0_1_0::set_module_options: "
                       "error parsing timeout %s", argv[i]);
        return false;
      } else {
        set_timeout(lp_lps, timeout);
      }
    } else if (strncmp(argv[i], "-sm,v", 5) == 0) {
      errno = 0;
      char *end;
      int verbosity = strtol(argv[i] + 5, &end, 10);
      if (errno != 0) {
				preport->print(1, "Sm_lps4_0_1_0::set_module_options: %s", strerror(errno));
        return false;
      } else if (*end != '\0') {
        preport->print(1, "Sm_lps4_0_1_0::set_module_options: "
                       "error parsing verbosity %s", argv[i]);
      } else if ((verbosity < 0) || (verbosity > 6)) {
        preport->print(1, "Sm_lps4_0_1_0::set_module_options: "
                       "unknown verbosity level: %s", argv[i]);
        return false;
      } else {
        set_verbose(lp_lps, verbosity);
      }
    } else {
      preport->print(1, "Sm_lps4_0_1_0::set_module_options: unknown option: %s", argv[i]);
      return false;
    }
  }
  return true;
} /* Sm_lps4_0_1_0::set_module_options */

/**
 * Get lp_solve's set accuracy
 *
 * @return
 * lp_solve's accuracy
 */
double
Sm_lps4_0_1_0::get_accuracy() 
{
  return DEF_EPSD;
} /* Sm_lps4_0_1_0::get_accuracy */

/**
 * Set the pointers of @p i to point to the member functions of Sm_lps4_0_1_0.
 *
 * @param[out]  i  receives the pointers to member functions
 */
extern "C" void
get_func_pointers(Solver_module_interface &i) 
{
  i.get_version             = Sm_lps4_0_1_0::get_version;
  i.print_version           = Sm_lps4_0_1_0::print_version;
  i.print_brief_version     = Sm_lps4_0_1_0::print_brief_version;
  i.print_options           = Sm_lps4_0_1_0::print_options;
  i.init                    = Sm_lps4_0_1_0::init;
  i.solve_original          = Sm_lps4_0_1_0::solve_original;
  i.solve_primal_perturbed  = Sm_lps4_0_1_0::solve_primal_perturbed;
  i.solve_dual_perturbed    = Sm_lps4_0_1_0::solve_dual_perturbed;
  i.get_dual_ray            = NULL;
  i.get_primal_ray	        = NULL;
  i.set_primal_phase1	      = NULL;
  i.set_dual_phase1	        = NULL;
  i.restore_primal          = Sm_lps4_0_1_0::restore_primal;
  i.restore_dual            = Sm_lps4_0_1_0::restore_dual;
  i.restore_primal_phase1   = NULL;
  i.restore_dual_phase1     = NULL;
  i.read_lp                 = Sm_lps4_0_1_0::read_lp;
  i.set_lp                  = Sm_lps4_0_1_0::set_lp;
  i.set_module_options      = Sm_lps4_0_1_0::set_module_options;
  i.get_accuracy            = Sm_lps4_0_1_0::get_accuracy;
} /* get_func_pointers */
