/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file declares the solver module for lp_solve 4.0.1.0, Sm_lps4_0_1_0.
 *
 * @file  Sm_lps4_0_1_0.h
 * @brief Declaration of lp_solve 4.0.1.0 solver module
 *
 * @author Christian Keil
 *  
 * $Id: Sm_lps4_0_1_0.h 533 2008-06-12 20:08:12Z keil $
 */

#ifndef SM_LPS4_0_1_0_H
#define SM_LPS4_0_1_0_H

#include <lurupa/Solver_module.h>
#include <lpkit.h>
/* lp_solve's definition of UNKNOWN clashes with libtool */
#undef UNKNOWN

/**
 * @class Lp_lp_solve
 *
 * @brief Information about lp_solve's problem
 *
 * This class stores additional information about lp_solve's representation of an %lp.
 */
class Lp_lp_solve : public Lp_solver {
public:
  Lp_lp_solve();
  ~Lp_lp_solve();

  lprec *lp;       /**< @brief Reference to lp_solve's structure representing the %lp. */
  int *mp_eq_con;  /**< @brief Map from equation to constraint */
  int *mp_le_con;  /**< @brief Map from less-equal to constraint */
};

/** @brief Get function pointers */
extern "C" void get_func_pointers(Solver_module_interface &i);

/**
 * @class Sm_lps4_0_1_0 Sm_lps4_0_1_0.h "Sm_lps4_0_1_0.h"
 *
 * @brief Solver module for lp_solve 4.0.1.0
 */
class Sm_lps4_0_1_0 : public Solver_module {
public:
  /** @brief Get module version string */
  static const char *get_version();

  /** @brief Print module version */
  static void print_version();
  /** @brief Print brief module version */
  static void print_brief_version();

  /** @brief Initialize the module */
  static bool init(Report *report);

  /** @brief Get accuracy used by lp_solve */
  static double get_accuracy();

  /** @brief Print supported options */
  static void print_options();

  /** @brief Set module options */
  static bool set_module_options(Lp *lp, int argc, char *argv[]);

  /** @brief Read lp */
  static bool read_lp(Lp *lp, FILE *in, const double relative_interval_radius, double &eta);

  /** @brief Set lp */
  static bool set_lp(Lp *lp, const double relative_interval_radius, double &eta);

  /** @brief Compute solution of unperturbed lp */
  static bool solve_original(Lp *lp, double &optimal_value);

  /** @brief Compute solution of primal perturbed lp */
  static bool solve_primal_perturbed(const Primal_deflation &deflation, Lp *lp, 
                                     const int iteration);

  /** @brief Compute solution of dual perturbed lp */
  static bool solve_dual_perturbed(const VECTOR &deflation_c, Lp *lp, const int iteration);

  /** @brief Restore lp after primal perturbation */
  static void restore_primal(Lp *lp);

  /** @brief Restore lp after dual perturbation */
  static void restore_dual(Lp *lp);

private:
  /** @brief Callback function for lp_solve's log messages */
  static void WINAPI lps_log(void *lp, void *userhandle, char *buf);

  /**
   * @name Lp transforming routines
   * 
   * The following routines read the lp from the data structures of lp_solve
   * 4.0.1.0 and transform it into Lurupa's representation. */
  /*@{*/
  /** @brief Read general data from lp_solve's structures */
  static void read_general_data(Lp *lp);

  /** @brief Build mapping between lp_solve's and Lurupa's constraint structures */
  static bool build_constraint_maps(int &c_eq, int &c_le, const int rows, const int cols, Lp *lp);

  /** @brief Find free variables */
  static void find_free_variables(const int vars, Lp *lp);

  /** @brief Resize Lurupa's data structures according to the lp dimensions */
  static void resize_lp(Lp *lp, const int cols, const int c_eq, const int c_le);

  /** @brief Read lp_solve's lp_mat structure */
  static void read_lp_mat(const int rows, Lp *lp);

  /** @brief Read right hand sides of the constraints */
  static void read_right_hand_sides(Lp *lp);

  /** @brief Read simple bounds */
  static void read_simple_bounds(Lp *lp);

  /** @brief Adjust eta to the lp */
  static void adjust_eta(double &eta, const Lp *lp);

  /** @brief Inflate lp parameters */
  static void inflate_lp(Lp *lp, const double r);

  /** @brief Transform lp from lp_solve to %Lurupa */
  static bool transform_lp(Lp *lp, const double relative_interval_radius, double &eta);
  /*@}*/


  /** @brief Postprocess lp */
  static void postprocess(Lp *lp, const int status);

  /** @brief Solve lp with lp_solve */
  static bool solve_lp(Lp *lp);

  /** @brief Read primal solution */
  static bool read_primals(Lp *lp, const int cols);

  /** @brief Read dual solution */
  static bool read_duals(Lp *lp);

  /** @brief Primal perturb lp */
  static void primal_perturb(const Primal_deflation &d, const Lp *lp);

  /** @brief Dual perturb lp */
  static void dual_perturb(const VECTOR &d_c, const Lp *lp);

  Sm_lps4_0_1_0();
  ~Sm_lps4_0_1_0();
};

#endif /* ifdef SM_LPS4_0_1_0_H */
