/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This file defines the solver module for lp_solve 5.5, Sm_lps5_5.
 *
 * @file  Sm_lps5_5.cpp
 * @brief Definition of lp_solve 5.5 solver module
 *
 * @author Christian Keil
 *
 * $Id: Sm_lps5_5.cpp 533 2008-06-12 20:08:12Z keil $
 */
#include "Sm_lps5_5.h"
#include <lurupa/Lp.h>
#include <BIAS/Bias0.h>
#include <errno.h>
#include <string.h>

/** @brief Version string */
static const char *version = "Lp_solve 5.5 module 1.1";

Report *Solver_module::preport = NULL;
Write_mps Sm_lps5_5::write_mps = wm_none;
bool   Sm_lps5_5::freset_bas   = false;
long   Sm_lps5_5::timeout      = 0;
int    Sm_lps5_5::verbosity    = 3;
MYBOOL Sm_lps5_5::trace        = FALSE;

Lp_lp_solve::Lp_lp_solve()
{
  mp_eq_con = NULL;
  mp_le_con = NULL;
}

Lp_lp_solve::~Lp_lp_solve()
{
  free_lp(&lp);
  delete [] mp_eq_con;
  delete [] mp_le_con;
}

/**
 * Get the version information of the solver module.
 */
const char *
Sm_lps5_5::get_version()
{
  return version;
} /* Sm_lps5_5::get_version */

void
Sm_lps5_5::print_version()
{
  int major, minor, release, build;
  lp_solve_version(&major, &minor, &release, &build);
  printf("solver module\n"
         "  version : %s\n"
         "  solver parameters\n"
         "    version : %d.%d.%d.%d\n"
         "    eps     : %g\n", 
         version, major, minor, release, build, DEF_EPSDUAL);
} /* Sm_lps5_5::print_version */

void
Sm_lps5_5::print_brief_version()
{
  int major, minor, release, build;
  lp_solve_version(&major, &minor, &release, &build);
  printf("solver module       : %s, solver version = %d.%d.%d.%d, solver eps = %g\n",
         version, major, minor, release, build, DEF_EPSDUAL);
} /* Sm_lps5_5::print_brief_version */

/**
 * Set #preport to \p *report. Check the version of lp_solve.
 *
 * \param[in] report  the #Report instance to use
 *
 * \retval true
 */
bool
Sm_lps5_5::init(Report *report)
{
  int major, minor, release, build;

  preport = report;
  preport->print(3, "Using %s", version);

  lp_solve_version(&major, &minor, &release, &build);
  if ((major != 5) || (minor != 5) || (release != 0) || (build < 6) || (build > 11)) {
    preport->print(2, "Lp_solve version differs, expected: 5.5.0.[6-11], found: %d.%d.%d.%d",
                   major, minor, release, build);
  }

  return true;
} /* Sm_lps5_5::init */

/**
 * Print the command line options supported by the solver module.
 */
void
Sm_lps5_5::print_options()
{
  printf("Options supported by %s:\n", version);
  printf("  -sm,resetbas        Try to re-solve lps in the case of numerical\n"
         "                      failure with a basis reset to all slacks.\n");
  printf("  -sm,s               Use lp_solve's default scaling\n"
         "                      (Numerical range-based scaling).\n");
  printf("  -sm,t               Trace pivot selection\n");
  printf("  -sm,timeout,i       Set solver timeout in seconds.\n");
  printf("  -sm,vn              Set solver verbosity:\n"
         "                        v0: NEUTRAL\n"
         "                        v1: CRITICAL\n"
         "                        v2: SEVERE\n"
         "                        v3: IMPORTANT (default)\n"
         "                        v4: NORMAL\n"
         "                        v5: DETAILED\n"
         "                        v6: FULL\n");
  printf("  -sm,wmps,s          Save intermediate problems in mps format:\n"
         "                        s = all : Save all problems\n"
         "                            ask : Ask before saving\n"
         "                            none: Save no problems\n");
} /* Sm_lps5_5::print_options */

void
Sm_lps5_5::write_lp(const char *sz_description, lprec *lp_lps, char *sz_file)
{
  if (wm_none == write_mps) return;

  bool f = true;
  if (wm_ask == write_mps) {
    printf("Save this LP to disk (%s):\n%s ? [y/n] ", sz_file, sz_description);
    char sz[2];
    scanf("%1s", sz);
    if (('n' == *sz) || ('N' == *sz)) f = false;
  }

  if (f) ::write_mps(lp_lps, sz_file);
  preport->print(5, "Wrote %s", sz_file);
} /* Sm_lps5_5::write_lp */

/**
 * Read general lp properties from lp_solve's structures and store them in @p lp. This includes the
 * value representing infinity, the direction of optimization, and the name of the lp.
 *
 * @param[out] lp  stores the read data
 */
void
Sm_lps5_5::read_general_data(Lp *lp)
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;
  lp->infinite = get_infinite(lp_lps);
  lp->maximize = is_maxim(lp_lps);
  lp->name = strdup(get_lp_name(lp_lps));
} /* Sm_lps5_5::read_general_data */

/**
 * Build a mapping from Lurupa's equality and inequality constraints to lp_solve's constraints,
 * which is storing all constraints in one big matrix. Store the mapping in @p lp->module->mp_eq_con
 * and @p lp->module->mp_le_con. Check if the number of constraints is less than the number of
 * variables.
 *
 * @param[out] c_eq  number of equations in lp
 * @param[out] c_le  number of inequalities in lp
 * @param[in]  rows  number of rows of constraint matrix
 * @param[in]  cols  number of columns of constraint matrix
 * @param[in]  lp    the %lp
 *
 * @retval true
 * if number of constraints is less than number of variables,
 * @retval false
 * otherwise
 */
bool
Sm_lps5_5::build_constraint_maps(int &c_eq, int &c_le, const int rows, const int cols, Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* count equations and inequalities */
  for (int i = 1; i <= rows; ++i) {
    if (get_constr_type(lp_lps->lp, i) == EQ) {
      ++c_eq;
    } else {
      ++c_le;
    }
  }

  if (c_eq > cols) {
    preport->print(1, "Sm_lps5_5::build_constraint_maps: model contains more "
                   "equations (%d) than variables (%d)", c_eq, cols);
    return false;
  }

  /* build maps */
  lp_lps->mp_eq_con = new int[c_eq];
  lp_lps->mp_le_con = new int[c_le];
  c_eq = -1; c_le = -1;
  for (int i = 1; i <= rows; ++i) {
    if (get_constr_type(lp_lps->lp, i) == EQ) {
      lp_lps->mp_eq_con[++c_eq] = i;
    } else {
      lp_lps->mp_le_con[++c_le] = i;
    }
  }

  ++c_le;
  ++c_eq;

  return true;
} /* Sm_lps5_5::build_constraint_maps */

/**
 * Find free variables in the lp. Store the indices of the free variables in @p lp->free_variables
 * and their number in @p lp->free_variables_size.
 *
 * @param[in]  vars       number of variables in lp
 * @param[out] lp         stores number and indices of the free variables
 * @param[out] c_inf_bnd  count of variables with infinite bounds
 */
void
Sm_lps5_5::find_free_variables(const int vars, Lp *lp, int *c_inf_bnd)
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  /* count free variables */
  for (int i = 1; i <= vars; ++i) {
    if (is_infinite(lp_lps, get_lowbo(lp_lps, i))) {
      ++(*c_inf_bnd);
      if (is_infinite(lp_lps, get_upbo(lp_lps, i))) {
        ++lp->free_variables_size;
      }
    } else if (is_infinite(lp_lps, get_upbo(lp_lps, i))) {
      ++(*c_inf_bnd);
    }
  }
  lp->free_variables = new int[lp->free_variables_size];

  /* store indices */
  int var = 1, free_var = 0;
  while (free_var < lp->free_variables_size) {
    if (is_unbounded(lp_lps, var)) lp->free_variables[free_var++] = var;
    ++var;
  }
}

/**
 * Resize the vectors and matrices in Lurupa's data structure to the sizes of
 * the lp.
 *
 * @param[out] lp     Lurupa's structure with the vectors and matrices to be resized
 * @param[in]  cols   number of variables in the lp
 * @param[in]  c_eq   number of equality constraints in the lp
 * @param[in]  c_le   number of inequality constraints in the lp
 */
void
Sm_lps5_5::resize_lp(Lp *lp, const int cols, const int c_eq, const int c_le)
{
  Resize(lp->IA, c_le, cols);
  Resize(lp->IB, c_eq, cols);
  Resize(lp->ic, cols);
  Resize(lp->ia, c_le);
  Resize(lp->ib, c_eq);
  Resize(lp->xl, cols);
  Resize(lp->xu, cols);
  Resize(lp->ix, cols);
  Resize(lp->iy, c_le);
  Resize(lp->iz, c_eq);
}

/**
 * Read lp_solve's @c lp_mat structure. This stores the left hand sides of the
 * constraints together with the objective function. Split the read coefficients
 * into equality and inequality constraints, transform inequality constraints to
 * all less equal, and the direction of optimization to minimize.
 *
 * @param[in]  rows   number of rows in constraint matrix
 * @param[out] lp     stores the constraints and objective function
 *
 * @retval true
 * if reading @c lp_mat succeeds,
 * @retval false
 * otherwise, lp_solve's @c get_column returns false
 */
bool
Sm_lps5_5::read_lp_mat(const int rows, Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  REAL *column = new REAL[rows + 1];
  for (int j = 1; j <= ColDimension(lp->IA); ++j) {
    /* get_column returns the parameters of the _original_ lp, so we need to
     * transform into our representation */
    if (get_column(lp_lps->lp, j, column) == FALSE) {
      return false;
    }

    /* maximizing linear programs become minimizing ones: multiply objective
     * function coefficients by -1 */
    lp->ic(j) = is_maxim(lp_lps->lp) ? -Hull(column[0]) : Hull(column[0]);

    for (int i = 1; i <= RowDimension(lp->IA); ++i) {
      /* >= constraints become <= constraints: multiply coefficients by -1 */
      if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
        lp->IA(i, j) =  Hull(column[lp_lps->mp_le_con[i - 1]]);
      } else {
        lp->IA(i, j) = -Hull(column[lp_lps->mp_le_con[i - 1]]);
      }
    }

    for (int i = 1; i <= RowDimension(lp->IB); ++i) {
      lp->IB(i, j) = Hull(column[lp_lps->mp_eq_con[i - 1]]);
    }
  }
  delete [] column;

  return true;
} /* Sm_lps5_5::read_lp_mat */

/**
 * Read the right hand sides of the constraints. Split the read right hand sides
 * into equality and inequality constraints, transform inequality constraints to
 * all less equal. Store the values in @p lp->ia and @p lp->ib.
 *
 * @param[out] lp  stores the right hand sides
 */
void
Sm_lps5_5::read_right_hand_sides(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  for (int i = 1; i <= Dimension(lp->ia); ++i) {
    /* transform >= constraints to <= */
    if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
      lp->ia(i) =  Hull(get_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1]));
    } else {
      lp->ia(i) = -Hull(get_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1]));
    }
  }

  for (int i = 1; i <= Dimension(lp->ib); ++i) {
    lp->ib(i) = Hull(get_rh(lp_lps->lp, lp_lps->mp_eq_con[i - 1]));
  }
} /* Sm_lps5_5::read_right_hand_sides */

/**
 * Read the simple bounds of the variables and store them in @p lp->xl and
 * @p lp->xu.
 *
 * @param[out] lp  stores the simple bounds
 */
void
Sm_lps5_5::read_simple_bounds(Lp *lp)
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  for (int i = 1; i <= Dimension(lp->xl); ++i) {
    lp->xl(i) = get_lowbo(lp_lps, i);
    lp->xu(i) = get_upbo(lp_lps, i);

    if (lp->xl(i) != lp->xu(i)) ++lp->non_fixed_vars;
  }
} /* Sm_lps5_5::read_simple_bounds */

/**
 * Adjust the algorithm parameter @p eta to be problem dependent.  It becomes
 * the maximum of its previous value and \f$10^{-20}\f$ times the
 * maximum norm of the right hand sides of the inequality constraints,
 * \f[
 *   \eta = \max\{\eta, 10^{-20} \|a\|_{\infty}\}.
 * \f]
 *
 * @param[out] eta  parameter to be adjusted
 * @param[in]  lp   lp @p eta is adjusted to
 */
void
Sm_lps5_5::adjust_eta(double &eta, const Lp *lp)
{
  /* <lp->ia> is a point vector here */
  VECTOR v = 1e-20 * Abs(lp->ia);

  for (int i = 1; i <= Dimension(v); ++i) {
    if (v(i) > eta) eta = v(i);
  }

  preport->print(5, "Adjusting eta to %.16e", eta);
} /* Sm_lps5_5::adjust_eta */

/**
 * Inflate all lp parameters to intervals with relative radius @p r,
 * @f[
 *   x \quad \rightarrow \quad \mathbf{x} \triangleq x \cdot [1-r, 1+r]
 * @f]
 *
 * @param[out] lp  lp to be inflated
 * @param[in]  r   relative interval radius
 */
void
Sm_lps5_5::inflate_lp(Lp *lp, const double r)
{
  lp->IA = Hull((1 - r) * Inf(lp->IA), (1 + r) * Sup(lp->IA));
  lp->IB = Hull((1 - r) * Inf(lp->IB), (1 + r) * Sup(lp->IB));
  lp->ia = Hull((1 - r) * Inf(lp->ia), (1 + r) * Sup(lp->ia));
  lp->ib = Hull((1 - r) * Inf(lp->ib), (1 + r) * Sup(lp->ib));
  lp->ic = Hull((1 - r) * Inf(lp->ic), (1 + r) * Sup(lp->ic));
} /* Sm_lps5_5::inflate_lp */

/**
 * Callback function for lp_solve's log functionality. Print the supplied log
 * message marked as coming from lp_solve.
 */
extern "C" void __WINAPI
Sm_lps5_5::lps_log(lprec *lp_lps, void *userhandle, char *buf)
{
  /* Remove \n's from the start of buf */
  while (*buf == '\n') ++buf;
  /* Remove the \n from the end of buf */
  buf[strlen(buf) - 1] = '\0';
  preport->print(0, "[lps55] %s", buf);
}

/**
 * Transform lp from the data structure of lp_solve 5.5 into the representation used in Lurupa.
 * Adjust algorithm parameters @p eta according to the lp, and inflate the lp to an interval valued
 * one if @p relative_interval_radius is greater than 0.
 *
 * Write the (midpoint) problem with #preport if Report#write_vm is set.
 *
 * @param[out]  lp                        receives the transformed data
 * @param[in]   relative_interval_radius  relative interval radius to inflate lp parameters to
 * @param[out]  eta                       gets adjusted to the model
 * 
 * @retval true
 * if the transformation succeeds,
 * @retval false
 * otherwise, @p lp->module->lp is not set or @c build_constraint_maps returns false
 *
 * @sa
 * #adjust_eta,
 * #build_constraint_maps,
 * #find_free_variables,
 * #inflate_lp,
 * #read_general_data,
 * #read_lp_mat,
 * #read_right_hand_sides,
 * #read_simple_bounds,
 * #resize_lp
 */
bool
Sm_lps5_5::transform_lp(Lp *lp, const double relative_interval_radius, double &eta)
{
  preport->print(3, "Reading lp from lp_solve 5.5");

  if (static_cast<Lp_lp_solve *>(lp->module)->lp == NULL) {
    preport->print(1, "Sm_lps5_5::transform_lp: model not set");
    return false;
  }

  /* read general data */
  read_general_data(lp);

  int c_eq = 0, c_le = 0;
  /* we work on the original not the presolved model, presolving is not
   * rigorous - at this time presolving hasn't been done so get_Nrows /
   * get_Ncolumns should return the same */
  int rows = get_Norig_rows(static_cast<Lp_lp_solve *>(lp->module)->lp);
  int cols = get_Norig_columns(static_cast<Lp_lp_solve *>(lp->module)->lp);

  /* build constraint maps and count equations and inequalities */
  if (!build_constraint_maps(c_eq, c_le, rows, cols, lp)) {
    return false;
  }

  /* collect free variables if present in the lp */
  int c_inf_bnd = 0;
  find_free_variables(cols, lp, &c_inf_bnd);

  if (c_inf_bnd > 0) {
    preport->print(3, "variables  : %d total\t %d with infinite bounds\t %d free",
                   cols, c_inf_bnd, lp->free_variables_size);
  } else {
    preport->print(3, "variables  : %d total\t only finite bounds",
                   cols);
  }

  preport->print(3, "constraints: %d total\t %d equations\t %d inequalities",
                 rows, c_eq, c_le);

  /* resize the data structures */
  resize_lp(lp, cols, c_eq, c_le);

  /* read the matrix and the objective function */
  read_lp_mat(rows, lp);

  /* read the right hand sides of the constraints */
  read_right_hand_sides(lp);

  /* read the simple bounds of the variables */
  read_simple_bounds(lp);

  /* make eta problem dependent */
  adjust_eta(eta, lp);

  if (preport->get_write_vm()) {
    /* write the model to a file */
    preport->write_matrix(lp->name, Mid(lp->IA), "A");
    preport->write_vector(lp->name, Mid(lp->ia), "a");
    preport->write_matrix(lp->name, Mid(lp->IB), "B");
    preport->write_vector(lp->name, Mid(lp->ib), "b");
    preport->write_vector(lp->name, Mid(lp->ic), "c");
    preport->write_vector(lp->name, lp->xl, "xl");
    preport->write_vector(lp->name, lp->xu, "xu");
  }

  if (relative_interval_radius > 0) {
    /* inflate the lp to an interval one */
    inflate_lp(lp, relative_interval_radius);
  }

  return true;
}

/**
 * Read an lp from a file into lp_solve and @p lp. Inflate the lp to an interval valued one if @p
 * relative_interval_radius is greater than 0, and adjust @p eta according to the lp's parameters.
 *
 * @param[out] lp                        receives the lp
 * @param[in]  in                        file pointer to %lp
 * @param[in]  relative_interval_radius  interval radius to inflate lp parameters to
 * @param[out] eta                       algorithm parameter to be adjusted
 *
 * @retval true
 * if reading and transforming the lp succeeds,
 * @retval false
 * otherwise, reading the lp fails or @c transform_lp returns false
 *
 * @sa
 * #transform_lp
 */
bool
Sm_lps5_5::read_lp(Lp *lp, FILE *in, const double relative_interval_radius, double &eta)
{
  bool success = true;
  lp->module = new Lp_lp_solve;
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  BiasRoundNear();
  if (NULL == (lp_lps->lp = read_mps(in, IMPORTANT))) {
    success = false;
  }

  if (success) {
    /* Print log messages ourself, this way we can prefix them and thus
     * distinguish from ours. */
    set_outputfile(lp_lps->lp, "");
    put_logfunc(lp_lps->lp, lps_log, NULL);
    /* Set default verbosity level to 3 */
    set_verbose(lp_lps->lp, 3);
    set_scaling(lp_lps->lp, SCALE_NONE);

    success = transform_lp(lp, relative_interval_radius, eta);
  }

  return success;
} /* Sm_lps5_5::read_lp */

bool
Sm_lps5_5::set_lp(Lp *lp, const double relative_interval_radius, double &eta)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  return transform_lp(lp, relative_interval_radius, eta);
} /* Sm_lps5_5::set_lp */

/**
 * Solve an lp with lp_solve. Store lp_solve's status in @p lp->feasibility.
 *
 * @param[out] lp  receives lp_solve's status 
 *
 * @retval true
 * if lp_solve returns a known status code (i.e., OPTIMAL, INFEASIBLE,
 * UNBOUNDED, TIMEOUT, NUMFAILURE, DEGENERATE, or SUBOPTIMAL)
 * @retval false
 * otherwise
 */
bool
Sm_lps5_5::solve_lp(Lp *lp) 
{
  /* No call to preprocess postprocess to compute last feasible points in case
   * of infeasibility or unboundedness. Lp_solve has to be patched to compute
   * these at this time. Problem is with postsolving, which is done whether
   * presolve was called or not. This way we cannot run the simplex, compute
   * last feasible points and then complete the postprocessing and postsolving.
   * */
  BiasRoundNear();

  bool fretry = false;
  bool success = true;

  do {
    int code = solve(static_cast<Lp_lp_solve *>(lp->module)->lp);

    switch(code) {
    case OPTIMAL:
      lp->feasibility = ss_feasible; 
      preport->print(5, "OPTIMAL, objective = %.16e",
                     get_objective(static_cast<Lp_lp_solve *>(lp->module)->lp));
      fretry = false;
      break;
    case INFEASIBLE:
      lp->feasibility = ss_infeasible; 
      preport->print(5, "INFEASIBLE");
      fretry = false;
      break;
    case UNBOUNDED:
      lp->feasibility = ss_unbounded;
      preport->print(5, "UNBOUNDED");
      fretry = false;
      break;
    case TIMEOUT:
      lp->feasibility = ss_timeout;
      preport->print(5, "TIMEOUT");
      fretry = false;
      break;
    case NUMFAILURE:
      lp->feasibility = ss_failure;
      preport->print(5, "NUMFAILURE");
      if (!fretry && freset_bas) {
        preport->print(4, "Resolving with all slack basis");
        default_basis(static_cast<Lp_lp_solve *>(lp->module)->lp);
        fretry = true;
      } else {
        fretry = false;
      }
      break;
    case DEGENERATE:
      lp->feasibility = ss_failure;
      preport->print(5, "DEGENERATE");
      fretry = false;
      break;
    case SUBOPTIMAL:
      lp->feasibility = ss_feasible;
      preport->print(5, "SUBOPTIMAL");
      fretry = false;
      break;
    default:
      lp->feasibility = ss_failure;
      preport->print(1, "Sm_lps5_5::solve_lp: unknown code %d", code);
      fretry = false;
      success = false;
    }
  } while (fretry);

  return success;
} /* Sm_lps5_5::solve_lp */

/**
 * Read the primal solution from lp_solve and store it in @p lp->ix.
 *
 * @param[out] lp  receives the primal solution
 *
 * @retval true
 * if reading the primal solution succeeds,
 * @retval false
 * otherwise, calling lp_solve's @c get_variables returns false
 */
bool
Sm_lps5_5::read_primals(Lp *lp)
{
  REAL *solution;
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);
  if (!get_ptr_primal_solution(lp_lps->lp, &solution)) {
    preport->print(1, "Sm_lps5_5::read_primals: calling get_ptr_primal_solution");
    return false;
  }

  /* get_ptr_primal_solution returns the objective value and constraint values
   * also:
   * +----------------------------------------------------------+
   * |   |                       |                              |
   * +----------------------------------------------------------+
   *   0   1     -     get_Nrows   get_Nrows+1     -         
   *  obj    constraints                        primals
   *  */
  int base = get_Nrows(lp_lps->lp);
  for (int i = 1; i <= Dimension(lp->ix); ++i)
    lp->ix(i) =  Hull(solution[base + i]);

  /* Don't delete solution, the memory belongs to lp_solve */

#ifndef NDEBUG
  /* What kind of output do we get in case of infeasible/unbounded linear
   * programs? */
  preport->print(6, "lp_solve's c'x = %.16e", get_objective(lp_lps->lp));
  preport->print(6, "read c'x       = %.16e", Inf(lp->ic * lp->ix));
  /* Set rounding mode to nearest - we only need this if we actually print these
   * debug infos. */
  BiasRoundNear();
#endif

  return true;
} /* Sm_lps5_5::read_primals */

/**
 * Read the dual solution from lp_solve and store it in @p lp->iy and @p lp->iz. Lp_solve's
 * convention to determine the sign of the duals is, that the costs must be a linear combination of
 * the gradients of the constraints. 
 *
 * The duals returned by lp_solve, however, are the duals for the original lp,
 * that is the duals for greater equal constraints are multiplied by \f$-1\f$,
 * and all duals are again multiplied by \f$-1\f$ if the lp is a maximizing one.
 * As %Lurupa stores all linear programs as minimizing and all inequality
 * constraints as less equal undo these multiplications.
 *
 * @param[out] lp  receives the dual solution
 *
 * @retval true
 * if reading the dual solution succeeded,
 * @retval false
 * otherwise, calling lp_solve's @c get_sensitivity_rhs returns false
 */
bool
Sm_lps5_5::read_duals(Lp *lp)
{
  REAL *duals;
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  if (!get_ptr_dual_solution(lp_lps->lp, &duals)) {
    preport->print(1, "Sm_lps5_5::read_duals: calling get_ptr_dual_solution");
    return false;
  }

  /* get_ptr_dual_solution returns a 1-based array!!! */
  /* if we have a maximizing lp all returned duals have to be multiplied by -1 */
  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->iy); ++i) {
      /* if the constraint is a <= one no additional multiplication is necessary */
      if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
        lp->iy(i) = -Hull(duals[lp_lps->mp_le_con[i - 1]]);
      } else {
        lp->iy(i) =  Hull(duals[lp_lps->mp_le_con[i - 1]]);
      }
    }

    for (int i = 1; i <= Dimension(lp->iz); ++i) {
      lp->iz(i) = -Hull(duals[lp_lps->mp_eq_con[i - 1]]);
    }
  } else {
    for (int i = 1; i <= Dimension(lp->iy); ++i) {
      /* if the constraint is a <= one no multiplication is necessary */
      if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
        lp->iy(i) =  Hull(duals[lp_lps->mp_le_con[i - 1]]);
      } else {
        lp->iy(i) = -Hull(duals[lp_lps->mp_le_con[i - 1]]);
      }
    }

    for (int i = 1; i <= Dimension(lp->iz); ++i) {
      lp->iz(i) = Hull(duals[lp_lps->mp_eq_con[i - 1]]);
    }
  }

  /* Don't delete duals, the memory belongs to lp_solve */

  return true;
} /* Sm_lps5_5::read_duals */

/**
 * Solve the original unperturbed lp with lp_solve and read the primal and dual
 * solution.
 *
 * @param[out] lp             receives the solver status and the solution
 * @param[out] optimal_value  receives the optimal value
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp, @c read_primals, or @c read_duals returns false
 *
 * @sa
 * #read_duals,
 * #read_primals,
 * #solve_lp
 */
bool
Sm_lps5_5::solve_original(Lp *lp, double &optimal_value)
{
  preport->print(3, "Solving original lp");

  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  bool success = solve_lp(lp);

  if (success) {
    optimal_value = get_objective(lp_lps);

    /* read the solution and the value of the duals */
    if (!read_primals(lp) || !read_duals(lp)) {
      success = false;
    }
  }

  return success;
} /* Sm_lps5_5::solve_original */

/**
 * Perturb the lp in the right hand sides of the inequalities and the simple bounds.
 *
 * @param[in]  d   deflation parameters
 * @param[out] lp  contains the base for the deflation
 */
void
Sm_lps5_5::primal_perturb(const Primal_deflation &d, const Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* perturb the right hand sides */
  for (int i = 1; i <= Dimension(d.a); ++i) {
    /* If the constraint is a "less or equal" one we can directly use the
     * perturbed right hand side. Otherwise we have to multiply it by -1 as
     * lp_solve expects the right hand side for a "greater or equal" constraint
     * in this case. 
     * We could cache this information in a sign vector and multiply with it to
     * get rid of the if's. If this would really be a bottleneck =). */
    if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1],   Mid(lp->ia(i)) - d.a(i));
    } else {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1], -(Mid(lp->ia(i)) - d.a(i)));
    }
  }

  /* perturb the simple bounds */
  for (int i = 1; i <= Dimension(d.xl); ++i) {
    if (d.xl(i) != 0) set_lowbo(lp_lps->lp, i, lp->xl(i) + d.xl(i));
    if (d.xu(i) != 0)  set_upbo(lp_lps->lp, i, lp->xu(i) - d.xu(i));
  }
} /* Sm_lps5_5::primal_perturb */

/**
 * Perturb the lp and solve the new one. Read the primal solution and store it
 * in @p lp->ix. Save the perturbed lp to <tt>lpname-ubii.mps</tt> according to setting of
 * #write_mps. Substitute the name of the lp for <tt>lpname</tt> and @p iteration for <tt>ii</tt>.
 *
 * @param[in]  d          deflation parameters
 * @param[out] lp         receives the solver status and the primal solution
 * @param[in]  iteration  number of the current iteration
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp or @c read_primals returns false
 *
 * @sa
 * #read_primals,
 * #solve_lp
 */
bool
Sm_lps5_5::solve_primal_perturbed(const Primal_deflation &d, Lp *lp, const int iteration)
{
  preport->print(3, "Solving primal perturbed lp");

  primal_perturb(d, lp);

  bool success = solve_lp(lp);

  /* Read the primal solution if the problem is feasible. */
  if (success && (lp->feasibility == ss_feasible)) {
    success = read_primals(lp);
  }

  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;
  char *sz = new char[strlen(lp->name) + 9 + 1];
  sprintf(sz, "%s-ub%02d.mps", lp->name, iteration);
  write_lp("Primal perturbed problem", lp_lps, sz);
	delete [] sz;

  return success;
}

/**
 * Perturb the lp in the cost vector.
 *
 * @param[in]  d_c  deflation parameters
 * @param[out] lp   contains the base for the deflation
 */
void
Sm_lps5_5::dual_perturb(const VECTOR &d_c, const Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  /* if the lp is a maximizing one we need to multiply our minimizing cost 
   * vector by -1 */
  if (lp->maximize) {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      if (0 != d_c(i)) {
        set_obj(lp_lps->lp, i, -(Mid(lp->ic(i)) + d_c(i)));
      }
    }
  } else {
    for (int i = 1; i <= Dimension(lp->ic); ++i) {
      if (0 != d_c(i)) {
        set_obj(lp_lps->lp, i, Mid(lp->ic(i)) + d_c(i));
      }
    }
  }
}

/**
 * Perturb the lp and solve the new one. Read the dual solution from lp_solve
 * and store it in @p lp->iy and @p lp->iz. Save the perturbed lp as <tt>lpnamel-lbii.mps</tt>
 * according to the setting of #write_mps. Substitute the name of the lp for <tt>lpname</tt> and @p
 * iteration for <tt>ii</tt>.
 *
 * @param[in]  d_c        deflation parameter
 * @param[out] lp         receives the solver status and the dual solution
 * @param[in]  iteration  number of the current iteration
 *
 * @retval true
 * if @c solve_lp returns true and the solution is read
 * @retval false
 * otherwise, @c solve_lp or @c read_duals returns false
 *
 * @sa
 * #read_duals,
 * #solve_lp
 */
bool
Sm_lps5_5::solve_dual_perturbed(const VECTOR &d_c, Lp *lp, const int iteration)
{
  preport->print(3, "Solving dual perturbed lp");

  dual_perturb(d_c, lp);

  bool success = solve_lp(lp);

  /* Read the dual solution if the problem is feasible. */
  if (success && (lp->feasibility == ss_feasible)) {
    success = read_duals(lp);
  }

  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;
  char *sz = new char[strlen(lp->name) + 9 + 1];
  sprintf(sz, "%s-lb%02d.mps", lp->name, iteration);
  write_lp("Dual perturbed problem", lp_lps, sz);
	delete [] sz;

  return success;
} /* Sm_lps5_5::solve_dual_perturbed */

bool
Sm_lps5_5::get_dual_ray(Lp *lp, int &iyzray)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);
  int *coltarget;
  REAL d[lp_lps->lp->sum + 1];

  Clear(lp->iy);
  Clear(lp->iz);
  coltarget = (int *) mempool_obtainVector(lp_lps->lp->workarrays, lp_lps->lp->columns+1, sizeof(*coltarget));
  get_colIndexA(lp_lps->lp, SCAN_USERVARS+USE_NONBASICVARS, coltarget, FALSE);

#ifndef NDEBUG
  if (1 == iyzray) {
    if (lp_lps->lp->rows <= 100) {
      set_outputfile(lp_lps->lp, NULL);
      print_tableau(lp_lps->lp);
      set_outputfile(lp_lps->lp, "");
    }
    printf("var_basic = {%d", lp_lps->lp->var_basic[1]);
    for (int i = 2; i <= lp_lps->lp->rows; ++i) {
      printf(",%d", lp_lps->lp->var_basic[i]);
    }
    printf("}\n");
    printf("rhs = {%g", lp_lps->lp->rhs[1]);
    for (int i = 2; i <= lp_lps->lp->rows; ++i) {
      printf(",%g", lp_lps->lp->rhs[i]);
    }
    printf("}\n");
  }
#endif

  for (int i = iyzray; i <= lp_lps->lp->rows; ++i) {
    if ((lp_lps->lp->rhs[i] < -lp_lps->lp->epsdual) 
        && (lp_lps->lp->var_basic[i] <= lp_lps->lp->rows)
        && ((LE == get_constr_type(lp_lps->lp, lp_lps->lp->var_basic[i])) 
            || (GE == get_constr_type(lp_lps->lp, lp_lps->lp->var_basic[i])))) {
      /* We have an infeasibility in an inequality constraint */
      preport->print(6, "Infeasibility in row %d, type = %s, rhs = %.16e\n",
                     i, get_constr_type(lp_lps->lp, lp_lps->lp->var_basic[i]) == LE ? "LE" : "GE",
                     lp_lps->lp->rhs[i]);
      /* Compute the corresponding tableau row */
      bsolve(lp_lps->lp, i, d, NULL, lp_lps->lp->epsmachine*DOUBLEROUND, 1.0);
      prod_xA(lp_lps->lp, coltarget, d, NULL, lp_lps->lp->epsmachine, 1.0, d, NULL, MAT_ROUNDDEFAULT);

#ifndef NDEBUG
      /* Print primal row with variable types */
      for (int j = 1; j <= lp_lps->lp->sum + 1; ++j) {
        /* Dual variable types Less-equal, Greater-equal, Equality constraint */
        char rgch[] = " LGE";
        if (d[j] == 0) continue;
        if (lp_lps->lp->var_basic[i] == j) {
          printf("(R,%d%c,%g)", j, 
                 j <= lp_lps->lp->rows ? rgch[get_constr_type(lp_lps->lp, j)] : 'V',
                 lp_lps->lp->rhs[i]);
        } else if (!lp_lps->lp->is_basic[j]) {
          printf("(%c,%d%c,%g)", lp_lps->lp->is_lower[j] ? 'L' : 'U', j,
                 j <= lp_lps->lp->rows ? rgch[get_constr_type(lp_lps->lp, j)] : 'V', d[j]);
        }
      }
      printf("\n");
#endif

      /* Construct improving ray */
      for (int j = 1; j <= Dimension(lp->iy); ++j) {
        if (lp_lps->lp->var_basic[i] == lp_lps->mp_le_con[j-1]) {
          lp->iy(j) = -1;
        } else if (!lp_lps->lp->is_basic[lp_lps->mp_le_con[j-1]]) {
          lp->iy(j) = -d[lp_lps->mp_le_con[j-1]];
        }
      }
      for (int j = 1; j <= Dimension(lp->iz); ++j) {
        if (!lp_lps->lp->is_basic[lp_lps->mp_eq_con[j-1]]) {
          lp->iz(j) = -d[lp_lps->mp_eq_con[j-1]];
        }
      }

      iyzray = i;
      mempool_releaseVector(lp_lps->lp->workarrays, (char *) coltarget, FALSE);
      return true;
    } else if ((fabs(lp_lps->lp->rhs[i]) > lp_lps->lp->epsdual)
               && (lp_lps->lp->var_basic[i] <= lp_lps->lp->rows)
               && (EQ == get_constr_type(lp_lps->lp, lp_lps->lp->var_basic[i]))) {
      /* We have an infeasibility in an equality constraint */
      preport->print(6, "Infeasibility in row %d, type = E, rhs = %.16e\n",
                     i, lp_lps->lp->rhs[i]);
      /* Compute the corresponding tableau row */
      bsolve(lp_lps->lp, i, d, NULL, lp_lps->lp->epsmachine*DOUBLEROUND, 1.0);
      prod_xA(lp_lps->lp, coltarget, d, NULL, lp_lps->lp->epsmachine, 1.0, d, NULL, MAT_ROUNDDEFAULT);

#ifndef NDEBUG
      /* Print primal row with variable types */
      for (int j = 1; j <= lp_lps->lp->sum + 1; ++j) {
        /* Dual variable types Less-equal, Greater-equal, Equality constraint */
        char rgch[] = " LGE";
        if (d[j] == 0) continue;
        if (lp_lps->lp->var_basic[i] == j) {
          printf("(R,%d%c,%g)", j, 
                 j <= lp_lps->lp->rows ? rgch[get_constr_type(lp_lps->lp, j)] : 'V',
                 lp_lps->lp->rhs[i]);
        } else if (!lp_lps->lp->is_basic[j]) {
          printf("(%c,%d%c,%g)", lp_lps->lp->is_lower[j] ? 'L' : 'U', j,
                 j <= lp_lps->lp->rows ? rgch[get_constr_type(lp_lps->lp, j)] : 'V', d[j]);
        }
      }
      printf("\n");
#endif

      /* Construct improving ray */
      for (int j = 1; j <= Dimension(lp->iy); ++j) {
        if (!lp_lps->lp->is_basic[lp_lps->mp_le_con[j-1]]) {
          lp->iy(j) = d[lp_lps->mp_le_con[j-1]];
        }
      }
      for (int j = 1; j <= Dimension(lp->iz); ++j) {
        if (lp_lps->lp->var_basic[i] == lp_lps->mp_eq_con[j-1]) {
          lp->iz(j) = 1;
        } else if (!lp_lps->lp->is_basic[lp_lps->mp_eq_con[j-1]]) {
          lp->iz(j) = d[lp_lps->mp_eq_con[j-1]];
        }
      }
      if (lp_lps->lp->rhs[i] < 0) {
        lp->iy = -lp->iy;
        lp->iz = -lp->iz;
      }

      iyzray = i;
      mempool_releaseVector(lp_lps->lp->workarrays, (char *) coltarget, FALSE);
      return true;
    } else if ((lp_lps->lp->rhs[i] < -lp_lps->lp->epsdual) 
               && (lp_lps->lp->var_basic[i] > lp_lps->lp->rows)) {
      /* We have an infeasibility in a variable */
      preport->print(6, "Infeasibility in row %d, type = V, rhs = %.16e\n",
                     i, lp_lps->lp->rhs[i]);
      /* Compute the corresponding tableau row */
      bsolve(lp_lps->lp, i, d, NULL, lp_lps->lp->epsmachine*DOUBLEROUND, 1.0);
      prod_xA(lp_lps->lp, coltarget, d, NULL, lp_lps->lp->epsmachine, 1.0, d, NULL, MAT_ROUNDDEFAULT);

#ifndef NDEBUG
      /* Print primal row with variable types */
      for (int j = 1; j <= lp_lps->lp->sum + 1; ++j) {
        /* Dual variable types Less-equal, Greater-equal, Equality constraint */
        char rgch[] = " LGE";
        if (d[j] == 0) continue;
        if (lp_lps->lp->var_basic[i] == j) {
          printf("(R,%d%c,%g)", j, 
                 j <= lp_lps->lp->rows ? rgch[get_constr_type(lp_lps->lp, j)] : 'V',
                 lp_lps->lp->rhs[i]);
        } else if (!lp_lps->lp->is_basic[j]) {
          printf("(%c,%d%c,%g)", lp_lps->lp->is_lower[j] ? 'L' : 'U', j,
                 j <= lp_lps->lp->rows ? rgch[get_constr_type(lp_lps->lp, j)] : 'V', d[j]);
        }
      }
      printf("\n");
#endif

      /* Construct improving ray */
      for (int j = 1; j <= Dimension(lp->iy); ++j) {
        if (!lp_lps->lp->is_basic[lp_lps->mp_le_con[j-1]]) {
          lp->iy(j) = -d[lp_lps->mp_le_con[j-1]];
        }
      }
      for (int j = 1; j <= Dimension(lp->iz); ++j) {
        if (!lp_lps->lp->is_basic[lp_lps->mp_eq_con[j-1]]) {
          lp->iz(j) = -d[lp_lps->mp_eq_con[j-1]];
        }
      }

      iyzray = i;
      mempool_releaseVector(lp_lps->lp->workarrays, (char *) coltarget, FALSE);
      return true;
    }
  }

  mempool_releaseVector(lp_lps->lp->workarrays, (char *) coltarget, FALSE);
  return false;
} /* Sm_lps5_5::get_dual_ray */

bool
Sm_lps5_5::get_primal_ray(Lp *lp, int &ixray)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);
  int *coltarget;
  REAL d[lp_lps->lp->sum + 1];
  REAL pcol[lp_lps->lp->rows + 1];

  Clear(lp->ix);
  coltarget = (int *) mempool_obtainVector(lp_lps->lp->workarrays, lp_lps->lp->columns+1, sizeof(*coltarget));
  get_colIndexA(lp_lps->lp, SCAN_USERVARS+USE_NONBASICVARS, coltarget, FALSE);

#ifndef NDEBUG
  set_outputfile(lp_lps->lp, NULL);
  print_tableau(lp_lps->lp);
  set_outputfile(lp_lps->lp, "");
#endif

  /* The default for lp_solve is to minimize. The reduced costs d - see lp_matrix.c for the logic
   * behind these two calls - reflect this, so we have to look for negative ones. */
  bsolve(lp_lps->lp, 0, d, NULL, lp_lps->lp->epsmachine*DOUBLEROUND, 1.0);
  prod_xA(lp_lps->lp, coltarget, d, NULL, lp_lps->lp->epsmachine, 1.0, d, NULL, MAT_ROUNDDEFAULT);

  for (int i = ixray; i <= lp_lps->lp->sum; ++i) {
    if (!lp_lps->lp->is_basic[i] && d[i] < lp_lps->lp->epsdual) {
      fsolve(lp_lps->lp, i, pcol, NULL, lp_lps->lp->epsmachine, 1.0, FALSE);

      /* We might check for floating-point unboundedness here, but lp_solve should have followed all
       * improving, bounded directions. */

      /* We are only interested in the values of the structural variables */
      if (i > lp_lps->lp->rows) {
        lp->ix(i - lp_lps->lp->rows) = 1;
      }
      for (int j = 1; j <= lp_lps->lp->rows; ++j) {
        if (lp_lps->lp->var_basic[j] > lp_lps->lp->rows) {
          lp->ix(lp_lps->lp->var_basic[j] - lp_lps->lp->rows) = -pcol[j];
        }
      }

      if (!lp_lps->lp->is_lower[i]) lp->ix = -lp->ix;

      ixray = i;
      mempool_releaseVector(lp_lps->lp->workarrays, (char *) coltarget, FALSE);
      return true;
    }
  }

  mempool_releaseVector(lp_lps->lp->workarrays, (char *) coltarget, FALSE);
  return false;
} /* Sm_lps5_5::get_primal_ray */

void
Sm_lps5_5::set_primal_phase1(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  int n = get_Norig_columns(lp_lps->lp);
  int nn = Dimension(lp->ic);
  int m = RowDimension(lp->IA);
  int p = RowDimension(lp->IB);

  REAL *col = new REAL[m + 2];
  int *rowno = new int[m + 2];

  ::resize_lp(lp_lps->lp, get_Nrows(lp_lps->lp), nn);
  set_minim(lp_lps->lp);
  /* we just want to clear the objective */
  set_obj_fnex(lp_lps->lp, 0, col, rowno);

  col[0] = 1.0; 
  rowno[0] = 0;
  int j = n + 1;
  if ((m > 0) && (0 != Inf(lp->IA(1,n+1)))) {
    ++j;
    for (int i = 1; i <= m; ++i) {
      col[i] = (LE == get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) ? -1.0 : 1.0);
      rowno[i] = lp_lps->mp_le_con[i - 1];
    }
    add_columnex(lp_lps->lp, m + 1, col, rowno);
    /* set_lowbo(lp_lps->lp, get_Ncolumns(lp_lps->lp), lp->xl(n + 1)); */
    /* set_upbo(lp_lps->lp, get_Ncolumns(lp_lps->lp), lp->xu(n + 1)); */
  }

  for (int i = 1; i <= p && j <= nn; ++i) {
    if (0 != Inf(lp->IB(i, j))) {
      col[1] = Inf(lp->IB(i, j));
      rowno[1] = lp_lps->mp_eq_con[i - 1];
      add_columnex(lp_lps->lp, 2, col, rowno);
      ++j;
      if ((j <= nn) && (0 != Inf(lp->IB(i, j)))) {
        col[1] = Inf(lp->IB(i, j));
        add_columnex(lp_lps->lp, 2, col, rowno);
        ++j;
      }
    }
    /* set_lowbo(lp_lps->lp, get_Ncolumns(lp_lps->lp), lp->xl(n + 1 + i)); */
    /* set_upbo(lp_lps->lp, get_Ncolumns(lp_lps->lp), lp->xu(n + 1 + i)); */
  }

  char *sz = new char[strlen(lp->name) + 8 + 1];
  sprintf(sz, "%s-pp1.mps", lp->name);
  write_lp("Primal phase 1 problem", lp_lps->lp, sz);

  delete [] col;
  delete [] rowno;
} /* Sm_lps5_5::set_primal_phase1 */

void
Sm_lps5_5::set_dual_phase1(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  REAL *rhs = new REAL[Dimension(lp->ia) + Dimension(lp->ib) + 1];
  memset(rhs, 0, (Dimension(lp->ia) + Dimension(lp->ib) + 1) * sizeof (REAL));

  set_rh_vec(lp_lps->lp, rhs);

  delete [] rhs;

  for (int ix = 1; ix <= Dimension(lp->xl); ++ix) {
    set_lowbo(lp_lps->lp, ix, lp->xl(ix));
    set_upbo(lp_lps->lp, ix, lp->xu(ix));
  }

  char *sz = new char[strlen(lp->name) + 8 + 1];
  sprintf(sz, "%s-dp1.mps", lp->name);
  write_lp("Dual phase 1 problem", lp_lps->lp, sz);
} /* Sm_lps5_5::set_dual_phase1 */

/**
 * Take the model parameters for the right hand sides and the simple bounds from @p lp and write to
 * @p lp->module->lp restoring the original model after computation of the upper bound.
 *
 * @param[in] lp  contains the original parameters
 */
void
Sm_lps5_5::restore_primal(Lp *lp) 
{
  preport->print(3, "Restoring original lp");

  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  for (int i = 1; i <= RowDimension(lp->IA); ++i) {
    if (get_constr_type(lp_lps->lp, lp_lps->mp_le_con[i - 1]) == LE) {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1],  Mid(lp->ia(i)));
    } else {
      set_rh(lp_lps->lp, lp_lps->mp_le_con[i - 1], -Mid(lp->ia(i)));
    }
  }

  for (int i = 1; i <= Dimension(lp->xl); ++i) {
    set_lowbo(lp_lps->lp, i, lp->xl(i));
    set_upbo(lp_lps->lp, i, lp->xu(i));
  }
} /* Sm_lps5_5::restore_primal */

/**
 * Take the model parameters for the costs from @p lp and write to @p lp->module->lp
 * restoring the original model after computation of the lower bound.
 *
 * @param[in] lp  contains the original parameters
 */
void
Sm_lps5_5::restore_dual(Lp *lp) 
{
  preport->print(3, "Restoring original lp");

  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  if (lp->maximize) {
    /* lp_solve expects a 1-based array so we move the pointer */
    set_obj_fn(lp_lps, (-Mid(lp->ic)).theElements - 1);
  } else {
    /* lp_solve expects a 1-based array so we move the pointer */
    set_obj_fn(lp_lps, (Mid(lp->ic)).theElements - 1);
  }
} /* Sm_lps5_5::restore_dual */

void
Sm_lps5_5::restore_primal_phase1(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  ::resize_lp(lp_lps->lp, get_Nrows(lp_lps->lp), Dimension(lp->ic));

  if (lp->maximize) {
    set_maxim(lp_lps->lp);
    /* lp_solve expects a 1-based array so we move the pointer */
    set_obj_fn(lp_lps->lp, (-Mid(lp->ic)).theElements - 1);
  } else {
    /* lp_solve expects a 1-based array so we move the pointer */
    set_obj_fn(lp_lps->lp, (Mid(lp->ic)).theElements - 1);
  }
} /* Sm_lps5_5::restore_primal_phase1 */

void
Sm_lps5_5::restore_dual_phase1(Lp *lp)
{
  Lp_lp_solve *lp_lps = static_cast<Lp_lp_solve *>(lp->module);

  int m = Dimension(lp->ia);
  int p = Dimension(lp->ib);

  for (int iia = 1; iia <= m; ++iia) {
    if (FALSE == set_rh(lp_lps->lp, lp_lps->mp_le_con[iia-1], Mid(lp->ia(iia)))) {
      preport->print(1, "Failed to restore rhs on inequality %d (constraint %d, %.16e)",
                     iia, lp_lps->mp_le_con[iia-1], Mid(lp->ia(iia)));
    }
  }
  for (int iib = 1; iib <= p; ++iib) {
    if (FALSE == set_rh(lp_lps->lp, lp_lps->mp_eq_con[iib-1], Mid(lp->ib(iib)))) {
      preport->print(1, "Failed to restore rhs on equation %d (constraint %d, %.16e)",
                     iib, lp_lps->mp_eq_con[iib-1], Mid(lp->ib(iib)));
    }
  }

  for (int ix = 1; ix <= Dimension(lp->xl); ++ix) {
    if (FALSE == ::set_bounds(lp_lps->lp, ix, lp->xl(ix), lp->xu(ix))) {
      preport->print(1, "Failed to restore bounds on variable %d (%.16e, %.16e)",
                     ix, lp->xl(ix), lp->xu(ix));
    }
     /*
    if (-lp->infinite == lp->xl(ix)) set_lowbo(lp_lps->lp, ix, -lp->infinite);
    if ( lp->infinite == lp->xu(ix))  set_upbo(lp_lps->lp, ix,  lp->infinite);
    */
  }
} /* Sm_lps5_5::restore_dual_phase1 */

/**
 * Set the specified module options. These are supplied in @p argc and @p argv
 * in a command line version. Each option is supplied as one string entry in the
 * string array @p argv, @p argc specifies the number of options.
 * 
 * The module supports the following options
 * - '<tt>-sm,resetbas</tt>': Resolve lps in the case of numerical failure with
 *   a default basis of all slacks. This may enable lp_solve to find a solution
 *   at the expense of increased runtime.
 * - '<tt>-sm,s</tt>': Use lp_solve's default scaling (Numerical range-based
 *   scaling).
 * - '<tt>-sm,timeout,n</tt>': Set lp_solve's timeout to <tt>n</tt> seconds.
 * - '<tt>-sm,vn</tt>': Set lp_solve's verbosity level to <tt>n</tt>. The
 *   available levels are
 *   - <tt>v0</tt> NEUTRAL
 *   - <tt>v1</tt> CRITICAL
 *   - <tt>v2</tt> SEVERE
 *   - <tt>v3</tt> IMPORTANT (default)
 *   - <tt>v4</tt> NORMAL
 *   - <tt>v5</tt> DETAILED
 *   - <tt>v6</tt> FULL
 *
 * @param[out] lp     lp to set the options for
 * @param[in]  argc   number of arguments
 * @param[in]  argv   array of arguments
 *
 * @retval true
 * if options are set
 * @retval false
 * otherwise
 */
bool
Sm_lps5_5::set_module_options(Lp *lp, int argc, char *argv[]) 
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  for (int i = 0; i < argc; ++i) {
    if (strcmp(argv[i], "-sm,resetbas") == 0) {
      freset_bas = true;
    } else if (strcmp(argv[i], "-sm,s") == 0) {
      set_scaling(lp_lps, SCALE_MEAN);
    } else if (strcmp(argv[i], "-sm,t") == 0) {
      trace = TRUE;
      set_trace(lp_lps, TRUE);
    } else if (strncmp(argv[i], "-sm,timeout,", 12) == 0) {
      errno = 0;
      char *end;
      long t = strtol(argv[i] + 12, &end, 10);
      if (errno != 0) {
				preport->print(1, "Sm_lps5_5::set_module_options: %s", strerror(errno));
        return false;
      } else if (*end != '\0') {
        preport->print(1, "Sm_lps5_5::set_module_options: "
                       "error parsing timeout %s", argv[i]);
        return false;
      } else {
        timeout = t;
        set_timeout(lp_lps, timeout);
      }
    } else if (strncmp(argv[i], "-sm,v", 5) == 0) {
      errno = 0;
      char *end;
      int v = strtol(argv[i] + 5, &end, 10);
      if (errno != 0) {
				preport->print(1, "Sm_lps5_5::set_module_options: %s", strerror(errno));
        return false;
      } else if (*end != '\0') {
        preport->print(1, "Sm_lps5_5::set_module_options: error parsing verbosity %s", argv[i]);
      } else if ((v < 0) || (v > 6)) {
        preport->print(1, "Sm_lps5_5::set_module_options: unknown verbosity level: %s", argv[i]);
        return false;
      } else {
        verbosity = v;
        set_verbose(lp_lps, verbosity);
      }
    } else if (strncmp(argv[i], "-sm,wmps,", 9) == 0) {
      if (strcmp(argv[i] + 9, "all") == 0) {
        write_mps = wm_all;
      } else if (strcmp(argv[i] + 9, "ask") == 0) {
        write_mps = wm_ask;
      } else if (strcmp(argv[i] + 9, "none") == 0) {
        write_mps = wm_none;
      } else {
        preport->print(1, "Sm_lps5_5::set_module_options: unknown setting for -sm,wmps: %s", 
                       argv[i] + 9);
        return false;
      }
    } else {
      preport->print(1, "Sm_lps5_5::set_module_options: unknown option: %s", argv[i]);
      return false;
    }
  }
  return true;
} /* Sm_lps5_5::set_module_options */

/**
 * Get lp_solve's set accuracy
 *
 * @return
 * lp_solve's accuracy
 */
double
Sm_lps5_5::get_accuracy() 
{
  return DEF_EPSDUAL;
} /* Sm_lps5_5::get_accuracy */


void
Sm_lps5_5::lp2solver(Lp *lp)
{
  preport->print(4, "writing model to lp_solve 5.5");
  int c_le = RowDimension(lp->IA);
  int c_eq = RowDimension(lp->IB);
  int vars = ColDimension(lp->IA);

  lp->module = new Lp_lp_solve;
  lprec *lp_lps = make_lp(0, vars);
  static_cast<Lp_lp_solve *>(lp->module)->lp = lp_lps;

  if (lp->maximize) set_maxim(lp_lps);

  set_add_rowmode(lp_lps, TRUE);

  set_obj_fn(lp_lps, (Mid(lp->ic)).theElements - 1);

  for (int i = 1; i <= c_le; ++i) {
    add_constraint(lp_lps, Mid(Row(lp->IA, i)).theElements - 1, LE,
                   Mid(lp->ia(i))); }

  for (int i = 1; i <= c_eq; ++i) {
    add_constraint(lp_lps, Mid(Row(lp->IB, i)).theElements - 1, EQ,
                   Mid(lp->ib(i)));
  }

  set_add_rowmode(lp_lps, FALSE);

  for (int i = 1; i <= vars; ++i) {
    ::set_bounds(lp_lps, i, lp->xl(i), lp->xu(i));
  }

  /* Apply our log settings and the set defaults in the source lp */
  set_outputfile(lp_lps, "/dev/null");
  put_logfunc(lp_lps, lps_log, NULL);
  set_verbose(lp_lps, verbosity);
  set_timeout(lp_lps, timeout);
  set_trace(lp_lps, trace);

  int **pmp_eq_con = &static_cast<Lp_lp_solve *>(lp->module)->mp_eq_con;
  int **pmp_le_con = &static_cast<Lp_lp_solve *>(lp->module)->mp_le_con;
  *pmp_eq_con = new int[c_eq];
  *pmp_le_con = new int[c_le];
  int con = 1;
  for (int eq = 0; eq < c_le; ++eq) {
    (*pmp_le_con)[eq] = con++;
  }
  for (int le = 0; le < c_eq; ++le) {
    (*pmp_eq_con)[le] = con++;
  }

  char name[strlen(lp->name) + 4 + 1];
  sprintf(name, "%s.mps", lp->name);
  write_lp("Distance to primal infeasibility base problem", lp_lps, name);
} /* Sm_lps5_5::lp2solver */

void
Sm_lps5_5::set_bounds(Lp *lp, int var, double lower, double upper)
{
  lprec *lp_lps = static_cast<Lp_lp_solve *>(lp->module)->lp;

  ::set_bounds(lp_lps, var, lower, upper);

  int size = strlen(lp->name) + 32;
  char name[size + 1];
  if (snprintf(name, size + 1, "%s_%03d_%+.1e_%+.1e.mps", lp->name, var, lower, upper) > size) {
    preport->print(2, "Truncating file name for rho_p problem to %s", name);
    preport->print(2, "(var = %d, lower = %f, upper = %f)", var, lower, upper);
  }
  write_lp("Distance to primal infeasibility iterated problem", lp_lps, name);
} /* Sm_lps5_5::set_bounds */

/**
 * Set the pointers of @p i to point to the member functions of Sm_lps5_5.
 *
 * @param[out]  i  receives the pointers to member functions
 */
extern "C" void
get_func_pointers(Solver_module_interface &i)
/* Sm_lps5_5_LTX_get_func_pointers(Solver_module_interface &i) */
{
  i.get_version             = Sm_lps5_5::get_version;
  i.print_version           = Sm_lps5_5::print_version;
  i.print_brief_version     = Sm_lps5_5::print_brief_version;
  i.print_options           = Sm_lps5_5::print_options;
  i.init                    = Sm_lps5_5::init;
  i.solve_original          = Sm_lps5_5::solve_original;
  i.solve_primal_perturbed  = Sm_lps5_5::solve_primal_perturbed;
  i.solve_dual_perturbed    = Sm_lps5_5::solve_dual_perturbed;
  i.get_dual_ray            = Sm_lps5_5::get_dual_ray;
  i.get_primal_ray	        = Sm_lps5_5::get_primal_ray;
  i.set_primal_phase1	      = Sm_lps5_5::set_primal_phase1;
  i.set_dual_phase1	        = Sm_lps5_5::set_dual_phase1;
  i.restore_primal          = Sm_lps5_5::restore_primal;
  i.restore_dual            = Sm_lps5_5::restore_dual;
  i.restore_primal_phase1   = Sm_lps5_5::restore_primal_phase1;
  i.restore_dual_phase1     = Sm_lps5_5::restore_dual_phase1;
  i.read_lp                 = Sm_lps5_5::read_lp;
  i.set_lp                  = Sm_lps5_5::set_lp;
  i.set_module_options      = Sm_lps5_5::set_module_options;
  i.get_accuracy            = Sm_lps5_5::get_accuracy;
  i.lp2solver               = Sm_lps5_5::lp2solver;
  i.set_bounds              = Sm_lps5_5::set_bounds;
} /* get_func_pointers */
