/*
 * Lurupa, a library for verified linear programming.
 * Copyright (C) 2006 by Christian Keil
 *
 * This file is part of Lurupa.
 *
 * Lurupa is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

/**
 * This is Lurupa's command line client. It uses the core logic and a user
 * specified solver module to compute verified bounds for a linear program. The
 * results and the time needed to compute them are displayed and written to disk
 * in csv or LaTeX format.
 *
 * @file  lurupa_cmd.cpp
 * @brief Lurupa's command line client
 *  
 * @author Christian Keil
 * 
 * $Id: lurupa_cmd.cpp 533 2008-06-12 20:08:12Z keil $
 */
#include "svnversion.h"
#include <lurupa/globals.h>
#include <lurupa/Lurupa.h>
#include <lurupa/Lp.h>
#include <lr/Lr.h>
#include <ltdl.h>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <errno.h>
#include <iostream>
#include <new>
#include <time.h>
using namespace std;

/** @brief Version string */
static const char *version = "Lurupa command line client 1.2";
/** @brief Copyright string */
static const char *copyright = "Copyright (C) 2006 Christian Keil";

/**
 * @brief Information about a %Bound
 *
 * This structure aggregates the value of a single (lower or upper) bound and
 * corresponding data, that is the status of the bound and the iterations and
 * the time necessary to compute it.
 */
struct Bound {
  bool          compute;      /**< @brief Whether the bound shall be computed */
  double        value;        /**< @brief Value of the bound */
  Bound_status  status;       /**< @brief Status of the bound computation (i.e.,
                                successfull or failed due to ...) */
  int           iterations;   /**< @brief Iteration count */
  double        realtime;     /**< @brief Wall clock time needed to compute bound in seconds */
  double        proc_time;    /**< @brief Process time needed to compute bound in seconds */
  const char   *time_status;  /**< @brief Status of the time taken (i.e., if any wrap around is
                                possible / occurred) */
};

/**
 * @brief A certificate
 *
 * This structure stores information about a certificate.
 */
struct Certificate {
	bool          compute;      /**< @brief Whether the certificate shall be computed. */
	Bound_status  status;       /**< @brief Status of the certificate computation. */
	double        realtime;     /**< @brief Wall clock time needed to compute certificate in seconds.
																*/
	double        proc_time;    /**< @brief Process time needed to compute certificate in seconds. */
	const char   *time_status;  /**< @brief Status of the time taken (i.e., if any wrap around is
																possible / occured) */
};

/**
 * @brief An enclosure
 *
 * This structure stores information about an enclosure of a value.
 */
struct Enclosure {
  bool compute;  /**< @brief Whether the enclosure shall be computed. */
  double lower;  /**< @brief The lower bound. */
  double upper;  /**< @brief The upper bound. */
};

/**
 * @brief Information about an %Lp
 *
 * This structure aggregates information about a linear program. The path to its
 * file representation, information about the model itself, about the lp-solver
 * solving the model and about the bounds for the optimal value.
 */
struct Lp_stats {
  char *path;              /**< @brief Path to the %lp file */
  const char *name;        /**< @brief Name of the model */
  bool is_maximize;        /**< @brief Whether objective is to be maximized */
  double optimal_value;    /**< @brief Approximate optimal value from solver */
  Solver_status status;    /**< @brief Status of the solver (e.g., found optimal
                             solution, model unbounded, failure while solving) */
  double solve_realtime;   /**< @brief Wall clock time needed to solve the lp in seconds */
  double solve_proc_time;  /**< @brief Process time needed to solve the lp in seconds */
  const char *solve_time_status; 
	                         /**< @brief Status of the time taken (i.e., if any wrap around is
                             possible / occurred) */
  double total_realtime;   /**< @brief Wall clock time needed to solve the lp in seconds */
  double total_proc_time;  /**< @brief Process time needed to solve the lp in seconds */
  const char *total_time_status; 
	                         /**< @brief Status of the time taken (i.e., if any wrap around is
                             possible / occurred) */
	bool automatic;          /**< @brief Whether bounds are automatically selected for computing based
														 on approximate solver output */
  Bound lower;             /**< @brief The lower bound's data */
  Bound upper;             /**< @brief The upper bound's data */
	Certificate primal;      /**< @brief The primal certificates data */
	Certificate dual;        /**< @brief The dual certificates data */
  Enclosure rho_p;         /**< @brief Data for enclosure of distance to primal infeasibility */
  Enclosure rho_d;         /**< @brief Data for enclosure of distance to dual infeasibility */
  Enclosure condition;     /**< @brief Data for enclosure of condition number */
  double bound_quality;    /**< @brief Relative accuracy of the computed bounds */
};

/**
 * @brief Out of memory handler
 */
void
out_of_memory()
{
  fprintf(stderr, "lurupa_cmd: Detected out of memory in new handler.\n");
  exit(EXIT_FAILURE);
}

/**
 * @brief Print version information
 *
 * Print version information for the application parts.
 *
 * @param[in] lurupa             Lurupa instance to print information about
 * @param[in] solver_module_set  whether a solver module is loaded
 */
void
print_version(const Lurupa &lurupa, bool solver_module_set)
{
  printf("o======================================================o\n"
         "|                                                      |\n"
         "|                       Lurupa                         |\n"
         "|                     ,--------'                       |\n"
         "| Linear UndeRestimator Using Perturbed Approximations |\n"
         "|                                                      |\n"
         "| %52s |\n"
         "|                                                      |\n"
         "`======================================================'\n",
         copyright);
  printf("command line client\n"
         "  version : %s\n", version);
  lurupa.print_core_version();
  if (solver_module_set) {
    lurupa.print_module_version();
  } else {
    printf("No solver module loaded.\n");
  }
  if ((char *)0 != svn_version) {
    printf("svn version %s\n", svn_version);
  }
  exit(EXIT_SUCCESS);
}

/**
 * @brief Print brief version information
 *
 * Print brief version information suitable for log output.
 *
 * @param[in] lurupa             Lurupa instance to print information about
 * @param[in] solver_module_set  whether a solver module is loaded
 */
void
print_brief_version(const Lurupa &lurupa, bool solver_module_set)
{
  printf("Lurupa - %s\n", copyright);
  printf("command line client : %s\n", version);
  lurupa.print_core_brief_version();
  if (solver_module_set) {
    lurupa.print_module_brief_version();
  }
  if ((char *)0 != svn_version) {
    printf("svn version         : %s\n", svn_version);
  }
  exit(EXIT_SUCCESS);
}

/**
 * @brief Print usage information
 * 
 * Print usage information for the command line client; which parameters with
 * which values are supported.
 *
 * @param[in] self               path to running program
 * @param[in] lurupa             Lurupa instance to access solver module
 * @param[in] solver_module_set  whether a solver module is loaded
 */
void
print_usage(char *self, const Lurupa &lurupa, bool solver_module_set)
{
  printf("Usage:\n");
  printf("  %s <options>\n\n", self);

  printf("Options supported by %s:\n", version);
  printf("Mandatory\n");
  printf("  -lp file            Read the linear program to be processed from\n"
         "                      <file>. Must be in a format that can be\n"
         "                      interpreted by the chosen solver module.\n"
         "                      If this switch is not present, the model is\n"
         "                      read from stdin.\n");
  printf("  -sm file            Use the solver module <file> to solve the\n"
         "                      linear programs.\n");
  printf("Optional\n");
  printf("  -alpha d            Set algorithm parameter alpha to d.\n");
  printf("  -cond               Compute the condition number.\n");
  printf("  -csv file           Append the results to the csv file <file>[.csv],\n"
         "                      with the extension being appended if not present.\n");
  printf("  -eta d              Set algorithm parameter eta to d.\n");
  printf("  -h / -? / --help    Print this message.\n");
  printf("  -i d                Compute bounds for an interval problem derived from\n"
         "                      the one specified. Change all parameters to\n"
         "                      intervals with a relative radius of r.\n");
  printf("  -inflate            Try inflating the model if a perturbed one\n"
         "                      seems to be infeasible.\n");
  printf("  -latex file         Append the results to the latex table in the file\n"
         "                      <file>[.tex] with the extension being appended if\n"
         "                      not present.\n");
  printf("  -lb                 Compute the lower bound.\n");
  printf("  -rho_d              Compute the distance to dual infeasibility.\n");
  printf("  -rho_p              Compute the distance to primal infeasibility.\n");
  printf("  -t                  Prepend time information to messages.\n");
  printf("  -ub                 Compute the upper bound.\n");
  printf("  -vn                 Select verbosity level:\n"
         "                        -v0  No messages\n"
         "                        -v1  Errors\n"
         "                        -v2  Warnings (default)\n"
         "                        -v3  Brief\n"
         "                        -v4  Normal\n"
         "                        -v5  Verbose\n"
         "                        -v6  Full\n");
  printf("  -V                  Print version information.\n");
  printf("  -Vb                 Print brief version information.\n");
  printf("  -write_vm style     Write intermediate vectors and matrices to\n"
         "                      disk.\n"
         "                      If <style> is 'octave' the vectors and\n"
         "                      matrices are stored in one file named after\n"
         "                      the model to be solved.\n"
         "                      If <style> is 'matlab' the vectors and\n"
         "                      matrices are stored in several files in a\n"
         "                      subdirectory named after the model to be\n"
         "                      solved.\n");

  printf("\n");
  if (solver_module_set) {
    lurupa.print_module_options();
  } else {
    printf("No solver module loaded.\n");
  }

  printf("\nReport bugs to c.keil@tu-harburg.de.\n");
  exit(EXIT_SUCCESS);
}

/**
 * @brief Print model data
 *
 * Print the name and the optimization direction of the model to be processed.
 *
 * @param[in] lp_stats  contains the model's data
 */
void
print_model_data(const Lp_stats &lp_stats)
{
  printf("model                 : %s\n", lp_stats.name);
  printf("direction             : ");
  if (lp_stats.is_maximize) printf("maximize\n");
  else printf("minimize\n");
}

/**
 * @brief Start timer
 *
 * Get realtime and process time counter values. These can be compared to the ones returned by
 * #stop_timer to compute the elapsed time.
 *
 * @param[out] pt  process time counter value
 * @param[out] rt  realtime counter value
 *
 * @sa
 * #stop_timer
 */
void
#if HAVE_CLOCK_GETTIME
start_timer(clock_t &pt, struct timespec &rt)
#elif HAVE_TIME
start_timer(clock_t &pt, time_t &rt)
#else
#warning No realtime counter available
start_timer(clock_t &pt, int &rt)
#endif
{
#if HAVE_CLOCK_GETTIME
  int errnum;
  if (errnum = clock_gettime(CLOCK_REALTIME, &rt)) {
    fprintf(stderr, "Could not take realtime: %s", strerror(errnum));
		rt.tv_sec = -1;
  }
#elif HAVE_TIME
	if ((time_t) -1 == time(&rt)) {
		fprintf(stderr, "Could not take realtime: %s", strerror(errno));
	}
#else
	rt = -1;
	fprintf(stderr, "No realtime counter available");
#endif
  pt = clock();
}

/**
 * @brief Stop timer
 *
 * Get realtime and process time counter values. These can be compared to the ones returned by
 * #start_timer to compute the elapsed time.
 *
 * @param[out] pt  process time counter value
 * @param[out] rt  realtime counter value
 *
 * @sa
 * #start_timer
 */
void
#if HAVE_CLOCK_GETTIME
stop_timer(clock_t &pt, struct timespec &rt)
#elif HAVE_TIME
stop_timer(clock_t &pt, time_t &rt)
#else
#warning No realtime counter available
stop_timer(clock_t &pt, int &rt)
#endif
{
  pt = clock();
#if HAVE_CLOCK_GETTIME
  int errnum;
  if (errnum = clock_gettime(CLOCK_REALTIME, &rt)) {
    fprintf(stderr, "Could not take realtime: %s", strerror(errnum));
		rt.tv_sec = -1;
  }
#elif HAVE_TIME
	if ((time_t) -1 == time(&rt)) {
		fprintf(stderr, "Could not take realtime: %s", strerror(errno));
	}
#else
	rt = -1;
#endif
}

/**
 * @brief Compute timer difference
 *
 * Compute the elapsed realtime and process time verifying that no wrap around in the process time
 * counter occured. If a wrap around occured or cannot be ruled out or the calls to the time
 * routines failed, return a short string explaining what happened.
 * 
 * @param[in]  start_pt  process time counter value at start
 * @param[in]  end_pt    process time counter value at end
 * @param[in]  start_rt  realtime counter value at start
 * @param[in]  end_rt    realtime counter value at end
 * @param[out] pt_diff   elapsed process time in seconds
 * @param[out] rt_diff   elapsed realtime in seconds
 *
 * @return time status string
 */
const char *
#if HAVE_CLOCK_GETTIME
timer_diff(clock_t         start_pt, clock_t         end_pt,
           struct timespec start_rt, struct timespec end_rt,
           double &pt_diff, double &rt_diff)
#elif HAVE_TIME
timer_diff(clock_t start_pt, clock_t end_pt,
           time_t  start_rt, time_t  end_rt,
           double &pt_diff, double &rt_diff)
#else
timer_diff(clock_t start_pt, clock_t end_pt,
           int     start_rt, int     end_rt,
           double &pt_diff, double &rt_diff)
#endif
{
#if HAVE_CLOCK_GETTIME
	if (-1 != start_rt.tv_sec && -1 != end_rt.tv_sec) {
		rt_diff = ((end_rt.tv_sec - start_rt.tv_sec) 
								 + 1e-9*(end_rt.tv_nsec - start_rt.tv_nsec));
	} else {
		rt_diff = -1;
	}
#elif HAVE_TIME
	if (-1 != start_rt && -1 != end_rt) {
		rt_diff = end_rt - start_rt;
	} else {
		rt_diff = -1;
	}
#else
	rt_diff = -1;
#endif
  if (rt_diff > LONG_MAX / CLOCKS_PER_SEC) {
    /* we've got a wrap around in the <clock()> counter */
    pt_diff = -1;
    return "(process time wrap around)";
  } else if ((start_pt != (clock_t) -1) && (end_pt != (clock_t) -1)) {
    /* there's no wrap around in the counter */ 
    pt_diff = (end_pt - start_pt) / (double) CLOCKS_PER_SEC;
		if (-1 == rt_diff) {
			return "(possible process time wrap around)";
		} else {
			return "";
		}
  }

  pt_diff = -1;
  return "(error taking time)";
}

/**
 * @brief Process solving statistics
 *
 * Print statistics regarding the solving of the lp like the approximate optimal
 * value, the time needed to compute it, the status of the lp-solver (e.g., the
 * solver's judgement of the model's feasibility or timeout while solving the
 * model). Set @p lp.lower.compute and @p lp.upper.compute to false if the
 * lp-solver did not judge the model to be either feasible, infeasible, or
 * unbounded, that is returned with any kind of error.
 * 
 * @param[in,out] lp_stats  contains the solving statistics
 */
void
process_solving_status(Lp_stats &lp_stats)
{
  printf("     .---------------------.\n");
  printf("     | original problem    |\n");
  printf("     +---------------------+----------------------.\n");
  if (lp_stats.status == ss_feasible) {
    printf("     | app. optimal value  | %20.10e |\n", lp_stats.optimal_value);
  }
  printf("     | lp-solver status    | %20s |\n", solver_status_string[lp_stats.status]);
  printf("     | process time        |");
  if (lp_stats.solve_proc_time != -1) {
    printf(" %19.3gs | %s\n", lp_stats.solve_proc_time, lp_stats.solve_time_status);
  } else {
    printf(" %20s | %s\n", "error", lp_stats.solve_time_status);
  }
  printf("     | wall clock time     |");
  if (lp_stats.solve_realtime != -1) {
    printf(" %19.3gs |\n", lp_stats.solve_realtime);
  } else {
    printf(" %20s |\n", "error");
  }
  
  printf("     `---------------------+----------------------'\n");
  printf("\n");


  if ((lp_stats.status != ss_feasible) 
      && (lp_stats.status != ss_infeasible)
      && (lp_stats.status != ss_unbounded)) {
    lp_stats.lower.compute = false;
    lp_stats.upper.compute = false;
  }
}

/**
 * @brief Print bound statistics
 *
 * Print statistics about a computed bound. These include of course the bound
 * itself along with the count of algorithm iterations needed, the time needed,
 * the ratio of the time needed to compute the bounds and the time needed to
 * compute the approximate optimal value, the status of the computed bound (e.g.
 * bound successfully computed, no bound computed because of rank deficiency).
 *
 * @param[in] bound     the bound's statistics
 * @param[in] lp_stats  contains the time to solve the model
 */
void
print_bound_stats(const Bound &bound, const Lp_stats &lp_stats) 
{
  printf("     | status              | %23s |\n", bound_status_string[bound.status]);
  printf("     | iterations          | %23d |\n", bound.iterations);

  printf("     | process time        | ");
  if (bound.proc_time != -1) {
    printf("%22.3gs | %s\n", bound.proc_time, bound.time_status);
  } else {
    printf("%23s | %s\n", "error", bound.time_status);
  }
  printf("     | time / solving time | ");
  if (lp_stats.solve_proc_time > 0) {
    printf("%23f |\n", bound.proc_time / lp_stats.solve_proc_time);
  } else {
    printf("%23s |\n", "-");
  }

  printf("     | wall clock time     | ");
  if (bound.realtime != -1) {
    printf("%22.3gs |\n", bound.realtime);
  } else {
    printf("%23s |\n", "error");
  }
  printf("     | time / solving time | ");
  if (lp_stats.solve_realtime > 0) {
    printf("%23f |\n", bound.realtime / lp_stats.solve_realtime);
  } else {
    printf("%23s |\n", "-");
  }
}

/**
 * @brief Compute bound quality
 *
 * Compute and print the accuracy of the verified bounds. It is computed by 
 * @f[
 *   \mu(a, b) = \frac{\lvert a - b \rvert}
 *                    {\max \{1, \lvert a + b \rvert / 2\}},
 * @f]
 * where @f$a@f$ and @f$b@f$ are the computed bounds or the approximate
 * optimal value if one of the bounds could not be computed. Note that @f$\mu@f$ is computed from
 * the bound values as they are computed and not from the printed ones. The printed values of the
 * bounds are rounded to display a rigorous enclosure of the optimal value.
 *
 * @param[in,out] lp_stats contains the verified bounds
 */
void
report_bound_quality(Lp_stats &lp_stats)
{
  if (((lp_stats.lower.value == BiasNegInf) || !lp_stats.lower.compute)
      && ((lp_stats.upper.value == BiasPosInf) || !lp_stats.upper.compute))
    /* We computed no bound, no point in printing u(optimal, optimal) ,o) */
    return;

  double up, lo;
  char u[19 + 1];
  short u_offset;

  if ((lp_stats.lower.value != BiasNegInf) && lp_stats.lower.compute) {
    lo = lp_stats.lower.value;
    sprintf(u, "u(lower");
    u_offset = 7;
  } else {
    lo = lp_stats.optimal_value;
    sprintf(u, "u(optimal");
    u_offset = 9;
  }
  if ((lp_stats.upper.value != BiasPosInf) && lp_stats.upper.compute) {
    up = lp_stats.upper.value;
    sprintf(u + u_offset, ", upper)");
  } else {
    up = lp_stats.optimal_value;
    sprintf(u + u_offset, ", optimal)");
  }

  double abs = 0.5 * (fabs(lo) + fabs(up));
  lp_stats.bound_quality = fabs(up - lo) / (abs > 1 ? abs : 1);

  printf("    %22s = %.10e\n", u, lp_stats.bound_quality);
  printf("\n");
}

/**
 * @brief Compute lower bound
 *
 * Compute and time a lower bound for the optimal value of @p lp.
 *
 * @param[in]  lurupa    Lurupa reference used to compute bound
 * @param[in]  lp        the %lp
 * @param[out] lp_stats  stores bound and time
 */
void
compute_lower(Lurupa *lurupa, Lp *lp, Lp_stats *lp_stats)
{
  clock_t start_pt, end_pt;
#if HAVE_CLOCK_GETTIME
  struct timespec start_rt, end_rt;
#elif HAVE_TIME
	time_t start_rt, end_rt;
#else
	int start_rt, end_rt;
#endif

  start_timer(start_pt, start_rt);
  lp_stats->lower.status = lurupa->lower_bound(lp, lp_stats->lower.value, lp_stats->lower.iterations);
  stop_timer(end_pt, end_rt);
  lp_stats->lower.time_status = timer_diff(start_pt, end_pt, start_rt, end_rt,
                                           lp_stats->lower.proc_time, lp_stats->lower.realtime);
}

/**
 * @brief Compute upper bound
 *
 * Compute and time an upper bound for the optimal value of @p lp.
 *
 * @param[in]  lurupa    Lurupa reference used to compute bound
 * @param[in]  lp        the %lp
 * @param[out] lp_stats  stores bound and time
 */
void
compute_upper(Lurupa *lurupa, Lp *lp, Lp_stats *lp_stats)
{
  clock_t start_pt, end_pt;
#if HAVE_CLOCK_GETTIME
  struct timespec start_rt, end_rt;
#elif HAVE_TIME
	time_t start_rt, end_rt;
#else
	int start_rt, end_rt;
#endif

  start_timer(start_pt, start_rt);
  lp_stats->upper.status = lurupa->upper_bound(lp, lp_stats->upper.value, lp_stats->upper.iterations);
  stop_timer(end_pt, end_rt);
  lp_stats->upper.time_status = timer_diff(start_pt, end_pt, start_rt,  end_rt,
                                           lp_stats->upper.proc_time, lp_stats->upper.realtime);
}

/**
 * @brief Compute primal certificate
 *
 * Compute and time a primal certificate for dual infeasibility of @p lp.
 *
 * @param[in]  lurupa    Lurupa reference used to compute certificate
 * @param[in]  lp        the %lp
 * @param[out] lp_stats  stores certificate status and time
 */
void
compute_primal(Lurupa *lurupa, Lp *lp, Lp_stats *lp_stats)
{
  clock_t start_pt, end_pt;
#if HAVE_CLOCK_GETTIME
  struct timespec start_rt, end_rt;
#elif HAVE_TIME
	time_t start_rt, end_rt;
#else
	int start_rt, end_rt;
#endif

  start_timer(start_pt, start_rt);
  lp_stats->primal.status = lurupa->primal_certificate(lp);
  stop_timer(end_pt, end_rt);
  lp_stats->primal.time_status = timer_diff(start_pt, end_pt, start_rt,  end_rt,
																						lp_stats->primal.proc_time, lp_stats->primal.realtime);
}

/**
 * @brief Compute dual certificate
 *
 * Compute and time a dual certificate for primal infeasibility @p lp.
 *
 * @param[in]  lurupa    Lurupa reference used to compute certificate
 * @param[in]  lp        the %lp
 * @param[out] lp_stats  stores certificate status and time
 */
void
compute_dual(Lurupa *lurupa, Lp *lp, Lp_stats *lp_stats)
{
  clock_t start_pt, end_pt;
#if HAVE_CLOCK_GETTIME
  struct timespec start_rt, end_rt;
#elif HAVE_TIME
	time_t start_rt, end_rt;
#else
	int start_rt, end_rt;
#endif

  start_timer(start_pt, start_rt);
  lp_stats->dual.status = lurupa->dual_certificate(lp);
  stop_timer(end_pt, end_rt);
  lp_stats->dual.time_status = timer_diff(start_pt, end_pt, start_rt,  end_rt,
																					lp_stats->dual.proc_time, lp_stats->dual.realtime);
}

/**
 * @brief Correctly round double to string
 *
 * Generate a correctly rounded string representation of @p x. The direction of rounding can be
 * selected via @p direction. Possible values are @c LR_RND_DOWN, LR_RND_UP, LR_RND_NEAR,
 * LR_RND_CHOP.
 *
 * @param[in] x          the double value to convert
 * @param[in] direction  direction of rounding
 * @return rounded string representation
 */
const char *
rounded_string(double x, int direction)
{
	LREAL lr = LrNewVar(6);
	dtolr(lr, &x);
	static char sz[10 + 14];
	PVOID wrk1, wrk2, wrk3;
	LONG dummy;
	USHORT Len = (USHORT) (( ((((LONG) 10 + 9) >> 2) << 16) + 83845L) / 78913L);
	wrk1 = malloc((Len + 14) * sizeof (LDIGIT) + sizeof (LREALHDR));
	wrk2 = malloc((Len + 14) * sizeof (LDIGIT) + sizeof (LREALHDR));
	wrk3 = malloc((Len + 33) * sizeof (LDIGIT));
	return lrtosze(lr, sz, 10, 2, direction, wrk1, wrk2, wrk3, &dummy);
}

/**
 * @brief Report lower bound
 *
 * Report value and stats of the lower bound on the optimal value of @p lp.
 *
 * @param[in] lp        the lp
 * @param[in] lp_stats  the stats
 */
void
report_lower(Lp *lp, Lp_stats &lp_stats)
{
  printf("     .---------------------.\n");
  printf("     | lower bound         |");
  if (lp_stats.status == ss_feasible) {
    printf("%22s\n", lp_stats.lower.value > lp_stats.optimal_value ? "!!!" : "");
  } else {
    printf("\n");
  }
  printf("     +---------------------+-------------------------.\n");
  printf("     | bound value         |       %s |\n",
				 rounded_string(lp_stats.lower.value, LR_RND_DOWN));
  print_bound_stats(lp_stats.lower, lp_stats);
  printf("     `---------------------+-------------------------'\n");
  printf("\n");
}

/**
 * @brief Report upper bound
 *
 * Report value and stats of the upper bound on the optimal value of @p lp.
 *
 * @param[in] lp        the lp
 * @param[in] lp_stats  the stats
 */
void
report_upper(Lp *lp, Lp_stats &lp_stats)
{
  printf("     .---------------------.\n");
  printf("     | upper bound         |");
  if (lp_stats.status == ss_feasible) {
    printf("%22s\n", lp_stats.upper.value < lp_stats.optimal_value ? "!!!" : "   ");
  } else {
    printf("\n");
  }
  printf("     +---------------------+-------------------------.\n");
  printf("     | bound value         |       %s |\n",
				 rounded_string(lp_stats.upper.value, LR_RND_UP));
  print_bound_stats(lp_stats.upper, lp_stats);
  printf("     `---------------------+-------------------------'\n");
  printf("\n");
}

/**
 * @brief Report primal certificate
 *
 * Report success and time of computing a primal certificate for an %lp.
 *
 * @param[in] lp_stats  contains the certificate to report
 */
void
report_primal(Lp_stats &lp_stats)
{
  printf("     .---------------------.\n");
  printf("     | primal certificate  |\n");
  printf("     +---------------------+-------------------------.\n");
  printf("     | status              | %23s |\n", bound_status_string[lp_stats.primal.status]);
  printf("     | process time        | ");
  if (lp_stats.primal.proc_time != -1) {
    printf("%22.3gs | %s\n", lp_stats.primal.proc_time, lp_stats.primal.time_status);
  } else {
    printf("%23s | %s\n", "error", lp_stats.primal.time_status);
  }
  printf("     | wall clock time     | ");
  if (lp_stats.primal.realtime != -1) {
    printf("%22.3gs |\n", lp_stats.primal.realtime);
  } else {
    printf("%23s |\n", "error");
  }
  printf("     `---------------------+-------------------------'\n");
  printf("\n");
}

/**
 * @brief Report dual certificate
 *
 * Report success and time of computing a dual certificate for an %lp.
 *
 * @param[in] lp_stats  contains the certificate to report
 */
void
report_dual(Lp_stats &lp_stats)
{
  printf("     .---------------------.\n");
  printf("     | dual certificate    |\n");
  printf("     +---------------------+-------------------------.\n");
  printf("     | status              | %23s |\n", bound_status_string[lp_stats.dual.status]);
  printf("     | process time        | ");
  if (lp_stats.dual.proc_time != -1) {
    printf("%22.3gs | %s\n", lp_stats.dual.proc_time, lp_stats.dual.time_status);
  } else {
    printf("%23s | %s\n", "error", lp_stats.dual.time_status);
  }
  printf("     | wall clock time     | ");
  if (lp_stats.dual.realtime != -1) {
    printf("%22.3gs |\n", lp_stats.dual.realtime);
  } else {
    printf("%23s |\n", "error");
  }
  printf("     `---------------------+-------------------------'\n");
  printf("\n");
}

/**
 * @brief Write LaTeX table
 *
 * Write a line containing the gathered results to disk in LaTeX table format
 * with precision tailored to viewing. This consists of 
 * - name of the model
 * - approximate optimal solution or the solver status if solving failed
 * - time needed to solve the model
 * - verified lower bound's data
 *   - bound value or the bound status if its computation failed
 *   - number of algorithm iterations to compute the bound
 *   - ratio of the time needed to compute the bound and the time needed to
 *     solve the model
 * - verified upper bound's data
 *   - bound value or the bound status if its computation failed
 *   - number of algorithm iterations to compute the bound
 *   - ratio of the time needed to compute the bound and the time needed to
 *     solve the model
 * - relative accuracy of the bounds
 *
 * @param[in] lp_stats     contains the model's and bound's statistics
 * @param[in] latex        file to be written to
 */
void
write_latex_table(const Lp_stats &lp_stats, FILE *latex)
{
  const char *rho = NULL;
  fprintf(latex, "%-12s & ", lp_stats.name);

  if (lp_stats.status != ss_feasible)
    fprintf(latex, "%13s & ", solver_status_string[lp_stats.status]);
  else
    fprintf(latex, "$% .4e$ & ", lp_stats.optimal_value);

  fprintf(latex, "$%9.3f$ & ", lp_stats.solve_proc_time);
  fprintf(latex, "$%9.3f$ & ", lp_stats.solve_realtime);

  if (lp_stats.lower.compute) {
    if (lp_stats.lower.status != bs_verified)
      fprintf(latex, "%13s & ", bound_status_string[lp_stats.lower.status]);
    else
      fprintf(latex, "$% .4e$ & ", lp_stats.lower.value);

    fprintf(latex, "$%2d$ & ", lp_stats.lower.iterations);

    if (0 >= lp_stats.solve_proc_time || -1 == lp_stats.lower.proc_time) {
      fprintf(latex, "%9s & ", "");
    } else {
      fprintf(latex, "$%7.3f$ & ", lp_stats.lower.proc_time / lp_stats.solve_proc_time);
    }

    if (0 >= lp_stats.solve_realtime || -1 == lp_stats.lower.realtime) {
      fprintf(latex, "%9s & ", "");
    } else {
      fprintf(latex, "$%7.3f$ & ", lp_stats.lower.realtime / lp_stats.solve_realtime);
    }
  } else {
    fprintf(latex, "%13s & %4s & %9s & %9s & ", "", "", "", "");
  }

  if (lp_stats.upper.compute) {
    if (lp_stats.upper.status != bs_verified) {
      fprintf(latex, "%13s & ", bound_status_string[lp_stats.upper.status]);
    } else {
      fprintf(latex, "$% .4e$ & ", lp_stats.upper.value);
    }

    fprintf(latex, "$%2d$ & ", lp_stats.upper.iterations);

    if (0 >= lp_stats.solve_proc_time || -1 == lp_stats.upper.proc_time) {
      fprintf(latex, "%9s & ", "");
    } else {
      fprintf(latex, "$%7.3f$ & ", lp_stats.upper.proc_time / lp_stats.solve_proc_time);
    }

    if (0 >= lp_stats.solve_realtime || -1 == lp_stats.upper.realtime) {
      fprintf(latex, "%9s & ", "");
    } else {
      fprintf(latex, "$%7.3f$ & ", lp_stats.upper.realtime / lp_stats.solve_realtime);
    }
  } else {
    fprintf(latex, "%13s & %4s & %9s & %9s & ", "", "", "", "");
  }

  if (((lp_stats.lower.value != BiasNegInf) && lp_stats.lower.compute)
      || ((lp_stats.upper.value != BiasPosInf) && lp_stats.upper.compute)) {
    fprintf(latex, "$%.4e$ \\\\\n", lp_stats.bound_quality);
  } else {
    fprintf(latex, "%12s \\\\\n", "", "", "", "");
  }
}

/**
 * @brief Write csv table
 *
 * Write a line containing the gathered results with maximal precision to disk
 * in csv format. This includes
 * - name of the model
 * - status of the solver solving the original model
 * - approximate optimal value
 * - time needed to compute the approximate value
 * - verified lower bound's data
 *   - bound status
 *   - bound value
 *   - number of iterations to compute the bound
 *   - time needed to compute the bound
 * - verified upper bound's data
 *   - bound status
 *   - bound value
 *   - number of iterations to compute the bound
 *   - time needed to compute the bound
 * - relative accuracy of the bounds
 *
 * @param[in] lp_stats      contains the model's and bound's statistics
 * @param[in] csv           file to be written to
 */
void
write_csv_table(const Lp_stats &lp_stats, FILE *csv)
{
  const char *rho = NULL;

  fprintf(csv,
          "%s,%s,%.16e,%.16e,%.16e",
          lp_stats.name, solver_status_string[lp_stats.status], lp_stats.optimal_value,
          lp_stats.solve_proc_time,lp_stats.solve_realtime);

  if (lp_stats.lower.compute) {
    fprintf(csv, ",%s,%.16e,%d,%.16e,%.16e", 
            bound_status_string[lp_stats.lower.status], lp_stats.lower.value,
            lp_stats.lower.iterations, lp_stats.lower.proc_time, lp_stats.lower.realtime);
  } else {
    fprintf(csv, ",NA,NA,NA,NA,NA,");
  }


  if (lp_stats.upper.compute) {
    fprintf(csv, ",%s,%.16e,%d,%.16e,%.16e", 
            bound_status_string[lp_stats.upper.status], lp_stats.upper.value,
            lp_stats.upper.iterations, lp_stats.upper.proc_time,lp_stats.upper.realtime);
  } else {
    fprintf(csv, ",NA,NA,NA,NA,NA,");
  }

  if (((lp_stats.lower.value != BiasNegInf) && lp_stats.lower.compute)
      || ((lp_stats.upper.value != BiasPosInf) && lp_stats.upper.compute)) {
    fprintf(csv, ",%.16e\n", lp_stats.bound_quality);
  } else {
    fprintf(csv, ",NA\n");
  }
}

/**
 * @brief Write tables
 *
 * Calls the routines to write LaTeX and csv tables as specified by tables.
 *
 * @param[in] tables        specifies the tables to be written (c.f. globals.h)
 * @param[in] lp_stats      contains the model's and bound's statistics
 * @param[in] latex         file to write LaTeX table to
 * @param[in] csv           file to write csv table to
 */
void
write_tables(short tables, const Lp_stats &lp_stats, FILE *latex, FILE *csv)
{
  if (tables & LU_T_LATEX) {
    write_latex_table(lp_stats, latex);
  }
  if (tables & LU_T_CSV) {
    write_csv_table(lp_stats, csv);
  }
}

/**
 * @brief Command line main
 *
 * Parse the command line arguments and call Lurupa appropriately. Print the
 * results and write them to disk.
 */
int
main(int argc, char *argv[])
{
  bool solver_module_set = false;
  bool arg_version = false;
  bool arg_brief_version = false;
  bool arg_usage = false;
  bool arg_print_time = false;
  bool arg_write_vm = false;
  short tables = 0;
  int solver_module_argc = 0;
  char **solver_module_argv = new char*[argc - 1];
  int verbosity = 2;
  double relative_interval_radius = 0.0;
  double alpha, eta;
  char **solver_options = NULL;
  int latex_arg, csv_arg;
  FILE *latex = NULL, *csv = NULL;
  Csv_style csv_style;
  Lp_stats lp_stats = {0};
	lp_stats.automatic = true;
  lp_stats.lower.compute = false;
  lp_stats.upper.compute = false;
	lp_stats.primal.compute = false;
	lp_stats.dual.compute = false;
  lp_stats.rho_p.compute = false;
  lp_stats.rho_d.compute = false;
  lp_stats.condition.compute = false;
  Lurupa lurupa;
  Lp *lp;

  set_new_handler(out_of_memory);
  LTDL_SET_PRELOADED_SYMBOLS();

  errno = 0;
  /* first of all lets parse the command line arguments */
  for (int i = 1; i < argc; ++i) {
    if (strcmp(argv[i], "-alpha") == 0) {
      if (i + 1 < argc) {
        ++i;
        alpha = strtod(argv[i], NULL);
        if (errno == ERANGE) {
          perror("Setting alpha");
          exit(EXIT_FAILURE);
        }
        lurupa.set_alpha(alpha);
      } else {
        fprintf(stderr, "Alpha is missing\n");
        exit(EXIT_FAILURE);
      }
    } else if (strcmp(argv[i], "-cond") == 0) {
      lp_stats.condition.compute = true;
    } else if (strcmp(argv[i], "-csv") == 0) {
      if ((i + 1 < argc) && (*argv[i + 1] != '-')) {
        ++i;
        tables |= LU_T_CSV;

        int l = strlen(argv[i]);
        char *csv_name = NULL;
        if (strcmp(argv[i] + l - 4 - 1, ".csv") != 0) {
          csv_name = new char[l + 4 + 1];
          sprintf(csv_name, "%s.csv", argv[i]);
        } else {
          csv_name = new char[l + 1];
          strcpy(csv_name, argv[i]);
        }

        if ((csv = fopen(csv_name, "a")) == NULL) {
          perror("Opening csv file");
          fprintf(stderr, "filename: %s\n", csv_name);
          exit(EXIT_FAILURE);
        }
      } else {
        fprintf(stderr, "Csv filename is missing\n");
        exit(EXIT_FAILURE);
      }
		} else if (strcmp(argv[i], "-dc") == 0) {
			lp_stats.dual.compute = true;
			lp_stats.automatic = false;
    } else if (strcmp(argv[i], "-eta") == 0) {
      if (i + 1 < argc) {
        ++i;
        eta = strtod(argv[i], NULL);
        if (errno == ERANGE) {
          perror("Setting eta");
          exit(EXIT_FAILURE);
        }
        lurupa.set_eta(eta);
      } else {
        fprintf(stderr, "Eta is missing\n");
        exit(EXIT_FAILURE);
      }
    } else if ((strcmp(argv[i], "-h") == 0)
               || (strcmp(argv[i], "--help") == 0)
               || (strcmp(argv[i], "-?") == 0)) {
      arg_usage = true;
    } else if (strcmp(argv[i], "-i") == 0) {
      if (i + 1 < argc) {
        ++i;
        relative_interval_radius = strtod(argv[i], NULL);
        if (errno == ERANGE) {
          perror("Setting interval radius");
          exit(EXIT_FAILURE);
        }
      } else {
        fprintf(stderr, "Interval radius is missing\n");
        exit(EXIT_FAILURE);
      }
    } else if (strcmp(argv[i], "-inflate") == 0) {
      lurupa.set_inflate(true);
    } else if (strcmp(argv[i], "-latex") == 0) {
      if ((i + 1 < argc) && (*argv[i + 1] != '-')) {
        ++i;
        tables |= LU_T_LATEX;

        int l = strlen(argv[i]);
        char *latex_name = NULL;
        if (strcmp(argv[i] + l - 4 - 1, ".tex") != 0) {
          latex_name = new char[l + 4 + 1];
          sprintf(latex_name, "%s.tex", argv[i]);
        } else {
          latex_name = new char[l + 1];
          strcpy(latex_name, argv[i]);
        }

        if ((latex = fopen(latex_name, "a")) == NULL) {
          perror("Opening latex file");
          fprintf(stderr, "filename: %s\n", latex_name);
          exit(EXIT_FAILURE);
        }
      } else {
        fprintf(stderr, "Latex filename is missing\n");
        exit(EXIT_FAILURE);
      }
    } else if (strcmp(argv[i], "-lb") == 0) {
      lp_stats.lower.compute = true;
			lp_stats.automatic = false;
    } else if (strcmp(argv[i], "-lp") == 0) {
      if (i + 1 < argc) {
        lp_stats.path = argv[++i];
      } else {
        fprintf(stderr, "Linear program filename is missing\n");
        exit(EXIT_FAILURE);
      }
		} else if (strcmp(argv[i], "-pc") == 0) {
			lp_stats.primal.compute = true;
			lp_stats.automatic = false;
    } else if (strcmp(argv[i], "-rho_d") == 0) {
      lp_stats.rho_d.compute = true;
    } else if (strcmp(argv[i], "-rho_p") == 0) {
      lp_stats.rho_p.compute = true;
    } else if (strcmp(argv[i], "-sm") == 0) {
      if (i + 1 < argc) {
        ++i;
        solver_module_set = true;
        if (!lurupa.set_module(argv[i])) {
          fprintf(stderr, "Solver module could not be opened\n");
          exit(EXIT_FAILURE);
        }
      } else {
        fprintf(stderr, "Solver module filename is missing\n");
        exit(EXIT_FAILURE);
      }
    } else if (strncmp(argv[i], "-sm,", 4) == 0) {
      /* it's a solver module option */
      solver_module_argv[solver_module_argc++] = argv[i];
    } else if (strcmp(argv[i], "-t") == 0) {
      arg_print_time = true;
    } else if (strcmp(argv[i], "-ub") == 0) {
      lp_stats.upper.compute = true;
			lp_stats.automatic = false;
    } else if (strncmp(argv[i], "-v", 2) == 0) {
      verbosity = atoi(argv[i] + 2);
      if ((verbosity < 0) || (verbosity > 6)) {
        fprintf(stderr, "Unknown verbosity level: %s\n", argv[i]);
        exit(EXIT_FAILURE);
      }
    } else if (strcmp(argv[i], "-V") == 0) {
      arg_version = true;
    } else if (strcmp(argv[i], "-Vb") == 0) {
      arg_brief_version = true;
    } else if (strcmp(argv[i], "-write_vm") == 0) {
      arg_write_vm = true;
      if (i + 1 < argc) {
        if (strcmp(argv[++i], "octave") == 0) {
          csv_style = octave;
        } else if (strcmp(argv[i], "matlab") == 0) {
          csv_style = matlab;
        } else {
          fprintf(stderr, "Unknown vector/matrix style: %s\n", argv[i]);
          exit(EXIT_FAILURE);
        }
      } else {
        fprintf(stderr, "Vector/matrix style is missing\n");
        exit(EXIT_FAILURE);
      }
    } else {
      fprintf(stderr, "Unknown option: %s\n", argv[i]);
      arg_usage = true;
    }
  } /* parse command line parameters */

  if (arg_version) print_version(lurupa, solver_module_set);
  if (arg_brief_version) print_brief_version(lurupa, solver_module_set);
  if (arg_usage) print_usage(argv[0], lurupa, solver_module_set);
  lurupa.report.set_verbosity(verbosity, arg_print_time, arg_write_vm);
  if (arg_write_vm) lurupa.report.set_csv_style(csv_style);

  /* Check that all necessary parameters where supplied. */
  if (!solver_module_set) {
    fprintf(stderr, "No solver module file specified\n");
    exit(EXIT_FAILURE);
  }

  /* As we've got all information we need we can start the processing. */
  time_t timestamp = time(NULL);
  printf("start                 : %s", ctime(&timestamp));
  if (NULL != lp_stats.path) {
    printf("file                  : %s\n", lp_stats.path);
  } else {
    printf("reading LP from stdin\n");
  }

  /* Now we'll let the lp solver parse the lp file and print some statistics
   * about the contained model to the screen and, if specified, to the latex
   * tables. */
  if (verbosity > 2) printf("\n");
  FILE *in = NULL;
  if (lp_stats.path == NULL) {
    in = stdin;
  } else if ((in = fopen(lp_stats.path, "r")) == NULL) {
    printf("Could not open linear program file\n");
    exit(EXIT_FAILURE);
  }
  if ((lp = lurupa.read_lp(in, relative_interval_radius)) == NULL) {
    printf("Error reading linear program file\n");
    exit(EXIT_FAILURE);
  }
  if (in != stdin) fclose(in);
  if (verbosity > 2) printf("\n");
  lp_stats.name = lp->name;
  lp_stats.is_maximize = lp->maximize;
  print_model_data(lp_stats);

  /* Set solver module options. */
  if ((0 != solver_module_argc) 
      && !lurupa.set_module_options(lp, solver_module_argc, solver_module_argv)) {
    fprintf(stderr, "Could not set module options\n");
    delete [] solver_module_argv;
    exit(EXIT_FAILURE);
  }
  delete [] solver_module_argv;

  printf("\n======================================================\n");
  printf("\n");

  clock_t start_pt, start_total_pt, end_pt;
#if HAVE_CLOCK_GETTIME
  struct timespec start_rt, start_total_rt, end_rt;
#elif HAVE_TIME
	time_t start_rt, start_total_rt, end_rt;
#else
	int start_rt, start_total_rt, end_rt;
#endif
  start_timer(start_total_pt, start_total_rt);
  /* Let's get to work. */
  if (lp_stats.automatic
			|| lp_stats.lower.compute || lp_stats.upper.compute 
			|| lp_stats.primal.compute || lp_stats.dual.compute) {
    /* Compute the solution of the lp as an approximation for the bound computing algorithm to start
     * with, and print some statistics about the solving process. */

    printf(" [*] approximative computations\n");

    printf("\n");
    start_timer(start_pt, start_rt);
    lp_stats.status = lurupa.solve_lp(lp, lp_stats.optimal_value);
    if (ss_failure == lp_stats.status) {
      printf("Failed to solve original linear program\n");
      exit(EXIT_FAILURE);
    }
    stop_timer(end_pt, end_rt);
    if (verbosity > 2) printf("\n");
		lp_stats.solve_time_status = timer_diff(start_pt, end_pt, start_rt, end_rt, 
                                            lp_stats.solve_proc_time, lp_stats.solve_realtime);
    process_solving_status(lp_stats);

    printf(" [*] verified bounds");
    if (0 != relative_interval_radius) {
      printf(" (%.0e interval extension)", relative_interval_radius);
    }
    printf("\n\n");

		if (lp_stats.automatic) {
			if (ss_feasible == lp_stats.status) {
				compute_lower(&lurupa, lp, &lp_stats);
				if (verbosity > 2) printf("\n");
				compute_upper(&lurupa, lp, &lp_stats);
				if (verbosity > 2) printf("\n");
				report_lower(lp, lp_stats);
				report_upper(lp, lp_stats);
				report_bound_quality(lp_stats);
			} else if (ss_unbounded == lp_stats.status) {
				compute_primal(&lurupa, lp, &lp_stats);
				report_primal(lp_stats);
			} else if (ss_infeasible == lp_stats.status) {
				compute_dual(&lurupa, lp, &lp_stats);
				report_dual(lp_stats);
			}
		} else {
			if (lp_stats.lower.compute) {
				compute_lower(&lurupa, lp, &lp_stats);
				if (verbosity > 2) printf("\n");
			}
			if (lp_stats.upper.compute) {
				compute_upper(&lurupa, lp, &lp_stats);
				if (verbosity > 2) printf("\n");
			}

			if (lp_stats.primal.compute) {
				compute_primal(&lurupa, lp, &lp_stats);
				if (verbosity > 2) printf("\n");
			}
			if (lp_stats.dual.compute) {
				compute_dual(&lurupa, lp, &lp_stats);
				if (verbosity > 2) printf("\n");
			}

			if (lp_stats.lower.compute) {
				report_lower(lp, lp_stats);
				if (lp_stats.upper.compute) {
					report_upper(lp, lp_stats);
					report_bound_quality(lp_stats);
				}
			} else if (lp_stats.upper.compute) {
				report_upper(lp, lp_stats);
			}
			if (lp_stats.primal.compute) {
				report_primal(lp_stats);
			}
			if (lp_stats.dual.compute) {
				report_dual(lp_stats);
			}
		}
	}

  if (lp_stats.rho_p.compute || lp_stats.rho_d.compute || lp_stats.condition.compute) {
    printf(" [*] verified condition measures\n");
    printf("\n");

    if (lp_stats.rho_p.compute) {
      lurupa.rho_p(lp, lp_stats.rho_p.lower, lp_stats.rho_p.upper);
      if (verbosity > 2) printf("\n");
      printf("     .---------------------.\n");
      printf("     | rho_p               |\n");
      printf("     +---------------------+-------------------------.\n");
      printf("     | lower bound         |       %s |\n",
						 rounded_string(lp_stats.rho_p.lower, LR_RND_DOWN));
      printf("     | upper bound         |       %s |\n", 
						 rounded_string(lp_stats.rho_p.upper, LR_RND_UP));
      printf("     `---------------------+-------------------------'\n");
      printf("\n");
    }

    if (lp_stats.rho_d.compute) {
      lurupa.rho_d(lp, lp_stats.rho_d.lower, lp_stats.rho_d.upper);
      if (verbosity > 2) printf("\n");
      printf("     .---------------------.\n");
      printf("     | rho_d               |\n");
      printf("     +---------------------+-------------------------.\n");
      printf("     | lower bound         |       %s |\n", 
						 rounded_string(lp_stats.rho_d.lower, LR_RND_DOWN));
      printf("     | upper bound         |       %s |\n", 
						 rounded_string(lp_stats.rho_d.upper, LR_RND_UP));
      printf("     `---------------------+-------------------------'\n");
      printf("\n");
    }

    if (lp_stats.condition.compute) {
      lurupa.cond(lp, lp_stats.condition.lower, lp_stats.condition.upper);
      if (verbosity > 2) printf("\n");
      printf("     .---------------------.\n");
      printf("     | condition number    |\n");
      printf("     +---------------------+-------------------------.\n");
      printf("     | lower bound         |       %s |\n", 
						 rounded_string(lp_stats.condition.lower, LR_RND_DOWN));
      printf("     | upper bound         |       %s |\n", 
						 rounded_string(lp_stats.condition.upper, LR_RND_UP));
      printf("     `---------------------+-------------------------'\n");
      printf("\n");
    }
  }

  stop_timer(end_pt, end_rt);
	lp_stats.total_time_status = timer_diff(start_total_pt, end_pt, start_total_rt, end_rt,
                                          lp_stats.total_proc_time, lp_stats.total_realtime);

  printf("======================================================\n");
  printf("\n");
  printf("total process time    : ");
  if (lp_stats.total_proc_time != -1) {
    printf("%.3gs %s\n", lp_stats.total_proc_time, lp_stats.total_time_status);
  } else {
    printf("error %s\n", lp_stats.total_time_status);
  }
  printf("total wall clock time : %.3gs\n", lp_stats.total_realtime);

  write_tables(tables, lp_stats, latex, csv);

  /* Finish. Cleanup. */
  timestamp = time(NULL);
  printf("\n");
  printf("end                   : %s", ctime(&timestamp));

  delete lp;
  if (tables & LU_T_LATEX)
    fclose(latex);
  if (tables & LU_T_CSV)
    fclose(csv);

  exit(EXIT_SUCCESS);
}

